//
// CifTokenizer.java,v 1.5 2003/05/05 20:58:07 dsg Exp
//
// Copyright 2001 The Regents of the University of California
// All Rights Reserved
//
// OpenMMS was developed by Dr. Douglas S. Greer at the San Diego
// Supercomputer Center, a research unit of the University of California,
// San Diego.  Support for this effort was provided by NSF through the
// Protein Data Bank (Grant DBI-9814284) and the National Partnership for
// Advanced Computational Infrastructure (Grant ACI-9619020)
//
// Permission to use, copy, modify and distribute any part of OpenMMS for
// educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above
// copyright notice, this paragraph and the following paragraphs appear
// in all copies.
//
// Those desiring to incorporate this OpenMMS into commercial products or
// use for commercial purposes should contact the Technology Transfer
// Office, University of California, San Diego, 9500 Gilman Drive, La
// Jolla, CA 92093-0910, Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// In no event shall the University of California be liable to any party
// for direct, indirect, special, incidental, or consequential damages,
// including lost profits, arising out of the use of this OpenMMS, even
// if the University of California has been advised of the possibility of
// such damage.
//
// The OpenMMS provided herein is on an "as is" basis, and the
// University of California has no obligation to provide maintenance,
// support, updates, enhancements, or modifications.  The University of
// California makes no representations and extends no warranties of any
// kind, either implied or express, including, but not limited to, the
// implied warranties of merchantability or fitness for a particular
// purpose, or that the use of the OpenMMS will not infringe any patent,
// trademark or other rights.

package org.rcsb.openmms.cifparse;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.zip.*;

/**
 * Class that reads the cif/STAR formatted file and
 * breaks it up into tokens at "word" boundaries.
 *
 * @author Douglas S. Greer
 * @version 1.5
 */
public class CifTokenizer
{
  public static final boolean debugTokenizer = false;

  private final static int UNKNOWN_REPOSITORY    = 0x1;
  private final static int LOCAL_FILE_REPOSITORY = 0x2;
  private final static int FTP_REPOSITORY        = 0x3;
  private final static int HTTP_REPOSITORY       = 0x4;
  private int rType = UNKNOWN_REPOSITORY;

  public boolean tokenIsName;

  private String filename;
  private BufferedReader br;
  private StreamTokenizer st;
  private StringBuffer sbuf;
  private int linenum;

  public CifTokenizer(String urlStr) throws IOException
    {
      if (urlStr.startsWith("file:/"))
	{
	  rType = LOCAL_FILE_REPOSITORY;
	  filename = urlStr.substring(6);
	}
      else if (urlStr.startsWith("ftp:/"))
	{
	  rType = FTP_REPOSITORY;
	  filename = urlStr.substring(5);
	}
      else if (urlStr.startsWith("http:/"))
	{
	  rType = HTTP_REPOSITORY;
	  filename = urlStr.substring(5);
	}
      else
	{
	  // If no schema prefix assume it is a directory path (may be a
	  // relative or absolute path)
	  rType = LOCAL_FILE_REPOSITORY;
	  filename = urlStr;
	}
      linenum = 1;
      sbuf = new StringBuffer();
      switch (rType)
	{
	case LOCAL_FILE_REPOSITORY:
	  if (filename.endsWith(".gz"))
	    {
	      ZipFile zf = new ZipFile(filename);
	      Enumeration ee = zf.entries();
	      ZipEntry ze = (ZipEntry) ee.nextElement();
	      if (ze == null)
		{
		  throw new IOException("No ZipEntries in Zip File");
		}
	      br = new BufferedReader(new InputStreamReader
				      (zf.getInputStream(ze)));
	    }
	  else
	    {
	      br = new BufferedReader(new FileReader(filename));
	    }
	  break;
	case FTP_REPOSITORY:
	case HTTP_REPOSITORY:
	  URL url = new URL(urlStr);
	  InputStream inputStream = url.openStream();

	  if (urlStr.endsWith(".zip"))
	    {
	      // URL: Zip
	      ZipInputStream zipInputStream =
		  new ZipInputStream(inputStream);
	      ZipEntry zipEntry = zipInputStream.getNextEntry();
	      if (zipEntry == null)
		throw new IOException("No ZipEntries in Zip stream");
	      InputStreamReader inputStreamReader =
		  new InputStreamReader(zipInputStream);
	      br = new BufferedReader(inputStreamReader);
	    }
	  else if (urlStr.endsWith(".gz"))
	    {
	      // URL: GZip
	      GZIPInputStream gzipInputStream =
		  new GZIPInputStream(inputStream);
	      InputStreamReader inputStreamReader =
		  new InputStreamReader(gzipInputStream);
	      br = new BufferedReader(inputStreamReader);
	    }
	  else if (urlStr.endsWith(".cif"))
	    {
	      // URL: No compression
	      InputStreamReader inputStreamReader =
		  new InputStreamReader(inputStream);
	      br = new BufferedReader(inputStreamReader);
	    }
	  break;
	}

      st = new StreamTokenizer(br);
      st.resetSyntax();

      // numbers are treated as words
      st.wordChars('!', '!');
      st.wordChars('$', '&');
      st.wordChars('(', '9');
      st.wordChars(':', ':');
      st.wordChars('<', '>');
      st.wordChars('?', '?');

      st.wordChars('@', '@');
      st.wordChars('A', 'Z');
      st.wordChars('[', '_');

      st.wordChars('`', '`');
      st.wordChars('a', 'z');
      st.wordChars('{', '~');
      
      // All white space plus ; ' and " chars are defaulted to
      // "ordinary" by the StreamTokenizer

      // st.commentChar('#'); - Doesn't work "#"s may occur in quoted strings
      st.eolIsSignificant(true);
    }

  public String filename()
    {
      return filename;
    }

  public int getLineNumber()
    {
      return linenum;
    }
  
  final static int NULL_STATE = 1;
  final static int TOKEN_STATE = 2;
  final static int REG_QUOTE_STATE = 3;
  final static int SEMI_QUOTE_STATE = 4;

  final static int WHITESPACE_CHAR = 1;
  final static int REG_QUOTE_CHAR = 2;
  final static int NEWLINE_CHAR = 3;
  final static int OTHER_CHAR = 4;
  
  public String getToken() throws IOException, CifParseException
    {
      boolean tokenIsQuoted = false;
      tokenIsName=false;

      int state = NULL_STATE;
      int lastCharType = OTHER_CHAR;

      int lchar;
      int quoteChar = 0;
      String token;
      int ttype = 0;

      // Exit token_found while loop when finished
    token_found:
      while (true)
	{
	  lchar = ttype;
	  ttype = st.nextToken();
	  switch(ttype)
	    {
	    case StreamTokenizer.TT_EOF:
	      token = null;
	      break token_found;
	    case StreamTokenizer.TT_WORD:
	      // "ordinary" word
	      switch (state)
		{
		case NULL_STATE:
		  // Token data found but it may be followed by ' " or ;
		  state = TOKEN_STATE;
		  break;
		case TOKEN_STATE:
		case REG_QUOTE_STATE:
		case SEMI_QUOTE_STATE:
		  // newline *not* followed by ;
		  if (lastCharType == NEWLINE_CHAR)
		    {
		      addToToken(lchar);
		    }
		  break;
		}
	      addToToken(st.sval);
	      lastCharType = OTHER_CHAR;
	      break; // end "ordinary" word
	    case '#':
	      // beginning of comment if not quoted
	      switch (state)
		{
		case NULL_STATE:
		case TOKEN_STATE:
		comment:
		  while (true)
		    {
		      // gobble comment
		      lchar = ttype;
		      ttype = st.nextToken();
		      switch(ttype)
			{
			case StreamTokenizer.TT_EOF:
			  token = null;
			  break token_found;
			case StreamTokenizer.TT_EOL:
			  st.pushBack();
			  break comment;
			default:
			  continue;
			}
		    }
		  break;
		case REG_QUOTE_STATE:
		case SEMI_QUOTE_STATE:
		  if (lastCharType == NEWLINE_CHAR)
		    {
		      addToToken(lchar);
		    }
		  addToToken(ttype);
		  break;
		}
	      lastCharType = OTHER_CHAR;
	      break;
	    case ' ':
	    case '\11':
	    case '\13':
	    case '\14':
	      // white space
	      switch (state)
		{
		case NULL_STATE:
		  // ignore multiple white space
		  break;
		case TOKEN_STATE:
		  lastCharType = WHITESPACE_CHAR;
		  token = endToken();
		  st.pushBack();
		  break token_found;
		case REG_QUOTE_STATE:
		  if (lastCharType == REG_QUOTE_CHAR)
		    {
		      lastCharType = WHITESPACE_CHAR;
		      token = endToken();
		      st.pushBack();
		      break token_found;
		    }
		  else
		    {
		      if (lastCharType == NEWLINE_CHAR)
			{
			  addToToken(lchar);
			}
		    }
		  addToToken(ttype);
		  break;
		case SEMI_QUOTE_STATE:
		  if (lastCharType == NEWLINE_CHAR)
		    {
		      addToToken(lchar);
		    }
		  addToToken(ttype);
		  break;
		}
	      lastCharType = WHITESPACE_CHAR;
	      break; // end white space
	    case '\'':
	    case '\"':
	      // quote
	      switch (state)
		{
		case NULL_STATE:
		  if (lastCharType == WHITESPACE_CHAR
		      || lastCharType == NEWLINE_CHAR)
		    {
		      quoteChar = ttype;
		      tokenIsQuoted = true;
		      state = REG_QUOTE_STATE;
		    }
		  break;
		case TOKEN_STATE:
		  addToToken(ttype);
		  break;
		case REG_QUOTE_STATE:
		  // don't match single and double quote chars
		  if (ttype == quoteChar)
		    {
		      if (lastCharType == REG_QUOTE_CHAR)
			{
			  addToToken(ttype);
			}
		      // prepare to end quote
		      lastCharType = REG_QUOTE_CHAR;
		    }
		  else
		    {
		      addToToken(ttype);
		    }
		  break;
		case SEMI_QUOTE_STATE:
		  addToToken(ttype);
		  break;
		}
	      // prepare to end quote
	      if (lastCharType != REG_QUOTE_CHAR)
		{
		  lastCharType = OTHER_CHAR;
		}
	      break; // quote
	    case ';':
	      // semi-quote
	      switch (state)
		{
		case NULL_STATE:
		  if (lastCharType == NEWLINE_CHAR)
		    {
		      tokenIsQuoted = true;
		      state = SEMI_QUOTE_STATE;
		    }
		  break;
		case TOKEN_STATE:
		  addToToken(ttype);
		  break;
		case REG_QUOTE_STATE:
		  addToToken(ttype);
		  break;
		case SEMI_QUOTE_STATE:
		  if (lastCharType == NEWLINE_CHAR)
		    {
		      lastCharType = OTHER_CHAR;
		      token = endToken();
		      break token_found;
		    }
		  else
		    {
		      addToToken(ttype);
		    }
		  break;
		}
	      lastCharType = OTHER_CHAR;
	      break;
	    case StreamTokenizer.TT_EOL:
	      // end-of-line
	      linenum++;
	      switch (state)
		{
		case NULL_STATE:
		  // ignore multiple white space
		  break;
		case TOKEN_STATE:
		  lastCharType = NEWLINE_CHAR;
		  token = endToken();
		  st.pushBack();
		  break token_found;
		case REG_QUOTE_STATE:
		  if (lastCharType == REG_QUOTE_CHAR)
		    {
		      lastCharType = NEWLINE_CHAR;
		      token = endToken();
		      st.pushBack();
		      break token_found;
		    }
		  else
		    {
		      if (lastCharType == NEWLINE_CHAR)
			{
			  addToToken(lchar);
			}
		      addToToken(st.sval);
		    }
		  break;
		case SEMI_QUOTE_STATE:
		  if (lastCharType == NEWLINE_CHAR)
		    {
		      addToToken(lchar);
		    }
		  break;
		}
	      lastCharType = NEWLINE_CHAR;
	      break; // end end-of-line
	    default:
	      switch (state)
		{
		case REG_QUOTE_STATE:
		case SEMI_QUOTE_STATE:
		  addToToken(ttype); // all's fair in quotes
		  break;
		case NULL_STATE:
		case TOKEN_STATE:
		  throw new CifParseException
		    ("CifTokenizer: Unrecognized token: 0x" +
		     Integer.toString(ttype, 16) + "(" 
		     + ((char) ttype) + ")");
		}
	      break; // end default
	    }
	} // end while(true)
      tokenIsName = token != null &&  ! tokenIsQuoted
	&& token.startsWith("_");
      if (debugTokenizer)
	{
	  printToken(token);
	}
      return token;
    }

  private void  addToToken(int c)
    {
      sbuf.append((char)c);
    }

  private void addToToken(String s)
    {
      sbuf.append(s);
    }

  private String endToken()
    {
      String token = sbuf.toString();
      sbuf = new StringBuffer();
      return token;
    }

  public void close()
    throws IOException 
    {
      // BufferedReader br;
      br.close();
      br = null;
      // StreamTokenizer st;
      st = null;
    }

  // DEBUG METHOD
  BufferedWriter tof;
  public void printToken(String token)
    throws IOException 
    {
      if (debugTokenizer)
	{
	  if (tof == null)
	    {
	      int li = filename.lastIndexOf('/');

	      if (li < 0) li = 0;
	      tof = new BufferedWriter(new FileWriter(
	         filename.substring(li+1, filename.length()) 
		 + "_tokens.txt"));
	    }
	  tof.newLine();
	  if (tokenIsName)
	    {
	      tof.write(":N:" + token);
	    }
	  else
	    {
	      tof.write(":::" + token);
	    }
	}
      tof.flush();
    }
}
