//
// FileEntryFoundry.java,v 1.11 2003/05/09 22:32:19 dsg Exp
//
// Copyright 2001 The Regents of the University of California
// All Rights Reserved
//
// OpenMMS was developed by Dr. Douglas S. Greer at the San Diego
// Supercomputer Center, a research unit of the University of California,
// San Diego.  Support for this effort was provided by NSF through the
// Protein Data Bank (Grant DBI-9814284) and the National Partnership for
// Advanced Computational Infrastructure (Grant ACI-9619020)
//
// Permission to use, copy, modify and distribute any part of OpenMMS for
// educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above
// copyright notice, this paragraph and the following paragraphs appear
// in all copies.
//
// Those desiring to incorporate this OpenMMS into commercial products or
// use for commercial purposes should contact the Technology Transfer
// Office, University of California, San Diego, 9500 Gilman Drive, La
// Jolla, CA 92093-0910, Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// In no event shall the University of California be liable to any party
// for direct, indirect, special, incidental, or consequential damages,
// including lost profits, arising out of the use of this OpenMMS, even
// if the University of California has been advised of the possibility of
// such damage.
//
// The OpenMMS provided herein is on an "as is" basis, and the
// University of California has no obligation to provide maintenance,
// support, updates, enhancements, or modifications.  The University of
// California makes no representations and extends no warranties of any
// kind, either implied or express, including, but not limited to, the
// implied warranties of merchantability or fitness for a particular
// purpose, or that the use of the OpenMMS will not infringe any patent,
// trademark or other rights.

package org.rcsb.openmms.entry;

import java.io.*;
import java.lang.reflect.*;
import java.net.*;
import java.text.*;
import java.util.*;
import java.util.zip.*;

import org.omg.CORBA.ORB;
import org.omg.DsLSRMacromolecularStructure.*;
import org.omg.ccm.BaseIDL.*;

import org.rcsb.openmms.cifparse.*;
import org.rcsb.openmms.util.*;
import org.rcsb.openmms.loader.*;

/**
 * Implements the EntryFactory for the MMS Corba reference server.
 *
 * @author Douglas S. Greer
 * @version 1.11
 */
public class FileEntryFoundry extends EntryFoundry
  implements EntryDirectory
{
  private MessageHandler mh;
  private String dataPath;
  private String dataURL;
  private String manifest;

  private final static int UNKNOWN_REPOSITORY = 0x1;
  private final static int FILE_REPOSITORY    = 0x2;
  private final static int FTP_REPOSITORY     = 0x3;
  private final static int HTTP_REPOSITORY    = 0x4;
  private int repositoryType = UNKNOWN_REPOSITORY;

  private DataItemList dil;

  // Since CifParser doesn't know anything about Entries, a cifparser
  // and a builder need to kept together as a pair; the RefBuilder
  // keeps track of the EntryImpl.
  private CifParser cparser;
  private RefBuilder buildr;

  // Entry Group data, Entries and ModificationDate structures
  private boolean inited = false;
  private String entries[];			// all entries
  private String[] entryGroups;			// all entry groups
  private HashMap entryGroupMap = new HashMap();
  private ModificationDate[] modDates;		// all entries with mod dates

  /**
   * Useful for applications that only wish to parse an mmCIF
   * datafile and create an EntryImpl
   */
  public static FileEntryFoundry createStandAloneFoundry
      (String dp, String manifest, ORB orb)
    {
      DataItem di;
      String itemName;

      // Bind DataItems to CatLoaders
      LinkTable ltable = new LinkTable();
      LoaderLink[] links = ltable.createLinkTable();
      DataItemList dilist = new DataItemList();

      for (int i = 0; i < links.length; i++)
	{
	  itemName = "_" + links[i].getCatName() + 
	    "." + links[i].getItemName();
	  di = new DataItem(itemName);
	  di.setFieldCode(links[i].getFieldCode());
	  di.setCatLoader(links[i].getCatLoader());
	  dilist.add(di);
	}
      ServerHandler servh = new ServerHandler(0, orb);
      servh.setMessageHandler(new PrintMessageHandler());
      // ?? should set a default message Handler (servh.setMessageHandler())
      return new FileEntryFoundry(servh, dilist, dp, manifest);
    }

  public FileEntryFoundry(ServerHandler a, DataItemList dl,
			  String dpURL, String manfst)
    {
      super(a);
      mh = servh.getMessageHandler();
      dataPath = "UNKNOWN";
      dataURL = "file:UNKNOWN";
      manifest = manfst;
      if (dpURL.startsWith("file:/"))
	{
	  repositoryType = FILE_REPOSITORY;
	  dataURL = dpURL;
	  dataPath = dpURL.substring(6);
	}
      else if (dpURL.startsWith("ftp:/"))
	{
	  repositoryType = FTP_REPOSITORY;
	  dataURL = dpURL;
	  dataPath = dpURL.substring(5);
	}
      else if (dpURL.startsWith("http:/"))
	{
	  repositoryType = HTTP_REPOSITORY;
	  dataURL = dpURL;
	  dataPath = dpURL.substring(6);
	}
      else
	{
	  // If no schema prefix assume it is a directory path (may be a
	  // relative or absolute path)
	  repositoryType = FILE_REPOSITORY;
	  dataURL = "file:/" + dpURL;
	  dataPath = dpURL;
	}
      dil = dl;
      buildr = new RefBuilder(mh);
      cparser = new CifParser();
      cparser.setBuilder(buildr);
    }

  public void setDebug(int d)
    {
      super.setDebug(d);
      buildr.setDebug(debug);
    }

  public boolean allVariablesRead()
    {
      return true;
    }

  public boolean cacheValue(EntryImpl ei, int modType,  int sFlag,
			    Object[] var)
    {
      return true;
    }

  public Object[] getList(EntryImpl ei, int modType, int sFlag)
    throws DataAccessException
    {
      // once a file is read all data is present, so for File Foundaries 
      // if this routine gets called there is no data for this category/item
      return null;
    }

  public int getListSize(EntryImpl ei, int modType, int sFlag)
    throws DataAccessException
    {
      // once a file is read all data is present, so for File Foundaries 
      // if this routine gets called there is no data for this category/item
      return -1;
    }

  public Object[] getListBlock(EntryImpl ei, int modType, int sFlag,
			       int from, int to)
    throws DataAccessException
    {
      // once a file is read all data is present, so for File Foundaries 
      // if this routine gets called there is no data for this category/item
      return null;
    }

  public EntryImpl createEntryImpl(String id)
   throws DataAccessException
    {
      String fname;

      try
	{
	  fname = findDataURL(id);
	}
      catch (DataAccessException cpe)
	{
	  EntryFoundry next = getNext();
	  if (next == null)
	    {
	      // We're at the end of the foundry queue -> Give up
	      String err = "FileEntryFoundry.createEntryImpl(): Entry " 
		+ id + " not found\n    Last data directory searched was: "
		+ dataPath;
		throw new DataAccessException("createEntryImpl", err);
	    }
	  else
	    {
	      return next.createEntryImpl(id);
	    }
	}
      // This does not actually read the data it just 
      // creates the entry.  The Cif file is read
      // the first time any of its contents are requested.
      EntryImpl ei = new EntryImpl(id, this, fname);
      return ei;
    }

  public void readEntryData(EntryImpl ei)
    throws DataAccessException
    {
      // file based foundries parse the whole file (all categories)
      readEntry(ei);
    }

  public String findDataURL(String id)
    throws DataAccessException
    {
      String fname = "unknown";
      switch (repositoryType)
	{
	case FILE_REPOSITORY:
	  fname = dataPath + "/" + id + ".cif";
	  File f = new File(fname);
	  if( ! f.exists())
	    {
	      // look for a compressed version
	      fname = fname + ".gz";
	      f = new File(fname);
	      if( ! f.exists())
		{
		  // try it with the Group-code
		  fname = dataPath + "/" + idToGroup(id) + "/" + id + ".cif";
		  f = new File(fname);
		  if( ! f.exists())
		    {
		      // look for a compressed version with a Group-code
		      fname = fname + ".gz";
		      f = new File(fname);
		      if( ! f.exists())
			{
			  // Give up
			  String err = 
			      "FileEntryFoundry.findDataURL(): Entry " 
			      + id + " not found\n";
			  throw new DataAccessException
			      ("findDataURL", err);
			}
		    }
		}
	    }
	  fname = "file:/" + fname;
	  break;
	case FTP_REPOSITORY:
	case HTTP_REPOSITORY:
	  // for now we will allways assume that the gzip'ed files are
	  // available
	  fname = dataURL + "/" + id + ".cif.gz";
	  break;
	}
      return fname;
    }

  public void readEntry(EntryImpl ei)
    throws DataAccessException
    {
      CifTokenizer ct = null;
      try
	{
	  String filename = ei.getPath();
	  
	  if (debug > 1)
	    {
	      servh.logMessage("FileEntryFoundry:Loading Entry " 
			      + ei.getEntryId() + " from file " + filename);
	    }
	  switch (repositoryType)
	    {
	    case FILE_REPOSITORY:
	      File f = new File(filename);
	      ei.setDepositionDate(new Date(f.lastModified()));
	      break;
	    case FTP_REPOSITORY:
	    case HTTP_REPOSITORY:
	      ei.setDepositionDate(new Date());
	      break;
	    }
	  ct = new CifTokenizer(filename);
	}
      catch(FileNotFoundException e)
	{
	  // This exception will occur anytime a someone tries to
	  // access a non-existant Entry
	  String err = "Entry "
	      + ei.getEntryId() + " not found.\n";
// 	  if (debug >= 0) 
// 	    {
// 	      servh.logMessage(err);
// 	    }
	  if (ct != null)
	    {
	      try
		{
		  ct.close();
		  ct = null;
		}
	      catch(IOException e2)
		{ }
	    }
	  
	  // throw a Corba DataAccess exception
	  // will come here on a file-not-found (no entry data) error
	  throw new DataAccessException("No such entry", err);
	}
      catch(Exception e)
	{	
	  String err = "Access error for Entry Id: "
	      + ei.getEntryId() + "\n  Exception: " + e.toString();
	  if (debug >= 0) 
	    {
	      servh.logMessage(err);
	    }
	  if (ct != null)
	    {
	      try
		{
		  ct.close();
		  ct = null;
		}
	      catch(IOException e2)
		{ }
	    }
	  // throw a Corba DataAccess exception
	  // will come here on a file-not-found (no entry data) error
	  throw new DataAccessException("readEntry", err);
	}
      try
	{
	  buildr.setEntry(ei);
	  cparser.readDataBlock(ct, dil);
	  buildr.setEntry(null);
	}
      catch(CifParseException e)
	{
	  // something appears to be incorrect inside an mmCIF file
	  String err = "CifParseException during readDataBlock\n"
	    + "\tID: " + ei.getEntryId() + "line" + ct.getLineNumber() + "\n";
	  if (debug > 0)
	    {
	      servh.logMessage(err + e);
	      System.err.println(err);
	      e.printStackTrace();
	    }
	  // throw a Corba DataAccess exception
	  throw new DataAccessException("readEntry", err);
	}
      catch (Throwable e)
	{
	  // something is seriously wrong 
	  String err = "Exception during readDataBlock\n"
	    + "\tID: " + ei.getEntryId() + "\nThrowable caught=" + e + ";\n";
	  if (debug > 0)
	    {
	      servh.logMessage(err);
	      System.err.println(err);
	      e.printStackTrace();
	    }
	  // throw a Corba DataAccess exception
	  throw new DataAccessException("readEntry", err);
	}
      finally
	{
	  if (ct != null)
	    {
	      try
		{
		  ct.close();
		}
	      catch(IOException e2)
		{ }
	    }
	  ct = null;
	}
    }

  protected void checkDirectoryInit()
    throws DataAccessException
    {
      if ( ! inited)
	{
	  initEntryLists();
	  inited = true;
	}
    }

  private void initEntryLists()
    throws DataAccessException
    {
      ArrayList eall = new ArrayList();
      ArrayList eg = new ArrayList();
      ArrayList edate = new ArrayList();
      int i, j;
      long mdate;
      String gdir;
      String en;
      File gfile, of;
      String[] fs;
      ModificationDate mdi;

      String datDir = dataPath;
      try
	{
	  // read groups and entries
	  switch (repositoryType)
	    {
	    case FILE_REPOSITORY:
	      File dFile = new File(datDir);
	      if ( ! dFile.isDirectory())
		{
		  throw new IOException("Data dir not a directory");
		}
	      entryGroups = dFile.list();
	      if (debug >= 7) 
		{
		  for (i = 0; i < entryGroups.length; i++)
		    {
		      mh.logMessage
			  ("Group[" + i + "] = " + entryGroups[i]);
		    }
		}
	      for (i = 0; i < entryGroups.length; i++)
		{
		  gdir = datDir + "/" + entryGroups[i];
		  gfile = new File(gdir);
		  if ( ! gfile.isDirectory())
		    {
		      mh.logMessage("Entry Group dir " 
				    + entryGroups[i] + " is not a directory");
		      continue;
		    }
		  fs = gfile.list();
		  for (j = 0; j < fs.length; j++)
		    {
		      if (fs[j].endsWith(".cif"))
			{
			  // Entry Found.
			  // have to open each file to find the 
			  // modification date
			  of = new File(gdir + "/" + fs[j]);
			  mdate = of.lastModified();
			  en = fs[j].substring(0, fs[j].length() - 4);
			  eall.add(en);
			  eg.add(en);
			  
			  // For Valuetypes
			  // mdi = new ModificationDateImpl();
			  
			  // For structs
			  mdi = new ModificationDate();
			  
			  mdi.entry_id = en;
			  mdi.date = mdate;
			  edate.add(mdi);
			}
		    }
		  // create a string array of all entries in this group
		  fs = new String[eg.size()];
		  for (j = 0; j < eg.size(); j++)
		    {
		      fs[j] = (String) eg.get(j);
		    }
		  entryGroupMap.put(entryGroups[i], fs);
		  eg.clear();
		}
	      // build array for get-all-entries method
	      entries = new String[eall.size()];
	      for (i = 0; i < eall.size(); i++)
		{
		  entries[i] = (String) eall.get(i);
		}
	      // build array for get-modification-dates method
	      modDates = new ModificationDate[edate.size()];
	      for (i = 0; i < edate.size(); i++)
		{
		  modDates[i] = (ModificationDate) edate.get(i);
		}
	      if (modDates.length <= 0)
		{
		  throw new IOException("No Entries Found");
		}
	      if (debug >= 7)
		{
		  i = Math.min(modDates.length-1, 147);
		  Date dt = new Date();
		  dt.setTime(modDates[i].date);
		  mh.logMessage
		      ("Entry[" + i + "] = " + modDates[i].entry_id
		       + " modified " + dt.toString());
		}
	      break;
	    case FTP_REPOSITORY:
	    case HTTP_REPOSITORY:
	      HashSet nullHashSet = new HashSet();
	      entries = readManifest(manifest, nullHashSet);
	      // make a single empty entry to avoid null value CORBA exceptions
	      entryGroups = new String[1];
	      entryGroups[0] = "";
	      // make a single empty entry to avoid null value CORBA exceptions
	      mdi = new ModificationDate();
	      mdi.entry_id = "";
	      mdi.date = 0;
	      modDates = new ModificationDate[1];
	      modDates[0] = mdi;
	      break;
	    }
	}
      catch(IOException ioe)
	{
	  String err = "Cannot read Data Directory: " + datDir 
	    + "\n\tIOException: " + ioe;
	  ioe.printStackTrace();
	    if (debug >= 0) 
	      {
		mh.logMessage(err);
	      }
	    // throw a Corba DataAccess exception
	    throw new DataAccessException("initEntryLists", err);
	}
    }

  public static String[] readManifest(String manifest, HashSet exclude)
      throws IOException, MalformedURLException
    {
      String s, id;
      ArrayList es = new ArrayList();
      
      URL manUrl = new URL(manifest);
      InputStreamReader inSReader 
	  = new InputStreamReader(manUrl.openStream());
      LineNumberReader lnr = new LineNumberReader
	  (new BufferedReader(inSReader));
      while ((s = lnr.readLine()) != null)
	{
	  // skip over any symbolic links
	  if (s.indexOf('>') >= 0)
	    {
	      continue;
	    }
	  int endid = s.indexOf(".cif.Z");
	  id = null;
	  if (endid >= 20)
	    {
	      id = s.substring(endid-4, endid);
	      // have to special case this because of "Components.cif.Z" file
	      if (id.equalsIgnoreCase("ents"))
	      {
		continue;
	      }
	      if ( ! exclude.contains(id) && ! exclude.contains(id + ".cif"))
		{
		  es.add(id);
		}
	    }
	}
      // build array for get-all-entries method
      String[] entryList = new String[es.size()];
      for (int i = 0; i < es.size(); i++)
	{
	  entryList[i] = (String) es.get(i);
	}
      return entryList;
    }

  public String[] getEntryIdList()
    throws DataAccessException
    {
      checkDirectoryInit();
      return entries;
    }

  public ModificationDate[] getEntryModificationDates()
    throws DataAccessException
    {
      checkDirectoryInit();
      return modDates;
    }

  public String[] getEntryGroupList()
    throws DataAccessException
    {
      checkDirectoryInit();
      return entryGroups;
    }

  public String[] getEntriesInGroup(String group)
    throws DataAccessException
    {
      String[] rval = null;

      checkDirectoryInit();
      java.lang.Object o = entryGroupMap.get(group);
      if (o == null)
	{
	  // This exception will occur anytime a someone tries to
	  // access a non-existant Entry Group
	  String err = "Non-existant Entry Group: "
	    + group + "\n";
	    if (debug >= 2) 
	      {
		mh.logMessage(err);
	      }
	    // throw a Corba DataAccess exception
	    throw new DataAccessException("get_entries_in_group", err);
	}
      else
	{
	  rval = (String[]) o;
	}
      return rval;
    }

  public String[] getNativeFormatsSupported()
    {
      String[] r = new String[1];
      switch (repositoryType)
	{
	case FILE_REPOSITORY:
	  r[0] = "mmCIF";
	  break;
	case FTP_REPOSITORY:
	case HTTP_REPOSITORY:
	  // If we already have a ftp or http repository we don't really need
	  // to provide additional native format file retrieval support.
	  r[0] = "<NONE>";
	  break;
	}
      return r;
    }

  public byte[] getNativeEntryRepresentation(String format,
					     String entry_id)
    throws DataAccessException
    {
      if ( ! format.equalsIgnoreCase("mmCIF"))
	{
	  // unsupported native format
	  String err = "Native format type "
	    + format + "not supported\n";
	    if (debug >= 2) 
	      {
		mh.logMessage(err);
	      }
	    // throw a Corba DataAccess exception
	    throw new DataAccessException
	      ("getNativeEntryRepresentation", err);
	}
      byte buf[] = null;
      try
	{
	  String fname = findDataURL(entry_id);
	  switch (repositoryType)
	    {
	    case FILE_REPOSITORY:
	      // strip off "file:/"
	      fname = fname.substring(6);
	      File f = new File(fname);
	      if( ! f.exists())
		{
		  // This exception will occur anytime a someone tries to
		  // access a non-existant Entry
		  String err = "Native format error for Entry Id: "
		      + entry_id + " - Entry not found\n";
		  if (debug >= 2) 
		    {
		      mh.logMessage(err);
		    }
		  // throw a Corba DataAccess exception
		  throw new DataAccessException
		      ("getNativeEntryRepresentation", err);
		}
	      buf = new byte[(int)f.length()];
	      FileInputStream in = new FileInputStream(f);
	      in.read(buf);
	      break;
	    case FTP_REPOSITORY:
	    case HTTP_REPOSITORY:
	      // not supported -- just use the ftp or http servers
	      break;
	    }
	}
      catch (Exception e)
	{
	  // Probably an IO error
	  String err = "Native format error for Entry Id: "
	    + entry_id + " - Cannot access entry\n";
	    if (debug >= 2) 
	      {
		mh.logMessage(err);
	      }
	    // throw a Corba DataAccess exception
	    throw new DataAccessException
	      ("getNativeEntryRepresentation", err);
	}
      return buf;
    }
}
