//
// RefBuilder.java,v 1.3 2003/05/05 20:58:15 dsg Exp
//
// Copyright 2001 The Regents of the University of California
// All Rights Reserved
//
// OpenMMS was developed by Dr. Douglas S. Greer at the San Diego
// Supercomputer Center, a research unit of the University of California,
// San Diego.  Support for this effort was provided by NSF through the
// Protein Data Bank (Grant DBI-9814284) and the National Partnership for
// Advanced Computational Infrastructure (Grant ACI-9619020)
//
// Permission to use, copy, modify and distribute any part of OpenMMS for
// educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above
// copyright notice, this paragraph and the following paragraphs appear
// in all copies.
//
// Those desiring to incorporate this OpenMMS into commercial products or
// use for commercial purposes should contact the Technology Transfer
// Office, University of California, San Diego, 9500 Gilman Drive, La
// Jolla, CA 92093-0910, Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// In no event shall the University of California be liable to any party
// for direct, indirect, special, incidental, or consequential damages,
// including lost profits, arising out of the use of this OpenMMS, even
// if the University of California has been advised of the possibility of
// such damage.
//
// The OpenMMS provided herein is on an "as is" basis, and the
// University of California has no obligation to provide maintenance,
// support, updates, enhancements, or modifications.  The University of
// California makes no representations and extends no warranties of any
// kind, either implied or express, including, but not limited to, the
// implied warranties of merchantability or fitness for a particular
// purpose, or that the use of the OpenMMS will not infringe any patent,
// trademark or other rights.

package org.rcsb.openmms.entry;

import java.util.*;
import java.io.*;

import org.rcsb.openmms.cifparse.*;
import org.rcsb.openmms.util.*;

/**
 * Reference Builder Class that parses an mmCIF data file and
 * calls the appropriate CatLoader for each Category contained
 * in the file to load corresponding the Entry data structure.
 *
 * @author Douglas S. Greer
 * @version 1.3
 */
public class RefBuilder implements Builder
{
  private static boolean trace;
  private static int debug = 0;

  MessageHandler mHandler;

  DataItem singleDataItem;  // for single data values outside loops
  DataItem loopDataItem;    // for loop construct data values

  EntryImpl xEntry;
  private ArrayList catArrayList = new ArrayList();
  private CatLoader lastCat = null;

  public RefBuilder(MessageHandler mh)
    {
      mHandler = mh;

    }

  public void setDebug(int d)
    {
      debug = d;
    }

  public void setEntry(EntryImpl ei)
    {
      xEntry = ei;
    }
  
//   public EntryImpl getEntry()
//     {
//       return xEntry;
//     }

  //
  // Builder Methods called by CifParser
  //
  public void beginCompound()
    throws CifParseException
    {
      if (trace)
	{
	  mHandler.logMessage("RefBuilder.beginCompound()");
	}
      if (xEntry == null)
	{
	  throw new CifParseException
	    ("ReBuilder.beginCompound(): No target Entry specified)");
	}
      catArrayList.clear();
      lastCat = null;
      loopDataItem = null;
      singleDataItem = null;
    }

  public void endCompound()
    throws CifParseException
    {
      CatLoader cl;

      if (trace)
	{
	  mHandler.logMessage("RefBuilder.endCompound()");
	}
      for (int i = 0; i < catArrayList.size(); i++)
	{
	  cl = (CatLoader) catArrayList.get(i);
	  cl.endCompound(xEntry);
	}
      catArrayList.clear();
    }

  private void checkCatLoader(CatLoader cl)
    throws CifParseException
    {
      if (cl != lastCat)
	{
	  cl.beginCategory();
	  catArrayList.add(cl);
	  lastCat = cl;
	}
    }

  public void setSingleItem(DataItem di)
    throws CifParseException
    {
      CatLoader cl;

      if (trace)
	{
	  mHandler.logMessage
	    ("RefBuilder.setSingleValue(): " + di.getItemName());
	}
      singleDataItem = di;
      loopDataItem = null;
      cl = di.getCatLoader();
      if (cl != null)
	{
	  checkCatLoader(cl);
	  cl.defineSingleItem(xEntry, di.getFieldCode());
	}
    }

  public void insertSingleValue(DataItem d, String val) 
    throws CifParseException
    {
      if (trace)
	{
	  mHandler.logMessage
	    ("RefBuilder.insertSingleValue(): "
	     + d.getItemName() + ":" + val);
	}
      if (singleDataItem == null)
	{
	  throw new CifParseException
	    ("ReBuilder.insertSingleValue(): No valid DataItem");
	}
      else
	{
	  if (singleDataItem.getCatLoader() != null)
	    {
	      singleDataItem.getCatLoader().insertSingleValue
		(xEntry, singleDataItem.getFieldCode(), val);
	    }
	}
    }

  public void beginLoop()
    throws CifParseException
    {
      if (trace)
	{
	  mHandler.logMessage("RefBuilder.beginLoop()");
	}
      loopDataItem = null;
      singleDataItem = null;
    }

  public void endLoop()
    throws CifParseException
    {
      if (trace)
	{
	  mHandler.logMessage("RefBuilder.endLoop()");
	}
      if (loopDataItem != null &&
	  loopDataItem.getCatLoader() != null)
	{
	  loopDataItem.getCatLoader().endLoop(xEntry);
	}
      loopDataItem = null;
    }

  public void beginRow()
    throws CifParseException
    {
      if (trace)
	{
	  mHandler.logMessage("RefBuilder.beginRow()");
	}
      if (loopDataItem != null &&
	  loopDataItem.getCatLoader() != null)
	{
	  loopDataItem.getCatLoader().beginRow();
	}
    }

  public void endRow()
    throws CifParseException
    {
      if (trace)
	{
	  mHandler.logMessage("RefBuilder.endRow()");
	}
      if (loopDataItem != null &&
	  loopDataItem.getCatLoader() != null)
	{
	  loopDataItem.getCatLoader().endRow();
	}
    }

  public void setLoopItem(int column, DataItem di)
    throws CifParseException
    {
      CatLoader cl;

      if (trace)
	{
	  String iname = "(Null DataItem)";
	  if (di != null)
	    {
	      iname = di.getItemName();
	    }
	  mHandler.logMessage("RefBuilder.setLoopItem(): " + iname);
	}
      if (loopDataItem == null && di != null ) 
	{
	  loopDataItem = di;
	}
      if (di != null && (cl = di.getCatLoader()) != null)
	{
	  checkCatLoader(cl);
	  cl.defineLoopItem(xEntry, di.getFieldCode());
	}
    }

  public void insertLoopValue(int column, DataItem d, String val)
    throws CifParseException
    {
      if (trace)
	{
	  mHandler.logMessage
	    ("RefBuilder.insertLoopValue(): " + d.getItemName() + ":" + val);
	}
      if (loopDataItem == null)
	{
	  throw new CifParseException
	    ("ReBuilder.insertLoopValue(): no valid LoopItem");
	}
      if (d.getCatLoader() != null)
	{
	  d.getCatLoader().insertLoopValue(d.getFieldCode(), val);
	}
    }
}
