//
// GeneralCategory.java,v 1.4 2003/05/28 01:09:03 dsg Exp
//
// Copyright 2002 The Regents of the University of California
// All Rights Reserved
//
// OpenMMS was developed by Dr. Douglas S. Greer at the San Diego
// Supercomputer Center, a research unit of the University of California,
// San Diego.  Support for this effort was provided by NSF through the
// Protein Data Bank (Grant DBI-9814284) and the National Partnership for
// Advanced Computational Infrastructure (Grant ACI-9619020)
//
// Permission to use, copy, modify and distribute any part of OpenMMS for
// educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above
// copyright notice, this paragraph and the following paragraphs appear
// in all copies.
//
// Those desiring to incorporate this OpenMMS into commercial products or
// use for commercial purposes should contact the Technology Transfer
// Office, University of California, San Diego, 9500 Gilman Drive, La
// Jolla, CA 92093-0910, Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// In no event shall the University of California be liable to any party
// for direct, indirect, special, incidental, or consequential damages,
// including lost profits, arising out of the use of this OpenMMS, even
// if the University of California has been advised of the possibility of
// such damage.
//
// The OpenMMS provided herein is on an "as is" basis, and the
// University of California has no obligation to provide maintenance,
// support, updates, enhancements, or modifications.  The University of
// California makes no representations and extends no warranties of any
// kind, either implied or express, including, but not limited to, the
// implied warranties of merchantability or fitness for a particular
// purpose, or that the use of the OpenMMS will not infringe any patent,
// trademark or other rights.

package org.rcsb.openmms.entry;

import java.sql.*;

import org.omg.CORBA.ORB;
import org.omg.DsLSRMacromolecularStructure.*;
import org.omg.DsLSRMmsReference.*;
import org.omg.DsLSRXRayCrystallography.*;

import org.rcsb.openmms.util.*;
import org.rcsb.openmms.dbinsert.*;
import org.rcsb.openmms.meta.*;
import org.rcsb.openmms.loader.*;

/**
 * Customizable class to interface to non-OpenMMS databases or
 * to compute specific category data.
 *
 * @author Douglas S. Greer
 * @version 1.4
 */
public class GeneralCategory implements TypeNamesSql
{
  protected Connection conn;
  protected MessageHandler messagePanel;
  protected SelectCategoryFinal scf;

  protected Statement stmt;

  public GeneralCategory(MessageHandler ps, Connection c,
				SelectCategoryFinal s)
    {
      messagePanel = ps;
      conn = c;
      scf = s;
    }

  public void readPresenceFlags(EntryImpl ei)
    throws DataAccessException
    {
      // Retreive Flags for fields that will be populated from the database.
      scf.readPresenceFlags(ei);

      // "OR" in any categories or items in the Entry that will be processed
      // or computed here but may not be retrieved by SelectCategoryFoundry.
      // If the Flag is set, then some data records *must* be provided
      // in order to remain compliant with the OMG Corba specification.

      // TEST DATA
      // This Test Code helps to illustrate how add presence flags for data that
      // may be missing from the database but could be computed on the fly
      // as needed for categories such as GEOM_*, CHEM_LINK_* or ATOM_TYPE
      byte[] pf  = ei.getPresenceFlags(ModuleTypes.MMS_CORE);
      byte[] rpf = ei.getPresenceFlags(ModuleTypes.MMS_REFERENCE);
      byte[] xpf = ei.getPresenceFlags(ModuleTypes.MMS_XRAY_CRYSTALLOGRAPHY);
      byte[] npf = ei.getPresenceFlags
	  (ModuleTypes.MMS_NUCLEAR_MAGNETIC_RESONANCE);

      // MmsReference Module Flags
      scf.setFlag(rpf, MmsReferenceEntry.S_SOFTWARE);
      scf.setFlag(rpf, MmsReferenceEntry.F_SOFTWARE_CLASSIFICATION);
      scf.setFlag(rpf, MmsReferenceEntry.F_SOFTWARE_CONTACT_AUTHOR);
      scf.setFlag(rpf, MmsReferenceEntry.F_SOFTWARE_CONTACT_AUTHOR_EMAIL);
      scf.setFlag(rpf, MmsReferenceEntry.F_SOFTWARE_LOCATION);
    }

  public Object[] getList(EntryImpl ei, int modType, int sFlag)
    throws DataAccessException, SQLException
    {
      Object[] oa = null;
      DataAccessException caught = null;

      try
	{
	  oa = scf.getList(ei, modType, sFlag);
	}
      catch (DataAccessException e)
	{
	  // re-throw if we do not generate this data below
	  caught = e;
	}

      if (oa == null)
	{
	  switch (modType)
	    {
	    case ModuleTypes.MMS_CORE:
	      switch (sFlag)
		{
		}
	      break;
	    case ModuleTypes.MMS_XRAY_CRYSTALLOGRAPHY:
	      switch (sFlag)
		{
		}
	      break;
	    case ModuleTypes.MMS_NUCLEAR_MAGNETIC_RESONANCE:
	      switch (sFlag)
		{
		}
	      break;
	    case ModuleTypes.MMS_REFERENCE:
	      switch (sFlag)
		{
		case MmsReferenceEntry.S_SOFTWARE:
		  oa = read_SOFTWARE_list();
		  break;
		}
	      break;
	    }
	}
      if (oa == null && caught != null)
	{
	  throw caught;
	}
      return oa;
    }

  public Object[] getListBlock(EntryImpl ei, int modType,
			       int sFlag, int from, int to)
    throws DataAccessException, SQLException
    {
      Object[] oa = null;
      DataAccessException caught = null;

      try
	{
	  oa = scf.getListBlock(ei, modType, sFlag, from, to);
	}
      catch (DataAccessException e)
	{
	  // re-throw if we do not generate this data below
	  caught = e;
	}

      if (oa == null)
	{
	  switch (modType)
	    {
	    case ModuleTypes.MMS_CORE:
	      switch (sFlag)
		{
		}
	      break;
	    case ModuleTypes.MMS_XRAY_CRYSTALLOGRAPHY:
	      switch (sFlag)
		{
		}
	      break;
	    case ModuleTypes.MMS_NUCLEAR_MAGNETIC_RESONANCE:
	      switch (sFlag)
		{
		}
	      break;
	    case ModuleTypes.MMS_REFERENCE:
	      switch (sFlag)
		{
		case MmsReferenceEntry.S_SOFTWARE:
		  if (from < 1 && to >= 1)
		    {
		      oa = read_SOFTWARE_list();
		    }
		  break;
		}
	      break;
	    }
	}

      if (oa == null && caught != null)
	{
	  throw caught;
	}
      return oa;
    }

  public int getListSize(EntryImpl ei, int modType, int sFlag)
    throws DataAccessException, SQLException
    {
      int sz = -1;
      DataAccessException caught = null;

      try
	{
	  sz = scf.getListSize(ei, modType, sFlag);
	}
      catch (DataAccessException e)
	{
	  // re-throw if we do not generate this data below
	  caught = e;
	}

      if (sz == -1)
	{
	  switch (modType)
	    {
	    case ModuleTypes.MMS_CORE:
	      switch (sFlag)
		{
		}
	      break;
	    case ModuleTypes.MMS_XRAY_CRYSTALLOGRAPHY:
	      switch (sFlag)
		{
		}
	      break;
	    case ModuleTypes.MMS_NUCLEAR_MAGNETIC_RESONANCE:
	      switch (sFlag)
		{
		}
	      break;
	    case ModuleTypes.MMS_REFERENCE:
	      switch (sFlag)
		{
		case MmsReferenceEntry.S_SOFTWARE:
		  // TEST DATA
		  // This Test Code shows how data that is missing from
		  // the database can be computed on the fly. E.g. for
		  // categories such as  GEOM_* or CHEM_LINK_*
		  // This code is only executed if the database contains
		  // no data for this category (sz == -1)
		  sz = 1;
		  break;
		}
	      break;
	    }
	}
      if (sz == -1 && caught != null)
	{
	  throw caught;
	}
      return sz;
    }
  
  private SoftwareImpl[] read_SOFTWARE_list()
    throws SQLException 
    {
      // TEST DATA
      // This Test Code shows how data that is missing from the database can
      // be computed on the fly as needed e.g for the GEOM_*, CHEM_LINK_* or
      // ATOM_TYPE categories.
      // This code is only executed if the database contains
      // no data for this category (oa == null)
      SoftwareImpl[] el = new SoftwareImpl[1];
      el[0] = new SoftwareImpl();

      el[0].compiler_name = "";
      el[0].compiler_version = "";
      el[0].date = "";
      el[0].description = "";
      el[0].dependencies = "";
      el[0].hardware = "";
      el[0].language = "";
      el[0].location = "";
      el[0].mods = "";
      el[0].name = "";
      el[0].os = "";
      el[0].os_version = "";
      el[0].type = "";

      el[0].name = "OpenMMS";
      el[0].location = "http://openmms.sdsc.edu";
      el[0].classification = "Corba Server";
      el[0].citation = new IndexId();
      el[0].citation.id = "Test Data Citation";
      el[0].citation.index = -1;
      el[0].contact_author = "Douglas S. Greer";
      el[0].contact_author_email = "dsg@sdsc.edu";
      el[0].version = EntryFactory.MODEL_IDENTIFIER + "_"
	+ EntryFactory.REVISION_NUMBER;
      return el;
    }

  public boolean isFlagSet(byte[] bits, int flag)
    {
      return scf.isFlagSet(bits, flag);
    }
}


