//
// CatWriter.java,v 1.7 2003/05/30 18:57:01 dsg Exp
//
// Copyright 2001 The Regents of the University of California
// All Rights Reserved
//
// OpenMMS was developed by Dr. Douglas S. Greer at the San Diego
// Supercomputer Center, a research unit of the University of California,
// San Diego.  Support for this effort was provided by NSF through the
// Protein Data Bank (Grant DBI-9814284) and the National Partnership for
// Advanced Computational Infrastructure (Grant ACI-9619020)
//
// Permission to use, copy, modify and distribute any part of OpenMMS for
// educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above
// copyright notice, this paragraph and the following paragraphs appear
// in all copies.
//
// Those desiring to incorporate this OpenMMS into commercial products or
// use for commercial purposes should contact the Technology Transfer
// Office, University of California, San Diego, 9500 Gilman Drive, La
// Jolla, CA 92093-0910, Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// In no event shall the University of California be liable to any party
// for direct, indirect, special, incidental, or consequential damages,
// including lost profits, arising out of the use of this OpenMMS, even
// if the University of California has been advised of the possibility of
// such damage.
//
// The OpenMMS provided herein is on an "as is" basis, and the
// University of California has no obligation to provide maintenance,
// support, updates, enhancements, or modifications.  The University of
// California makes no representations and extends no warranties of any
// kind, either implied or express, including, but not limited to, the
// implied warranties of merchantability or fitness for a particular
// purpose, or that the use of the OpenMMS will not infringe any patent,
// trademark or other rights.

package org.rcsb.openmms.meta;

import java.util.*;
import java.io.*;

import org.rcsb.openmms.cifparse.*;

/**
 * Metamodel visitor that generates the Java code for the
 * category loaders which process
 * the parsed string data in an mmCIF file and place
 * it into a structure specific for that mmCIF category.
 *
 * This code is perhaps best understood by simultaneously
 * reviewing an example
 * category loader it generates.  E.g. AtomSiteCatLoader.java
 *
 * The public methods called by the mmCIF builder ( see the
 * CifParser.java and Builder.java) code come first in this
 * file.  The various private methods that follow write sections
 * to Character buffers that are then writen to the Category loader
 * file when closeCategory() is called.
 *
 * @see org.rcsb.openmms.cifparse.Builder
 * @see org.rcsb.openmms.cifparse.CifParser
 * @see org.rcsb.openmms.loader.AtomSiteCatLoader
 *
 * @author Douglas S. Greer
 * @version 1.7
 */
public class CatWriter extends TransVisitor
{
  private static boolean trace = false;

  final String loaderPackage = "loader";
  String loaderPath;
  String catBaseName;
  String interfaceName;
  String lsrPackageName, msaPackageName;

  RootTrans inRoot;
  ModuleTrans inMod;
  String moduleName;
  String coreModuleName = "";
  String baseModuleName = "";
  Stack moduleNameStack = new Stack();
  String modVarName;
  String modEntryName;

  String oClassName;
  String oClassFilename;
  CifDictionary cifDic;
  int currentFieldCode = 1;
  
  StringStruct  initBuffer;
  StringStruct  tagBuffer;
  StringStruct  beginEndBuffer;
  StringStruct  defineBuffer;
  StringStruct  insert1Buffer;
  StringStruct  insert2Buffer;
  StringStruct  finBuffer;
  
  ArrayList catStructs;
  ArrayList indexFields;

  String hashTableName;
  CategoryTrans CompositeKeyCat;
  ArrayList atomIndices;
  ArrayList seqIndices;

  public CatWriter(CifDictionary d, String ldp, String lsr, String msa,
		   String cname)
    {
      setIndentString("   ");  // set indent size to three spaces
      
      cifDic = d;
      loaderPath = ldp;

      lsrPackageName = lsr;
      msaPackageName = msa;

      catBaseName = cname;
      
      catStructs = new ArrayList();
      indexFields = new ArrayList();

      // special cased due to composite seq_id key
      atomIndices = new ArrayList();
      seqIndices = new ArrayList();
    }
  
  public void setTrace(boolean t)
    {
      trace = t;
    }

  public void openRoot(RootTrans rt)
    throws IOException, TransGenException
    {
      inRoot = rt;

      // check all the variables
      if (catBaseName == null)
	{
	  throw new TransGenException("CatWriter: Null CatLoader name");
	}
      if (loaderPath == null)
	{
	  throw new TransGenException(
	      "CatWriter: loader target directoy path not specified");
	}
    }

  public void openModule(ModuleTrans mt) 
    throws IOException, TransGenException
    {
      inMod = mt;
      moduleName = "";
      Stack scp;

      ModuleTrans cMod= mt.getCoreModule();
      if (cMod == null)
	{
	  coreModuleName = mt.getName();
	}
      else
	{
	  coreModuleName = cMod.getName();
	}

      modVarName = mt.getVarName();
      if (modVarName.length() > 0)
	{
	  modVarName += ".";
	}
      modEntryName = mt.getEntryName();
      if (moduleNameStack.size() == 0)
	{
	  baseModuleName = mt.getName();
	}
      moduleNameStack.push(mt.getName());
      scp = (Stack) moduleNameStack.clone();
      
      while ( ! scp.empty())
	{
	  moduleName = ((String) scp.pop()) + "." +  moduleName;
	}
    }
      
  public void closeModule(ModuleTrans mt) 
    throws IOException, TransGenException
    {
      moduleNameStack.pop();
    }

  public void openCategory(CategoryTrans ct) 
    throws IOException, TransGenException
    {
      if (trace)
	{
	  System.out.println("CatWriter.openCategory called");
	}

      indentLevel = 0;
      catStructs.clear();
      indexFields.clear();
      atomIndices.clear();
      seqIndices.clear();
      CompositeKeyCat = isCompositeKeyCat(ct) ? ct : null;

      // begin writing to "segmented" set of char buffers
      initCharBuffers(ct);

      writeCatFileOpening(ct);
      writeDefineMethods();
      writeSwitchOpenings();
    }

  private boolean isCompositeKeyCat(CategoryTrans ct) 
    {
      // The only composite key at present is EntityPolySeq (seq_id)
      boolean rv = ct.getName().equalsIgnoreCase("entity_poly_seq");
      return rv; 
    }

  public void openStructure(StructureTrans st) 
    throws IOException, TransGenException
    {
      if (trace)
	{
	  System.out.println("CatWriter.openStructure called");
	}
      if (st.isCifStructure())
	{
	  catStructs.add(st);
	}
    }

  public void openValuetype(ValuetypeTrans vt) 
    throws IOException, TransGenException
    {
      openStructure(vt);
    }

  public void openField(FieldTrans ft) 
    throws IOException, TransGenException
    {
      if (trace)
	{
	  System.out.println("CatWriter.openField called");
	}
    }

  public void closeField(FieldTrans ft) 
    throws IOException, TransGenException
    {
      if (trace)
	{
	  System.out.println("CatWriter.closeField called");
	}

      if (ft.isIndexField())
	{
	  CategoryTrans ct 
	    = ft.getCifParentIndex().getStructure().getCategory();
	  if ( ! isCompositeKeyCat(ct))
	    {
	      indexFields.add(ft);
	    }
	}

      StructureTrans fst = ft.getFieldStructure();
      if (fst != null)
	  {
	   String sname = fst.getName();
	   if (sname.equals("AtomIndex"))
	     {
	       atomIndices.add(ft);
	     }
	   if (sname.equals("SeqIndex"))
	     {
	       seqIndices.add(ft);
	     }
	  }

      if (ft.isCifField())
	{
	  writeTypedefFields(ft);
	  writeDefineFields(ft);
	  writeInsertFields(ft);
	}
    }

  public void openSubfield(SubfieldTrans ft) 
    throws IOException, TransGenException
    {
      openField(ft);
    }

  public void closeSubfield(SubfieldTrans ft) 
    throws IOException, TransGenException
    {
      closeField(ft);
    }

  public void openVector(VectorTrans vt) 
    throws IOException, TransGenException
    {
      writeDefineFields(vt);
    }

  public void openMatrix(MatrixTrans vt) 
    throws IOException, TransGenException
    {
      writeDefineFields(vt);
    }

  // Called when the Category is closed, this method does
  // most of the work writing out representations of
  // data structures created during the Trans structure traversal.
  public void closeCategory(CategoryTrans ct) 
    throws IOException, TransGenException
    {
      writeVariables();
      writeBeginCatEndCompound();
      writeBeginEndRow();
      writeEndLoop();
      writeInsertMethods();
      closeSwitchStatements();
      writeFin();

      // copy everything from the "segmented" buffers to the Cat file
      copyCharArraysToFile();

      // close the file and reset variables
      close();
      indentLevel = 0;
      catStructs.clear();
      indexFields.clear();
      seqIndices.clear();
      atomIndices.clear();
    }

  private void initCharBuffers(CategoryTrans ct)
    throws IOException, TransGenException
    {
      if (trace)
	{
	  System.out.println("CatWriter.initCharBuffers called");
	}

      oClassName = ct.getCatClassPrefix() + catBaseName;
      oClassFilename = oClassName + ".java";
      open(loaderPath + "/" + oClassFilename);

      // Beginning of file
      initBuffer = new StringStruct();
      initBuffer.setIndentString(indent);
      // Typedef Block
      tagBuffer = new StringStruct();
      tagBuffer.setIndentString(indent);
      // beginCategory(), endCompound(), begin/endRow() & endLoop()
      beginEndBuffer = new StringStruct();
      beginEndBuffer.setIndentString(indent);
      // Set Flags Switch Block
      defineBuffer = new StringStruct();
      defineBuffer.setIndentString(indent);
      // Insert Switch Block
      insert1Buffer = new StringStruct();
      insert1Buffer.setIndentString(indent);
      insert2Buffer = new StringStruct();
      insert2Buffer.setIndentString(indent);
      // File ending
      finBuffer = new StringStruct();
      finBuffer.setIndentString(indent);
    }

  private void copyCharArraysToFile()
    throws IOException, TransGenException
    {
      // init
      writeBuffer(initBuffer);
      initBuffer = null;
      // tag
      writeBuffer(tagBuffer);
      tagBuffer = null;
      // beginEnd
      writeBuffer(beginEndBuffer);
      beginEndBuffer = null;
      // define
      writeBuffer(defineBuffer);
      defineBuffer = null;
      // insert 1
      writeBuffer(insert1Buffer);
      insert1Buffer = null;
      // insert 2
      writeBuffer(insert2Buffer);
      insert2Buffer = null;
      // fin
      writeBuffer(finBuffer);
      finBuffer = null;
    }

  private void writeCatFileOpening(CategoryTrans ct)
    throws IOException, TransGenException
    {
      Date dt = new Date();
      GregorianCalendar cal = new GregorianCalendar();
      cal.setTime(dt);

      // write preable

      String[] preamble =
      {
	"//",
	"// **** DERIVED FILE - DO NOT EDIT ****",
	"//",
      };
      writeStr(initBuffer, preamble, 0);

      writeCopyright(initBuffer);

      String[] packimport =
      {
	"",
	"package " + msaPackageName + "." + loaderPackage + ";",
	"",
	"import java.util.*;",
	"import " + msaPackageName + ".cifparse.*;",
	"import " + msaPackageName + ".util.*;",
	"import " + lsrPackageName + "." +  coreModuleName + ".*;"
      };
      writeStr(initBuffer, packimport, 0);

      if ( ! inMod.isCoreModule())
	{
	  writeStr(initBuffer,
		   "import " + lsrPackageName + "." +  baseModuleName + ".*;",
		   0);
	}
      if (moduleNameStack.size() > 1 )
	{
	  writeStr(initBuffer,
		   "import " + lsrPackageName + "." +  moduleName + ".*;", 0);
	}
      String[] opening =
      {
	"",
	"/**",
	" * Category Loader to load data into the " 
	+ modEntryName + "MethodImpl." + ct.getCatClassPrefix(),
	" * structure.",
	" * <p>",
	" * Generated by the program MMSExpr. Model: "
	  + inRoot.getModelIdentifier() + ", Revision: "
	  + inRoot.getRevisionNumber(),
	" * <p>",
	" * Date: " + dt.toString(),
	" *",
	" * @author Douglas S. Greer",
	" * @version OpenMMS Release V_1_5_1",
	" */",
	"public class " + oClassName,
	indent + "extends CatUtil implements " + catBaseName,
	"{",
	indent + "public " + oClassName + "()",
	indent + indent + "{",
	indent + indent + "}",
	""
      };
      writeStr(initBuffer, opening, 0);
    }
      
  private void writeDefineMethods()
    throws IOException, TransGenException
    {
      // define Single and Loop methods
      String[] defineItemBody =
      {
	"",
	"public void defineSingleItem(Object e, int fieldCode)",
	indent + "{",
	indent + indent + "EntryMethodImpl xe = (EntryMethodImpl) e;",
	indent + indent + "defineItem(xe, fieldCode);",
	indent + "}",
	"",
	"public void defineLoopItem(Object e, int fieldCode)",
	indent + "{",
	indent + indent + "EntryMethodImpl xe = (EntryMethodImpl) e;",
	indent + indent + "defineItem(xe, fieldCode);",
	indent + "}"
      };
      writeStr(defineBuffer, defineItemBody, 1);
    }
      
  private void writeSwitchOpenings()
    throws IOException, TransGenException
    {
      String[] insertSwitch =
      {
	"public void insertValue(int fieldCode, String value)",
	indent + "{",
	indent + "switch (fieldCode)",
	indent + indent+"{"
      };
      writeStr(insert2Buffer, insertSwitch, 1);

      String[] flagDefineItem =
      {
	"",
	"public void defineItem(EntryMethodImpl xe, int fieldCode)",
	indent + "{",
	indent + "switch (fieldCode)",
	indent + indent + "{"
      };
      writeStr(defineBuffer, flagDefineItem, 1);
    }

  private void writeTypedefFields(FieldTrans ft) 
    throws IOException, TransGenException
    {
      DictionaryItem di = ft.getDictionaryItem();
      // tag code
      writeStr(tagBuffer, 
	       "static final int " + ft.getTagName() + 
	       " = " + currentFieldCode + ";",
	       1);
      currentFieldCode++;
    }

  private void writeDefineFields(FieldTrans ft) 
    throws IOException, TransGenException
    {
      if (ft.isPseudoField())
	{
	  // Matrix or Vector fields defined with the [1] or [1][1]
	  // element in openVector() and openMatrix() calls.
	  return;
	}
      int smand = ft.getStructure().getDictionaryCategory()
	.getCategoryMandatoryCode();
      int fmand = ft.getDictionaryItem().getItemMandatoryCode();

      if (smand == DictionaryCategory.CATEGORY_MANDATORY_CODE_NO ||
	  fmand == DictionaryItem.ITEM_MANDATORY_CODE_NO)
	{
	  String fieldFlagName = ft.getFlagName();
	  String structFlagName = ft.getStructure().getFlagName();
	  
	  writeStr(defineBuffer, "case " + ft.getTagName() + ":", 3);
	  if (smand == DictionaryCategory.CATEGORY_MANDATORY_CODE_NO)
	    {
	      writeStr(defineBuffer,
		       "xe." + modVarName + "_presence_flags["
		       + modEntryName + "."
		       + structFlagName + "/8] |=", 4);
	      writeStr(defineBuffer,
		       "1 << (" + modEntryName + "."
		       + structFlagName + "%8);", 5);
	    }
	  if (fmand == DictionaryItem.ITEM_MANDATORY_CODE_NO)
	    {
	      writeStr(defineBuffer, 
		       "xe." + modVarName + "_presence_flags["
		       + modEntryName + "." + fieldFlagName + "/8] |=", 4);
	      writeStr(defineBuffer,
		       "1 << (" + modEntryName + "."
		       + fieldFlagName + "%8);", 5);
	    }
	  writeStr(defineBuffer, "break;", 4);
	  writeStr(defineBuffer, "", 0);
	}
    }

  private void writeInsertFields(FieldTrans ft) 
    throws IOException, TransGenException
    {
      String copyName;
      String convertedValue;

      DictionaryItem di = ft.getDictionaryItem();
      // inserting code
      String conv = cifTypeConverter(di.getItemType());
      convertedValue = conv + "(value)";

      writeStr(insert2Buffer, "case " + ft.getTagName() + ":", 3);

      writeStr(insert2Buffer, "var" + ft.getStructure().getName()
	       + "." + ft.getFullFieldName() + 
	       " = " + convertedValue + ";", 4);
      // copies of values needed for symmetric matrices
      copyName = ft.getCopyName();
      if (copyName != null)
	{
	  writeStr(insert2Buffer, 
		   "var" + ft.getStructure().getName() + "." + copyName
		   + " = var" + ft.getStructure().getName() 
		   + "." + ft.getFullFieldName() +  ";", 4);
	}

      if (ft.isIndexField()  && ! isCompositeKeyCat
	  (ft.getCifParentIndex().getStructure().getCategory()))
	{
	  if (conv != "cifString")
	    {
	      throw new TransGenException
		("CatWriter.writeInsertFields(): Indexing non-string field");
	    }
	  writeStr(insert2Buffer,
		   "str_indx_" + ft.getFullRefName()
		   + ".add(var" + ft.getStructure().getName() 
		   + "." + ft.getFullFieldName() + ");", 4);
	}
      writeStr(insert2Buffer, "break;", 4);
    }
  
  private void writeVariables()
    throws IOException, TransGenException
    {

      // ArrayLists
      for(int i = 0; i < catStructs.size(); i++)
	{
	  StructureTrans st = (StructureTrans) catStructs.get(i);
	  writeStr(initBuffer,
		   st.getTypeName() + " var" + st.getName() + ";", 1);
	  writeStr(initBuffer,
		   "ArrayList " + "array" + st.getName()
		   + " = new ArrayList();", 1);
	  writeStr(initBuffer, "", 0);
	}
      if (CompositeKeyCat != null)
	{
	  hashTableName = "hash_key_" + CompositeKeyCat.getName();
	  writeStr(initBuffer, "public static Hashtable " + hashTableName
		   + " = new Hashtable();", 1);
	  writeStr(initBuffer, "", 0);
	}

      if (indexFields.size() > 0)
	{
	  writeStr(beginEndBuffer, "", 0);
	}
      for(int i = 0; i < indexFields.size(); i++)
	{
	  FieldTrans ft = (FieldTrans) indexFields.get(i);
	  writeStr(beginEndBuffer, "ArrayList str_indx_" + ft.getFullRefName()
		   + " = new ArrayList();", 1);
	  writeStr(beginEndBuffer, "Index_" + ft.getFullRefName()
		   + " ndx_" + ft.getFullRefName()
		   + " = new Index_" + ft.getFullRefName() + "();", 1);

	}
      if (seqIndices.size() > 0 || atomIndices.size() > 0)
	{
	  writeStr(beginEndBuffer, "", 0);
	}
      for(int i = 0; i < seqIndices.size(); i++)
	{
	  FieldTrans ft = (FieldTrans) seqIndices.get(i);
	  writeStr(beginEndBuffer, "ArrayList seq_indx_" + ft.getFullRefName()
		   + " = new ArrayList();", 1);
	}
      for(int i = 0; i < atomIndices.size(); i++)
	{
	  FieldTrans ft = (FieldTrans) atomIndices.get(i);
	  writeStr(beginEndBuffer, "ArrayList atom_indx_" + ft.getFullRefName()
		   + " = new ArrayList();", 1);
	}
    }

  private void writeBeginCatEndCompound()
    throws IOException, TransGenException
    {
      // define member classes
      for(int i = 0; i < indexFields.size(); i++)
	{
	  FieldTrans ft = (FieldTrans) indexFields.get(i);
	  FieldTrans parent = ft.getCifParentIndex();
	  String parentConverter 
	    = cifTypeConverter(parent.getDictionaryItem().getItemType());
	  String pval =  "((EntryMethodImpl)xe)."
	    + parent.getStructure().getFullListVarName()
	      + "[i]." + parent.getFullRefName();
	  String comparer, javaType;

	  if (parentConverter.equalsIgnoreCase("cifString"))
	    {
	      comparer = "findVar.equalsIgnoreCase(" + pval + ")";
	      javaType = "String";
	    }
	  else if (parentConverter.equalsIgnoreCase("cifInt"))
	    {
	      comparer = "findVar == " + pval;
	      javaType = "int";
	    }
	  else if (parentConverter.equalsIgnoreCase("cifFloat"))
	    {
	      comparer = "findVar == " + pval;
	      javaType = "float";
	    }
	  else
	    {
	      throw new TransGenException
		("CatWriter: - Unknown parent type: " + parentConverter);
	    }

	  writeStr(beginEndBuffer, "", 1);
	  writeStr(beginEndBuffer,
		   "public class Index_" + ft.getFullRefName()
		   + " implements StringToIndex", 1);
	  writeStr(beginEndBuffer, "{", 2);
	  writeStr(beginEndBuffer, javaType + " findVar;", 3);
	  writeStr(beginEndBuffer,
		   "public void setFindValue(String s)", 3);
	  writeStr(beginEndBuffer,
		   "{findVar = " + parentConverter + "(s);};", 4);
	  writeStr(beginEndBuffer,
		   "public int getParentSize(Object xe)", 3);
	  writeStr(beginEndBuffer,
		   "{if (((EntryMethodImpl)xe)."
		   + parent.getStructure().getFullListVarName()
		   + " == null) return -1;", 4);
	  writeStr(beginEndBuffer,
		   "else return ((EntryMethodImpl)xe)."
		   + parent.getStructure().getFullListVarName()
		   + ".length;};", 5);
	  writeStr(beginEndBuffer,
		   "public boolean compareToParent"
		   + "(Object xe, int i)",
		   3);
	  writeStr(beginEndBuffer,
		   "{return " + comparer + ";};", 4);
	  writeStr(beginEndBuffer, 
		   "public void insertIndex("
		   + "Object xe, int i, int val)", 3);
	  writeStr(beginEndBuffer, 
		   "{((EntryMethodImpl)xe)." + modVarName
		   + ft.getStructure().getListVarName() + "[i]."
		   + ft.getContainingField().getFullFieldName()
		   + ".index = val;};", 4);
	  writeStr(beginEndBuffer, "}", 2);
	}

      String[] beginCatOpen =
      {
	"",
	"public void beginCategory()",
	indent + "{"
      };
      writeStr(beginEndBuffer, beginCatOpen, 1);

      for(int i = 0; i < catStructs.size(); i++)
	{
	  StructureTrans st = (StructureTrans) catStructs.get(i);
	  writeStr(beginEndBuffer,
	  "var" + st.getName() + " = null;", 3);
	}
      for(int i = 0; i < indexFields.size(); i++)
	{
	  FieldTrans ft = (FieldTrans) indexFields.get(i);
	  writeStr(beginEndBuffer,
		   "str_indx_" + ft.getFullRefName() + ".clear();", 3);
	}
      for(int i = 0; i < seqIndices.size(); i++)
	{
	  FieldTrans ft = (FieldTrans) seqIndices.get(i);
	  writeStr(beginEndBuffer,
		   "seq_indx_" + ft.getFullRefName() + ".clear();", 3);
	}
      for(int i = 0; i < atomIndices.size(); i++)
	{
	  FieldTrans ft = (FieldTrans) atomIndices.get(i);
	  writeStr(beginEndBuffer,
		   "atom_indx_" + ft.getFullRefName() + ".clear();", 3);
	}
      if (CompositeKeyCat != null)
	{
	  writeStr(beginEndBuffer, hashTableName + ".clear();", 3);
	}

      String[] endCompoundOpen =
      {
	indent + "}",
	"",
	"public void endCompound(Object e)",
	indent + "throws CifParseException",
	indent + "{",
      };
      writeStr(beginEndBuffer, endCompoundOpen, 1);
      if (indexFields.size() > 0 || seqIndices.size() > 0
	  || atomIndices.size() > 0)
	{
	  writeStr(beginEndBuffer, 
		   "EntryMethodImpl xe = (EntryMethodImpl) e;", 3);
	}
      for(int i = 0; i < indexFields.size(); i++)
	{
	  FieldTrans ft = (FieldTrans) indexFields.get(i);
	  writeStr(beginEndBuffer, "setChildIndex(xe,"
		   + " str_indx_" + ft.getFullRefName() + ", "
		   + " ndx_" + ft.getFullRefName() + ");", 3);
	}
      for(int i = 0; i < seqIndices.size(); i++)
	{
	  FieldTrans ft = (FieldTrans) seqIndices.get(i);
	  writeStr(beginEndBuffer, "xe.setCompositeSeqIndexList(seq_indx_"
		   + ft.getFullRefName() + ");", 3);
	}
      for(int i = 0; i < atomIndices.size(); i++)
	{
	  FieldTrans ft = (FieldTrans) atomIndices.get(i);
	  writeStr(beginEndBuffer, "xe.setCompositeAtomIndexList(atom_indx_"
		   + ft.getFullRefName() + ");", 3);
	}
      String[] endComp =
      {
	indent + "}"
      };
      writeStr(beginEndBuffer, endComp, 1);
    }

  private void writeBeginEndRow()
    throws IOException, TransGenException
    {
      String[] beginLoopOpen =
      {
	"",
	"public void beginRow()",
	indent + "{"
      };
      writeStr(beginEndBuffer, beginLoopOpen, 1);

      for(int i = 0; i < catStructs.size(); i++)
	{
	  StructureTrans st = (StructureTrans) catStructs.get(i);
	  writeStr(beginEndBuffer, "var" + st.getName() + " = new "
		   + st.getTypeName() + "();", 3);
	  initStructChildren(st);
	}

      String[] endRowOpen =
      {
	indent + "}",
	"",
	"public void endRow()",
	indent + "{"
      };
      writeStr(beginEndBuffer, endRowOpen, 1);
      for(int i = 0; i < catStructs.size(); i++)
	{
	  StructureTrans st = (StructureTrans) catStructs.get(i);
	  writeStr(beginEndBuffer, "array" + st.getName() + ".add(" +
		   "var" + st.getName() + ");", 3);
	}
      for(int i = 0; i < seqIndices.size(); i++)
	{
	  FieldTrans ft = (FieldTrans) seqIndices.get(i);
	  StructureTrans cst = ft.getStructure();
	  writeStr(beginEndBuffer, "seq_indx_" + ft.getFullRefName() 
		   + ".add(var"
		   + cst.getName() + "." + ft.getFullFieldName() + ");", 3);
	}
      for(int i = 0; i < atomIndices.size(); i++)
	{
	  FieldTrans ft = (FieldTrans) atomIndices.get(i);
	  StructureTrans cst = ft.getStructure();
	  writeStr(beginEndBuffer, "atom_indx_" + ft.getFullRefName()
		   + ".add(var"
		   + cst.getName() + "." + ft.getFullFieldName() + ");", 3);
	}
      writeStr(beginEndBuffer, indent + "}", 1);
    }

  private void initStructChildren(Trans pt)
    throws IOException, TransGenException
    {
      String nameSoFar;

      for (int j = 0; j < pt.numberOfChildren(); j++)
	{
	  Trans t = pt.getChild(j);
	  if (t.getTransType() == Trans.TYPE_FIELD
	      || t.getTransType() == Trans.TYPE_SUBFIELD)
	    {
	      FieldTrans ft = (FieldTrans) t;
	      if (! ft.isLeaf())
		{
		  // containing structure
		  StructureTrans cst = ft.getStructure();
		  // structure declaration for this field
		  StructureTrans fstt = ft.getFieldStructure();
		  // create the sub-structure
		  nameSoFar = "var" + cst.getName()
			   + "." + ft.getFullFieldName();
		  writeStr(beginEndBuffer,
			   nameSoFar + " = new "
			   + fstt.getTypeName() + "();",
			   3);
		  // recurse structure subfields
		  StructureTrans fst = ft.getFieldStructure();
		  if (fst != null)
		    {
		      initSubstructure(fst, nameSoFar);
		    }
		  // recurse structure subfields
		}
	      else
		{
		  // Leaf node
		  if (ft.isCifField())
		    {
		      DictionaryItem di = ft.getDictionaryItem();
		      if (di == null)
			{
			  throw new TransGenException
			    ("CatWriter.initStructChildern()"
			     + "- No dictionary entry"
			     + "\n\tFullFieldName()=" + ft.getFullFieldName()
			     + "\n\tgetCifItemName()=" + ft.getCifItemName()
			     + "\n\tgetContainingField()="
			        + ft.getContainingField());
			}
		      // In *THEORY* only need to initialize optional
		      // strings, however in practice some mandatory
		      // fields are missing in the data files
		      // if ((di.getItemMandatoryCode()
		      //   == DictionaryItem.ITEM_MANDATORY_CODE_NO) ...
		      if (cifTypeConverter(di.getItemType())
			      == "cifString")
			{
			  // need to initialize all non-mandatory strings
			  // with a zero lenght string.  Otherwise Corba will
			  // throw a BAD_PARAM exception on the structure
			  writeStr(beginEndBuffer,
				   "var" + ft.getStructure().getName()
				   + "." + ft.getFullFieldName()
				   + " = \"\";", 3);
			}
		    }
		}
	    }
	  else if (t.getTransType() == Trans.TYPE_VECTOR)
	    {
	      VectorTrans vt = (VectorTrans) t;
	      writeStr(beginEndBuffer,
		       "var" + vt.getStructure().getName()
		       + "." + vt.getFullFieldName()
		       + " = new float[3];", 3);
	    }
	  else if (t.getTransType() == Trans.TYPE_MATRIX)
	    {
	      MatrixTrans mt = (MatrixTrans) t;
	      writeStr(beginEndBuffer,
		       "var" + mt.getStructure().getName()
		       + "." + mt.getFullFieldName() 
		       + " = new float[3][3];", 3);
	    }
	}
    }

  private void initSubstructure(StructureTrans st, String nameSoFar)
    throws IOException, TransGenException
    {
      for (int j = 0; j < st.numberOfChildren(); j++)
	{
	  Trans t = st.getChild(j);

	  if (t instanceof VectorTrans)
	    {
	      VectorTrans vt = (VectorTrans) t;
	      writeStr(beginEndBuffer,
		       nameSoFar + vt.getName()
		       + " = new float[3];", 3);
	    }
	  else if (t instanceof MatrixTrans)
	    {
	      MatrixTrans mt = (MatrixTrans) t;
	      writeStr(beginEndBuffer,
		       nameSoFar + mt.getName()
		       + " = new float[3][3];", 3);
	    }
	  else if (t instanceof FieldTrans)
	    {
	      FieldTrans ft = (FieldTrans) t;
	      // recurse structure subfields
	      StructureTrans fst = ft.getFieldStructure();
	      if (fst != null)
		{
		  writeStr(beginEndBuffer, nameSoFar + "." + ft.getName()
			   + " = new " + fst.getTypeName() + "();",
			   3);
		  initSubstructure(fst, nameSoFar + "." + ft.getName());
		}
	      else
		{
		  if (ft.getFieldType(inMod).equals("string"))
		    {
		      writeStr(beginEndBuffer,
			       nameSoFar + "." + ft.getName()
			       + " = \"\";", 3);
		    }
		}
	    }
	  else
	    {
	      throw new TransGenException
		("CatWriter.initSubstructure: unknown Structure field type:"
		 + t.getTransType()
		 + "\n\tStruct name: " + st.getName()
		 + " Field name: " + t.getName());
	    }
	}
    }

  private void writeEndLoop()
    throws IOException, TransGenException
    {
      String[] endLoopOpen = {
      "",
      "public void endLoop(Object e)",
	indent + "{",
	indent + indent + "int i;",
	indent + indent + "EntryMethodImpl xe = (EntryMethodImpl) e;"
      };
      writeStr(beginEndBuffer, endLoopOpen, 1);

      for(int i = 0; i < catStructs.size(); i++)
	{
	  StructureTrans st = (StructureTrans) catStructs.get(i);
	  String s = "array" + st.getName();
	  String slName = st.getListVarName();

	  String[] endLoopBodyOpen = {
	    "xe." + modVarName + slName 
	    	+ " = new " + st.getTypeName() + "[" + s + ".size()];",
	    "for (i = 0; i < " + s + ".size(); i++)",
	    indent + "{",
	    indent + indent + "xe." + modVarName + slName + "[i]",
	    indent + indent + indent + "= ("
	    		+ st.getTypeName() + ") " + s + ".get(i);"
	  };
	  writeStr(beginEndBuffer, endLoopBodyOpen, 3);

	  if (CompositeKeyCat != null && i == 0) 
	    {
	      // only composite key in town...
	      if (CompositeKeyCat.getName()
		  .equalsIgnoreCase("entity_poly_seq"))
		{
		  String[] putInHash = {
		    hashTableName + ".put(",
		    indent + "xe." + modVarName 
		    	+ slName + "[i].entity.id + '&'",
		    indent + "+ xe." + modVarName + slName + "[i].num + '&'",
		    indent + "+ xe." + modVarName + slName + "[i].mon.id,",
		    indent + "new Integer(i));"
		  };
		  writeStr(beginEndBuffer, putInHash, 5);
		}
	    }

	  String[] endLoopBodyClose = {
	    indent + "}",
	    "array" + st.getName() + ".clear();" 
	  };
	  writeStr(beginEndBuffer, endLoopBodyClose, 3);
	}

      writeStr(beginEndBuffer, indent + "}", 1);
    }

  private void writeInsertMethods()
    throws IOException, TransGenException
    {
      // Insert Single and Loop methods
      String[] insertSingleValueOpen =
      {
	"",
	"public void insertSingleValue(Object e, "
	    + "int fieldCode, String value)",
	indent + "{",
	indent + "switch (fieldCode)",
	indent + indent + "{"
      };
      writeStr(insert1Buffer, insertSingleValueOpen, 1);

      for(int i = 0; i < catStructs.size(); i++)
	{
	  StructureTrans st = (StructureTrans) catStructs.get(i);
	  insertStructChildren(st);
	  String[] insertSingleValueBody =
	  {
	    "if(var" + st.getName() + " == null)",
	    indent + "{",
	    indent + indent + "beginRow();",
	    indent + indent + "EntryMethodImpl xe = (EntryMethodImpl) e;",
	    indent + indent + "xe." + modVarName + st.getListVarName() 
	    	+ " = new " + st.getTypeName() + "[1];",
	    indent + indent
	        + "xe." + modVarName + st.getListVarName() 
	        + "[0] = var" + st.getName() + ";",
	    indent + "}",
	    "break;"

	  };
	  writeStr(insert1Buffer, insertSingleValueBody, 4);
	}

      String[] insertLoopBody =
      {
	indent + indent + "}",
	indent + "insertValue(fieldCode, value);",
	indent + "}",
	"",
	"public void insertLoopValue(int fieldCode, String value)",
	indent + "{",
	indent + indent + "insertValue(fieldCode, value);",
	indent + "}",
	""
      };
      writeStr(insert1Buffer, insertLoopBody, 1);
    }

  private void insertStructChildren(Trans pt)
    throws IOException, TransGenException
    {
      Trans t, t2;
      FieldTrans ft;
      int ty;

      for (int j = 0; j < pt.numberOfChildren(); j++)
	{
	  t = pt.getChild(j);
	  ty = t.getTransType();
	  if (ty == Trans.TYPE_FIELD
	      || ty == Trans.TYPE_SUBFIELD)
	    {
	      ft = (FieldTrans) t;
	      if (! ft.isLeaf())
		{
		  // recurse down structure
		  insertStructChildren(t);
		}
	      else
		{
		  if (ft.isCifField())
		    {
		      writeStr(insert1Buffer,"case "
			       + ft.getTagName() + ":", 3);
		    }
		}
	    }
	  else if (ty == Trans.TYPE_VECTOR ||
		   ty == Trans.TYPE_MATRIX)
	    {
	      for (int k = 0; k < t.numberOfChildren(); k++)
		{
		  t2 = t.getChild(k);
		  if (t2.getTransType() == Trans.TYPE_FIELD)
		    {
		      ft = (FieldTrans) t2;
		      if (ft.isCifField())
			{
			  writeStr(insert1Buffer,"case "
				   + ft.getTagName() + ":", 3);
			}
		    }
		}
	    }
	}
    }

  private void closeSwitchStatements()
    throws IOException, TransGenException
    {
      // Insert Switch Block
      String[] closeSwitch =
      {
	indent+"}",
	"}",
      };
      writeStr(insert2Buffer, closeSwitch, 2);
      writeStr(defineBuffer, closeSwitch, 2);
    }

  private void writeFin()
    throws IOException, TransGenException
    {
      String[] closing =
      {
	"}",
	""
      };
      writeStr(finBuffer, closing, 0);
    }

  static public String cifTypeConverter(String ct) throws TransGenException
    {
      // no conversion necessary for strings
      if (ct == null)
	{
	  return "cifString";
	}
      if (ct.compareToIgnoreCase("float") == 0)
	{
	  // return "Float.parseFloat";
	  return "cifFloat";
	}
      else if (ct.compareToIgnoreCase("int") == 0)
	{
	  // return "Integer.parseInt";
	  return "cifInt";
	}
      else if (ct.compareToIgnoreCase("text") == 0 ||
	       ct.compareToIgnoreCase("name") == 0 ||
	       ct.compareToIgnoreCase("idname") == 0 ||
	       ct.compareToIgnoreCase("ucode") == 0 ||
	       ct.compareToIgnoreCase("code") == 0 ||
	       ct.compareToIgnoreCase("atcode") == 0 ||
	       ct.compareToIgnoreCase("yyyy-mm-dd") == 0 ||
	       ct.compareToIgnoreCase("line") == 0 ||
	       ct.compareToIgnoreCase("uline") == 0 ||
	       ct.compareToIgnoreCase("any") == 0 ||
	       ct.compareToIgnoreCase("uchar3") == 0 ||
	       ct.compareToIgnoreCase("uchar1") == 0 ||
	       ct.compareToIgnoreCase("symop") == 0 ||
	       ct.compareToIgnoreCase("float-range") == 0 ||
	       ct.compareToIgnoreCase("yyyy-mm-dd:hh:mm") == 0 ||
	       ct.compareToIgnoreCase("fax") == 0 ||
	       ct.compareToIgnoreCase("phone") == 0 ||
	       ct.compareToIgnoreCase("email") == 0)
	{
	  return "cifString";
	}
      else
	{
	  throw new TransGenException
	    ("cifTypeConverter - Unknown type: " + ct);
	}
    }
}
