//
// FieldTrans.java,v 1.6 2003/05/30 18:57:01 dsg Exp
//
// Copyright 2001 The Regents of the University of California
// All Rights Reserved
//
// OpenMMS was developed by Dr. Douglas S. Greer at the San Diego
// Supercomputer Center, a research unit of the University of California,
// San Diego.  Support for this effort was provided by NSF through the
// Protein Data Bank (Grant DBI-9814284) and the National Partnership for
// Advanced Computational Infrastructure (Grant ACI-9619020)
//
// Permission to use, copy, modify and distribute any part of OpenMMS for
// educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above
// copyright notice, this paragraph and the following paragraphs appear
// in all copies.
//
// Those desiring to incorporate this OpenMMS into commercial products or
// use for commercial purposes should contact the Technology Transfer
// Office, University of California, San Diego, 9500 Gilman Drive, La
// Jolla, CA 92093-0910, Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// In no event shall the University of California be liable to any party
// for direct, indirect, special, incidental, or consequential damages,
// including lost profits, arising out of the use of this OpenMMS, even
// if the University of California has been advised of the possibility of
// such damage.
//
// The OpenMMS provided herein is on an "as is" basis, and the
// University of California has no obligation to provide maintenance,
// support, updates, enhancements, or modifications.  The University of
// California makes no representations and extends no warranties of any
// kind, either implied or express, including, but not limited to, the
// implied warranties of merchantability or fitness for a particular
// purpose, or that the use of the OpenMMS will not infringe any patent,
// trademark or other rights.

package org.rcsb.openmms.meta;

import java.io.IOException;

import org.rcsb.openmms.cifparse.*;

/**
 * Class for "Fields" in Cif Metamodel
 *
 * @author Douglas S. Greer
 * @version 1.6
 */
public class FieldTrans extends Trans
{
  static final int MAX_PARAGRAPHS = 100;

  private String cifItemName;
  private String tagName;
  /**
   * "real" name used to generate vector/matrix code that correctly
   * indexes values with "n" instead of "n+1"
   */
  protected String realName;

  /**
   * prefixed with list of embedded field structure names using "."
   */
  protected String fullFieldName;
  /**
   * prefixed with list of embedded field structure names using "_"
   */
  protected String fullRefName;

  /**
   * Name of "extra" copy used in constructing symetric matrices
   */
  protected String copyName = null;

  /**
   * names for building sql tables
   */
  protected String columnName = null;
  protected String columnIndexName = null;
 
  private StructureTrans containingStruct;
  /**
   * Dictionary Item associated with this field
   */
  private DictionaryItem dictItem;

  protected boolean isCifField = true;
  /**
   * Non-Cif fields must explictly declare their type
   */
  protected String fieldType;

  /**
   * Pointer to the structure node when a structure type
   * used as a field element.
   */
  protected StructureTrans structType;

  /**
   * Indicates fields used to roll-up vectors and matrices in
   * mmCIF into the correct Vectore and Matrix3 structures
   */
  protected boolean pseudoField;

  private boolean indexField;
  private FieldTrans parentIndexField;
  
  private int flags = 0;
  public static final int IS_A_STREAM_DATATYPE = 0x1;

  public int descriptionParagraphs;

  public FieldTrans(String n)
    {
      this(n, n, MAX_PARAGRAPHS);
    }

  public FieldTrans(String varname, String cifname)
    {
      this(varname, cifname, MAX_PARAGRAPHS);
    }

  public FieldTrans(String rtype, String varname, boolean icf)
    {
      // Constructor used for non-Cif structures/fields (w/ icf == false)
      this(varname, null, 0);
      fieldType = rtype;
      isCifField = icf;
    }

  /**
   * Constructor used for embedded structs
   */
  public FieldTrans(StructureTrans st, String varname)
    {
      this(varname, null, 0);
      structType = st;
      fieldType = st.getName();
      isCifField = false;
    }

  public FieldTrans(String varname, String cifname, int descp)
    {
      super(varname, Trans.TYPE_FIELD);
      if (cifname != null)
	{
	  cifItemName = cifname.trim();
	}
      else
	{
	  cifItemName = null;
	}
      // real name is the same as regular name unless set with setRealName()
      realName = varname;
      descriptionParagraphs = descp;
    }

  public void accept(TransVisitor v)
    throws IOException, TransGenException
    {
      v.openField(this);
      // visit children
      Trans t;
      for (int i = 0; i < numberOfChildren(); i++)
	{
	  t = getChild(i);
	  int type = t.getTransType();
	  switch (type)
	    {
	    case Trans.TYPE_SUBFIELD:
	    case Trans.TYPE_VECTOR:
	    case Trans.TYPE_MATRIX:
	      t.accept(v);
	      break;
	    default:
	      throw new TransGenException
		("Invalid FieldTrans Child Type: " + type 
		 + ", Parent name: " + getName()
		 + ", Child name: " + t.getName());
	    }
	}
      v.closeField(this);
    }

  public void setStructure(StructureTrans st)
    {
      containingStruct = st;
    }

  public StructureTrans getStructure()
    {
      return containingStruct;
    }

  public void setDictionaryItem(DictionaryItem di)
    {
      dictItem = di;
    }

  public DictionaryItem getDictionaryItem()
    {
      return dictItem;
    }

  /**
   * This method must return a valid mmCIF dictionary reference
   */
public String getCifItemName()
    {
      return cifItemName;
    }

  /**
   * Pseudo Fields represent the individual elements in an array
   * (e.g. the unrolled mmCIF vectors/matrices)
   */
  public void setPseudoField(boolean pf)
    {
      pseudoField = pf;
    }

  /**
   * Pseudo Fields represent the individual elements in an array
   * (e.g. the unrolled mmCIF vectors/matrices)
   */
  public boolean isPseudoField()
    {
      return pseudoField;
    }

  public boolean isCifField()
    {
      return isCifField;
    }

  public void computeNames()
    {
      // Note that in the following two statements, the method 
      // computeFullName(...) is overridden by subclasses.
      fullFieldName = computeFullName('.');
      fullRefName = computeFullName('_');
      // tagName = cifNameToTagName(cifname); - cif Name may contain
      // illegal characters
      tagName = cifNameToTagName(fullRefName);
    }

  /**
   * Overridded by SubfieldTrans and MatrixTrans
   * For Fields and Subfields this will
   * return the name of a basic type only if
   * getFieldStructure() returns null.
   * Otherwise it has type getFieldStructure().getName()
   * Note that private methods are not overriden, so this
   * method must declared with at least protected access.
   */
  protected String computeFullName(char c)
    {
      return getRealName();
    }

  public String getFullFieldName()
    {
      // default
      return fullFieldName;
    }

  public String getFullRefName()
    {
      // default
      return fullRefName;
    }

  /**
   * A leaf field node is anything that does not have a structure type,
   * and therefore never has children.
   */
  public boolean isLeaf()
    {
      return structType == null;
    }

  public void setFieldStructure(StructureTrans st)
    {
      structType = st;
    }

  /**
   * for fields that are of type struct, returns a pointer to the struct
   * declaration; otherwise returns null
   */
  public StructureTrans getFieldStructure()
    {
      return structType;
    }

  /**
   * returns a string representing the type of this field type
   */
  public String getFieldType()
    {
      return fieldType;
    }

  /**
   * returns a string representing the type of this field
   * If the type is a structure name, then the fully qualified
   * name will be returned if the structure is not in the 
   * current module as specified by inMod.
   */
  public String getFieldType(ModuleTrans inMod)
    {
      String rtype;

      if (structType == null)
	{
	  rtype = fieldType;
	}
      else
	{
	  // rtype = RootTrans.coreModuleName + "::" + structType.getName();
	  rtype = structType.getFullyQualifiedName(inMod);
	}
      return rtype;
    }

  public String getFlagName()
    {
      return "F" 
	+ containingStruct.convertToFlagName(containingStruct.getName())
	  + "_" + getFullRefName().toUpperCase();
    }

  public String getTagName()
    {
      return tagName;
    }

  public void setRealName(String s)
    {
      realName = s;
    }

  public String getRealName()
    {
      return realName;
    }

  // Name of "extra" copy used in constructing symetric matrices
  public void setCopyName(String s)
    {
      copyName = s;
    }

  public String getCopyName()
    {
      return copyName;
    }

  public void setColumnName(String s)
    {
      columnName = s;
    }

  public String getColumnName()
    {
      return columnName;
    }

  public void setColumnIndexName(String s)
    {
      columnIndexName = s;
    }

  public String getColumnIndexName()
    {
      return columnIndexName;
    }

  public FieldTrans getContainingField()
    {
      return null;
    }

  public void setCifParentIndex(FieldTrans p)
    {
      indexField = true;
      parentIndexField = p;
    }
  
  public boolean isIndexField()
    {
      return indexField;
    }

  public FieldTrans getCifParentIndex()
    {
      return parentIndexField;
    }

  public void setFieldBit(int f)
    {
      flags = flags | f; 
    }

  public boolean isFieldBitSet(int f)
    {
      return (flags & f) != 0;
    }

  /**
   * Converts the left brackets and periods to underbars and deletes
   * the right brackets
   */
  public String cifNameToTagName(String cname)
    {
      if (cname == null)
	{
	  return null;
	}

      StringBuffer sb = new StringBuffer();
      char c;

      for (int i = 0; i < cname.length(); i++)
	{
	  c = cname.charAt(i);
	  if (c == '[' || c == '.')
	    {
	      sb.append('_');
	    }
	  else if (c != ']')
	    {
	      sb.append(c);
	    }
	}
      String s = new String(sb);
      return s.toUpperCase();
    }
}
