//
// MMSExpr.java,v 1.5 2003/05/12 22:09:00 dsg Exp
//
// Copyright 2001 The Regents of the University of California
// All Rights Reserved
//
// OpenMMS was developed by Dr. Douglas S. Greer at the San Diego
// Supercomputer Center, a research unit of the University of California,
// San Diego.  Support for this effort was provided by NSF through the
// Protein Data Bank (Grant DBI-9814284) and the National Partnership for
// Advanced Computational Infrastructure (Grant ACI-9619020)
//
// Permission to use, copy, modify and distribute any part of OpenMMS for
// educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above
// copyright notice, this paragraph and the following paragraphs appear
// in all copies.
//
// Those desiring to incorporate this OpenMMS into commercial products or
// use for commercial purposes should contact the Technology Transfer
// Office, University of California, San Diego, 9500 Gilman Drive, La
// Jolla, CA 92093-0910, Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// In no event shall the University of California be liable to any party
// for direct, indirect, special, incidental, or consequential damages,
// including lost profits, arising out of the use of this OpenMMS, even
// if the University of California has been advised of the possibility of
// such damage.
//
// The OpenMMS provided herein is on an "as is" basis, and the
// University of California has no obligation to provide maintenance,
// support, updates, enhancements, or modifications.  The University of
// California makes no representations and extends no warranties of any
// kind, either implied or express, including, but not limited to, the
// implied warranties of merchantability or fitness for a particular
// purpose, or that the use of the OpenMMS will not infringe any patent,
// trademark or other rights.

package org.rcsb.openmms.meta;

import java.awt.*;
import java.io.*;
import java.net.*;
import java.util.*;

import org.rcsb.openmms.cifparse.*;

/**
 * Generate Macromolecular Structure Expressions
 * <p>
 * The following command line options are recognized my MMSExpr:
 * <blockquote><dl> 
 *   <dt>debug=<i>debug_level</i> 
 *     <dd>Set error and status reporting message level<br>
 *     &nbsp;&nbsp;&nbsp;&nbsp;debug <= 0 no messages<br>
 *     &nbsp;&nbsp;&nbsp;&nbsp;debug == 1 basic status messages<br>
 *     &nbsp;&nbsp;&nbsp;&nbsp;debug >= 5 trace
 *   <dt>Corba
 *     <dd> If present on the command line, Corba
 *           IDL definitions are generated.
 *   <dt>XML
 *     <dd> If present on the command line, XML DTDs and files with name
 *          bindings for xml data file conversions are generated.
 *   <dt>Rdb
 *     <dd> If present on the command line, Relational
 *          Database SQL code and JDBC loaders are generated.
 *   <dt>NoMandatoryItems
 *     <dd> If present on the command line, mmCIF mandatory items are not
 *          specified as "NOT NULL" in RDB Table table definitions
 *   <dt>dct=<i>cif_mm_dictionary_name</i>
 *     <dd> Specifies the mmCIF dictionary to use in generating metamodel
 *   <dt>mmspackage=<i>OPENMMS_package_name</i>
 *     <dd> OpenMMS package prefex to use in generated code
 *   <dt>idlpath=<i>idl_dir</i>
 *     <dd> Directory where Corba IDL should be placed
 *   <dt>lsrname=<i>LSR_base_name</i>
 *     <dd> LSR package prefix (optional)
 *   <dt>lsrpackage=<i>OMG_package_name</i>
 *     <dd> OMG package name to use in generated code
 *   <dt>ldrpath=<i>loader_dir</i>
 *     <dd> Directory where Category loader files should be placed
 *   <dt>vtypepath=<i>vtype_dir</i>
 *     <dd> Directory where Valuetype Factory code should be placed
 *   <dt>xmlpath=<i>xml_dtd_dir</i>
 *     <dd> Directory where XML DTDs should be placed
 *   <dt>sqlpath=<i>deriv_sql_dir</i>
 *     <dd> Directory where SQL Table creating and initialization
 *          scripts should be placed
 *   <dt>insertpath=<i>dbinsert_dir</i>
 *     <dd> Directoy where JDBC database insertion code should be placed.
 * </dl>
 * </blockquote>
 *
 * @author Douglas S. Greer
 * @version 1.5
 */
public class MMSExpr
{
  // basic status debug messages
  public static int debug = 1;
  String rootType;

  boolean genCorba = false;
  boolean genXML = false;
  boolean genRdb = false;
  boolean noMandItem = false;

  String idlPath, lsrModuleName, ldrPath;
  String sqlPath, xmlPath, insertPath;
  String vtypePath;
  String lsrPackageName;
  String mmsPackageName;
  ArrayList dct;
  CifTokenizer ct;
  CifDictionary cifDic;
  RootTrans root;

  public static void main(String[] argv)
    throws Exception 
    {
      MMSExpr cif2Idl=new MMSExpr();
      cif2Idl.execute(argv);
    }

  public void execute(String[] argv)
    {
      CifParser cp = new CifParser();

      try
	{
	  init(argv);

	  // Create a list of all DictionaryItems we need
	  GenDictionary dlg = new GenDictionary();
	  root.accept(dlg);
	  cifDic = dlg.dictionary();
	  // "Entry" category special cased
	  cifDic.getDictionaryCategoryList()
	    .add(new DictionaryCategory("entry"));
	  cifDic.getDictionaryItemList()
	    .add(new DictionaryItem("_entry.id"));
	  
	  for(int i = 0; i < dct.size(); i++)
	    {
	      String s = (String) dct.get(i);
	      System.out.println("Reading dictionary from " + s);
	      try
		{
		  ct = new CifTokenizer(s);
		}
	      catch(IOException e)
		{
		  System.out.println("Error in opening File: " + s);
		  System.exit(1);
		}
	      cp.readDictionary(ct, cifDic, false);
	    }

	  // Check to make sure we found all the DictionaryItems
	  // we needed
	  cifDic.checkAllItemsDefined();

	  // Determine mmCIF parent fields
	  GenCifParents gcp = new GenCifParents(cifDic);
	  root.accept(gcp);
	  gcp.setCifParentIndices(cifDic);

	  // Generate the internal ConstantTrans structures
	  GenConstants gc = new GenConstants();
	  root.accept(gc);

	  // Generate required internal MethodTrans structures
	  GenAccessMethods gam = new GenAccessMethods();
	  root.accept(gam);
	  
	  // generate all the idl and java files
	  if (genCorba)
	    {
	      writeCorbaFiles(root);
	    }
	  if (genXML)
	    {
	      writeXMLFiles(root);
	    }
	  if (genRdb)
	    {
	      writeRdbFiles(root);
	    }
	  System.out.println("Finished");
	}
      catch(IOException e)
	{
	  System.out.println("IOException: " + e.getMessage());
	  e.printStackTrace();
	  System.exit(1);
	}
      catch(CifParseException e)
	{
	  if (ct != null)
	    {
	      System.out.println("Fatal Error parsing Cif file: " 
		 + ct.filename() + " at line number " + ct.getLineNumber());
	    }
	  else
	    {
	      System.out.println("Fatal Error");
	    }
	  e.printStackTrace();
	  System.exit(1);
	}
      catch(TransGenException e)
	{
	  System.out.println("IDL Generation Error");
	  e.printStackTrace();
	  System.exit(1);
	}
    }

  public void init(String[] argv)
    throws CifParseException
    {
      String debugStr = parseArguments("debug", argv);
      if(debugStr != null)
	{
	  // debug <= 0 no messages
	  // debug == 1 basic status messages
	  // debug >= 5 trace
	  debug = Integer.parseInt(debugStr);
	}
      genCorba = parseArgumentsBool("Corba", argv);
      genRdb = parseArgumentsBool("Rdb", argv);
      genXML = parseArgumentsBool("XML", argv);
      noMandItem = parseArgumentsBool("NoMandatoryItems", argv);

      rootType = parseArguments("root", argv);
      lsrModuleName = parseArguments("lsrname", argv);

      idlPath = parseArguments("idlpath", argv);
      if(idlPath == null && genCorba)
	{
	  throw new CifParseException
	    ("MMSExpr: Insufficient input - must specify idl pathname" +
	     "\n\t and idl filename (idlpath=<>)");
	}
      ldrPath = parseArguments("ldrpath", argv);
      if(ldrPath == null && genCorba)
	{
	  throw new CifParseException
	    ("MMSExpr: Insufficient input - must specify pathname" +
	     "\n\t for generated loader files (ldrpath=<>)");
	}
      vtypePath = parseArguments("vtypepath", argv);
      if(vtypePath == null && genCorba)
	{
	  throw new CifParseException
	    ("MMSExpr: Insufficient input - must specify pathname" +
	     "\n\t for generated valuetype inits (vtypepath=<>)");
	}
      xmlPath = parseArguments("xmlpath", argv);
      if(xmlPath == null && genXML)
	{
	  throw new CifParseException
	    ("MMSExpr: Insufficient input - must specify pathname" +
	     "\n\t for generated XML DTD files (xmlpath=<>)");
	}
      sqlPath = parseArguments("sqlpath", argv);
      if(sqlPath == null && genRdb)
	{
	  throw new CifParseException
	    ("MMSExpr: Insufficient input - must specify pathname" +
	     "\n\t for generated SQL files (sqlpath=<>)");
	}
      insertPath = parseArguments("insertpath", argv);
      if(insertPath == null && genRdb)
	{
	  throw new CifParseException
	    ("MMSExpr: Insufficient input - must specify pathname" +
	     "\n\t for generated loader files (jdbcloaderpath=<>)");
	}
      mmsPackageName = parseArguments("mmspackage", argv);
      if(mmsPackageName == null)
	{
	  throw new CifParseException
	    ("MMSExpr: Insufficient input " + 
	     "- must specify MMS package name" +
	     "\n\t utility generated files (mmspackage=<>)");
	}
      lsrPackageName = parseArguments("lsrpackage", argv);
      if(lsrPackageName == null)
	{
	  throw new CifParseException
	    ("MMSExpr: Insufficient input " + 
	     "- must specify LSR package name" +
	     "\n\t OMG/LSR generated files (lsrpackage=<>)");
	}
      dct = parseArgumentsMult("dct", argv);
      if(dct.size() == 0)
	{
	  throw new CifParseException
	    ("MMSExpr: Insufficient input " +
	     "- must specify dictionary (dct=<>)");
	}
      // Create the Hierarchy representing the 
      // entire MMS Corba interface
      if (debug > 0)
	{
	  System.out.println("Creating \"" 
			     + rootType + "\" IDL Trans Structure");
	}

      if (rootType != null && rootType.equalsIgnoreCase("Test"))
	{
	  // Build a much smaller version of the IDL definition
	  // and implementation for code development
	  root = new TestRootTrans(lsrModuleName);
	}
      else if (rootType != null && rootType.equalsIgnoreCase("Std"))
	{
	  // Build Standard MMS IDL
	  root = new StdRootTrans(lsrModuleName);
	}
      else if (rootType != null && rootType.equalsIgnoreCase("Lsr"))
	{
	  // Build OMG Life Sciences Research (LSR) MMS IDL
	  root = new LsrRootTrans(lsrModuleName);
	}
      else if (rootType != null && rootType.equalsIgnoreCase("Dev"))
	{
	  // Build Development version of MMS IDL
	  root = new DevRootTrans(lsrModuleName);
	}
      else if (rootType != null && rootType.equalsIgnoreCase("Compact"))
	{
	  // Build a Compact version of MMS IDL
	  root = new CompactRootTrans(lsrModuleName);
	}
      else if (rootType != null && rootType.equalsIgnoreCase("Full"))
	{
	  // Build ***COMPLETE*** MMS IDL
	  root = new AllRootTrans(lsrModuleName);
	}
      else
	{
	  throw new CifParseException
	    ("MMSExpr: must specify root type" +
	     "\n\t ( root=< Std | Dev | Compact | Test | All >)");
	}
    }

  public String parseArguments(String param, String[] argv)
    {
      for(int i=0; i<argv.length; i++)
	{
	  if(argv[i].startsWith(param))
	    {
	      int index=argv[i].indexOf('=');
	      if(index!=-1) return (argv[i].substring(index+1));
	    }
	}
      return null;
    }

  public boolean parseArgumentsBool(String param, String[] argv)
    {
      for(int i=0; i<argv.length; i++)
	if(argv[i].equalsIgnoreCase(param)) return true;

      return false;
    }

  public ArrayList parseArgumentsMult(String param, String[] argv)
    {
      ArrayList values = new ArrayList();
      for(int i=0; i<argv.length; i++)
	{
	  if(argv[i].startsWith(param))
	    {
	      int index=argv[i].indexOf('=');
	      if(index!=-1) 
		{
		  values.add(argv[i].substring(index+1));
		}
	    }
	}
      return values;
    }

  public void writeCorbaFiles(RootTrans rts) 
    throws TransGenException
    {
      try
	{
	  String catBaseName = "CatLoader";

	  if (debug > 0)
	    {
	      System.out.println("Writing IDL file");
	    }
	  // Generate the IDL file
	  IdlWriter idlw = new IdlWriter(cifDic, idlPath);
	  if (MMSExpr.debug >= 5)
	    {
	      idlw.setTrace(true);
	    }
	  rts.accept(idlw);
	  if (debug > 0)
	    {
	      System.out.println("Writing Java files");
	    }
	  // Generate the EntryMethodImpl.java file
	  EntryWriter ew = new EntryWriter(ldrPath,
					   lsrPackageName,
					   mmsPackageName,
					   "EntryMethodImpl", catBaseName);
	  rts.accept(ew);
	  // Generate the LinkTable.java file
	  LinkWriter lw = new LinkWriter(ldrPath,
					 lsrPackageName,
					 mmsPackageName,
					 "LinkTable", catBaseName);
	  rts.accept(lw);
	  // Generate all of the Category specific java files
	  CatWriter cw = new CatWriter(cifDic, ldrPath, 
				       lsrPackageName,
				       mmsPackageName,
				       catBaseName);
	  if (MMSExpr.debug >= 5)
	    {
	      cw.setTrace(true);
	    }
	  rts.accept(cw);
	  // Generate Valuetype Implementations
	  VImplWriter vw = new VImplWriter(vtypePath, 
					   lsrPackageName,
					   mmsPackageName);
	  if (MMSExpr.debug >= 5)
	    {
	      vw.setTrace(true);
	    }
	  rts.accept(vw);
	}
      catch(IOException e)
	{
	  System.out.println("writeIdlFile IOException " + e.getMessage());
	  e.printStackTrace();
	  System.exit(1);
	}
    }

  public void writeXMLFiles(RootTrans rts) 
    throws TransGenException
    {
      try
	{
	  if (debug > 0)
	    {
	      System.out.println("Writing XML DTD");
	    }

	  // Generate the DTD file
	  ModelDtdWriter dtdw = new ModelDtdWriter(xmlPath, "Model.dtd");
	  rts.accept(dtdw);
	}
      catch(IOException e)
	{
	  System.out.println("writeXMLFiles IOException " + e.getMessage());
	  e.printStackTrace();
	  System.exit(1);
	}
    }

  public void writeRdbFiles(RootTrans rts) 
    throws TransGenException
    {
      try
	{
	  if (debug > 0)
	    {
	      System.out.println("Writing Table Create/Init SQL");
	    }
	  // Generate the IDL file
	  TableWriter csqlw = new TableWriter(cifDic, sqlPath,
					      "Mms", noMandItem);
	  rts.accept(csqlw);

	  InitSqlWriter sysqw = new InitSqlWriter(cifDic, sqlPath, "Mms");
	  rts.accept(sysqw);
	  if (debug > 0)
	    {
	      System.out.println("Writing JDBC");
	    }
	  // Generate the InsertEntry.java file
 	  InsertWriter iw = new InsertWriter(cifDic, insertPath,
					     lsrPackageName,
					     mmsPackageName,
					     "InsertEntry");
 	  rts.accept(iw);
	  // Generate the SelectCategory.java file
 	  SelectWriter sw = new SelectWriter(cifDic, insertPath,
					     lsrPackageName,
					     mmsPackageName,
					     "SelectCategory");
 	  rts.accept(sw);
	}
      catch(IOException e)
	{
	  System.out.println("writeRdbFile IOException " + e.getMessage());
	  e.printStackTrace();
	  System.exit(1);
	}
    }
}	
