//
// RootTrans.java,v 1.5 2003/05/28 01:09:06 dsg Exp
//
// Copyright 2001 The Regents of the University of California
// All Rights Reserved
//
// OpenMMS was developed by Dr. Douglas S. Greer at the San Diego
// Supercomputer Center, a research unit of the University of California,
// San Diego.  Support for this effort was provided by NSF through the
// Protein Data Bank (Grant DBI-9814284) and the National Partnership for
// Advanced Computational Infrastructure (Grant ACI-9619020)
//
// Permission to use, copy, modify and distribute any part of OpenMMS for
// educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above
// copyright notice, this paragraph and the following paragraphs appear
// in all copies.
//
// Those desiring to incorporate this OpenMMS into commercial products or
// use for commercial purposes should contact the Technology Transfer
// Office, University of California, San Diego, 9500 Gilman Drive, La
// Jolla, CA 92093-0910, Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// In no event shall the University of California be liable to any party
// for direct, indirect, special, incidental, or consequential damages,
// including lost profits, arising out of the use of this OpenMMS, even
// if the University of California has been advised of the possibility of
// such damage.
//
// The OpenMMS provided herein is on an "as is" basis, and the
// University of California has no obligation to provide maintenance,
// support, updates, enhancements, or modifications.  The University of
// California makes no representations and extends no warranties of any
// kind, either implied or express, including, but not limited to, the
// implied warranties of merchantability or fitness for a particular
// purpose, or that the use of the OpenMMS will not infringe any patent,
// trademark or other rights.

package org.rcsb.openmms.meta;

import java.io.IOException;
import java.util.Date;

import org.rcsb.openmms.cifparse.*;

/**
 * Class for the "Root" of the mmCIF translation metamodel
 *
 * @author Douglas S. Greer
 * @version 1.5
 */
abstract public class RootTrans extends Trans
{
  // defined by Root types
  abstract public void doCoreModDefs(ModuleTrans mod);
  abstract public void doReferModDefs(ModuleTrans mod);
  abstract public void doXRayModDefs(ModuleTrans mod);
  abstract public void doNmrModDefs(ModuleTrans mod);
  abstract public void doDepoModDefs(ModuleTrans mod);

  ExceptionTrans[] basicExcepts;

  /**
   * COMPILER_HAS_VALUETYPES is true if the IDL compiler
   * being used can handle valuetype declarations.
   * Set to false if valuetypes are not available and
   * "vtype" typedef'ed as a struct will be used instead.
   */
  protected static final boolean COMPILER_IMPLEMENTS_VALUETYPES = true;

  /**
   * USE_VALUETYPES determines if structs or valuetypes 
   * are used for default mmCIF category container
   */
  protected static final boolean USE_VALUETYPES = true;

  public static String baseModuleName;
  public static String coreModuleName;

  FileTrans currentFile;

  static final int MAX_PARAGRAPHS = 100;
  static final boolean WRITE_IDL_COMMENTS = false;

  static public String vTypeName;
  static public String vFieldAccess;

  // structs used by the modules
  protected StructureTrans stIndexId;
  protected StructureTrans stVectorXYZ;
  protected StructureTrans seqIndex;
  protected StructureTrans atomIndex;

  // structs defined for XRayCrystallography module
  protected StructureTrans unitCell;
  protected StructureTrans millerIndices;

  public RootTrans(String baseModName)
    {
      super("Root Node", Trans.TYPE_ROOT);
      // must include " " after type if needed.
      if (USE_VALUETYPES)
	{
	  if (COMPILER_IMPLEMENTS_VALUETYPES)
	    {
	      vTypeName = "valuetype ";
	      vFieldAccess = "public ";
	    }
	  else
	    {
	      // marks #define type as a valuetype wannabe
	      vTypeName = "vtype ";
	      vFieldAccess = "";
	    }
	}
      else
	{
	  vTypeName = "struct ";
	  vFieldAccess = "";
	}
      init(baseModName);
    }

  public void accept(TransVisitor v)
    throws IOException, TransGenException
    {
      v.openRoot(this);
      // visit children
      Trans t;
      for (int i = 0; i < numberOfChildren(); i++)
	{
	  t = getChild(i);
	  int type = t.getTransType();
	  switch (type)
	    {
	    case Trans.TYPE_FILE:
	      t.accept(v);
	      break;
	    default:
	      throw new TransGenException
		("Invalid RootTrans Child Type: " + type );
	    }
	}
      v.closeRoot(this);
    }

  // Must be a short string value describing the model
  abstract public String getModelIdentifier();

  // Must be a value of the form "x.y" that can be converted to a float
  // for comparison purposes.
  // Additions to the model/schema can be done with a minor release
  // i.e incrementing "x.y" to "x.y+1".  Subtractions or substitutions
  // require a major release i.e changing "x.y" to "x+1.y".
  abstract public String getRevisionNumber();

  // A much smaller version of the IDL definition for
  // code development
  public void init(String baseModName)
    {
      int n;	// module number
      ModuleTrans coreMod;
      ModuleTrans mod;

      // CORE MODULE
      baseModuleName = baseModName;
      n = ModuleTypes.MMS_CORE;
      coreModuleName = baseModName + ModuleTypes.MODULE_NAMES[n];
      coreMod = createIdlFile(n, baseModName,
			      ModuleTypes.MODULE_NAMES[n],
			      ModuleTypes.MODULE_VARS[n],
			      null);
      doCoreModDefs(coreMod);
      doFlags(coreMod);
      doEntry(coreMod);
      finIdlFile(coreMod);

      // XRAY_CRYSTALLOGRAPHY MODULE
      n = ModuleTypes.MMS_XRAY_CRYSTALLOGRAPHY;
      mod = createIdlFile(n, baseModName,
			  ModuleTypes.MODULE_NAMES[n],
			  ModuleTypes.MODULE_VARS[n],
			  coreMod);
      doXRayModDefs(mod);
      doEntry(mod);
      finIdlFile(mod);

      // NMR_MODULE
      n = ModuleTypes.MMS_NUCLEAR_MAGNETIC_RESONANCE;
      mod = createIdlFile(n, baseModName,
			  ModuleTypes.MODULE_NAMES[n],
			  ModuleTypes.MODULE_VARS[n],
			  coreMod);
      doNmrModDefs(mod);
      doEntry(mod);
      finIdlFile(mod);

      // REFERENCE MODULE
      n = ModuleTypes.MMS_REFERENCE;
      mod = createIdlFile(n, baseModName,
			  ModuleTypes.MODULE_NAMES[n],
			  ModuleTypes.MODULE_VARS[n],
			  coreMod);
      doReferModDefs(mod);
      doEntry(mod);
      finIdlFile(mod);

      // DEPOSITION MODULE
      n = ModuleTypes.MMS_DEPOSITION;
      mod = createIdlFile(n, baseModName,
			  ModuleTypes.MODULE_NAMES[n],
			  ModuleTypes.MODULE_VARS[n],
			  coreMod);
      currentFile.setContainsEntry(false);
      doDepoModDefs(mod);
      finIdlFile(mod);
    }

  protected ModuleTrans createIdlFile(int number, String base, String ext,
				      String var, ModuleTrans core)
    {
      ModuleTrans mod;

      currentFile = new FileTrans(base, ext, ".idl", core);
      this.add(currentFile);
      mod = new ModuleTrans(number, base, ext, var, core);

      // opening goes before module in FileTrans's list of children
      doOpening(mod);
      if (mod.isCoreModule())
	{
	  doExceptions(mod);
	  doTypeDefs(mod);
	}
      currentFile.add(mod);
      return mod;
    }

  protected void finIdlFile(ModuleTrans mod)
    {
      if (mod.isCoreModule())
	{
	  doMMSServer(mod);
	}

      // close guard directive
      DirectiveTrans closeguard
	    = new DirectiveTrans("#endif //", currentFile.getGuardName(),
				 "", 1);
      currentFile.add(closeguard);
    }

  public void doOpening(ModuleTrans cmod)
    {
      String mnam;
      DirectiveTrans d;

      // OMG IDL style requiredment for first line
      d = new DirectiveTrans("// File:", currentFile.getName(), "", 0);
      currentFile.add(d);

      // OMG IDL required Guard
      d = new DirectiveTrans("#ifndef", currentFile.getGuardName(), "", 1);
      currentFile.add(d);
      d = new DirectiveTrans("#define", currentFile.getGuardName(), "", 0);
      currentFile.add(d);

      if (cmod.isCoreModule())
	{
	  mnam = "Core";
	}
      else
	{
	  mnam = cmod.getExtensionName();
	}
      String[] cr = {
	"",
	"Macromolecular Structure CORBA IDL",
	mnam + " Module",
	"Created " + (new Date()).toString(),
	"",
	"Copyright 2000",
	"    Douglas Greer, John Westbrook and Phil Bourne,",
	"    Research Collaboratory for Structural Bioinfomatics,",
	"    International Union of Crystallography",
	""
	  };
      d = new DirectiveTrans(null, "", "", 1);
      d.setComment(cr);
      currentFile.add(d);

      // includes
      if (cmod.isCoreModule())
	{
	  d = new DirectiveTrans("#include", "<CosPropertyService.idl>",
				 "", 1);
	  currentFile.add(d);
	  d = new DirectiveTrans("#include", "<TimeBase.idl>",
				 "", 0);
	  currentFile.add(d);
	  d = new DirectiveTrans("#include", "<BaseIDL.idl>",
				 "", 0);
	  currentFile.add(d);
	}
      if ( ! cmod.isCoreModule())
	{
	  d = new DirectiveTrans("#include", "\"" + coreModuleName + ".idl\"",
				 "", 1);
	  currentFile.add(d);
	}
      // %% - metamodel access
      // writeStr("#include <BaseIDL.idl>");

      d = new DirectiveTrans("#pragma", "prefix", "\"omg.org\"", 1);
      currentFile.add(d);

      // For vtype typedef
      if ( ! COMPILER_IMPLEMENTS_VALUETYPES)
	{
	  // temporary defines for Object-by-Value simulation
	  DirectiveTrans valuetypedef 
	    = new DirectiveTrans("#define", "vtype", "struct", 1);
	  String[] vtc = {
	    "ValueType is defined as a Struct for code development "
	    + "and testing"
	  };
	  valuetypedef.setComment(vtc);

	  // defines for simulation of valutypes with structs
	  cmod.add(valuetypedef);
	}

      // Preface finished - beginning of IDL proper
    }
      
  public void doExceptions(ModuleTrans cmod)
    {
      // Exceptions are defined in the primary file
      ExceptionTrans dataAccessExcept = 
	new ExceptionTrans("DataAccessException");
      dataAccessExcept.add(new ExceptionFieldTrans("string",
						   "method_name"));
      dataAccessExcept.add(new ExceptionFieldTrans("string",
						   "description"));
      if (WRITE_IDL_COMMENTS)
	{
	  String[] daeCom = {
	    "A DataAccessException is throw whenever requested data is",
	    "not available.  The reason for the exception is given",
	    "in the description field."
	  };
	  dataAccessExcept.setComment(daeCom);
	}

      basicExcepts = new ExceptionTrans[1];
      basicExcepts[0] = dataAccessExcept;

      cmod.add(dataAccessExcept); 
    }

  public void doTypeDefs(ModuleTrans cmod)
    {
      FieldTrans ft;

      // define vector and martix types
      TypeDefTrans identdef = new TypeDefTrans
	("Identifier", "string");
      identdef.setBlankLines(1);
      cmod.add(identdef);

      // define vector and martix types
      TypeDefTrans v3def = new TypeDefTrans
	("Vector3[3]", "float");
      v3def.setBlankLines(1);
      if (WRITE_IDL_COMMENTS)
	{
	  String[] v3Com = {
	    "Representation of a 3D spatial position or translation vector."
	  };
	  v3def.setComment(v3Com);
	}
      cmod.add(v3def);
      
      TypeDefTrans m3def = new TypeDefTrans
	("Matrix3[3]", "Vector3");
      if (WRITE_IDL_COMMENTS)
	{
	  String[] m3Com = {
	    "Representation of a rotation matrix in 3D Euclidean space."
	  };
	  m3def.setComment(m3Com);
	}
      cmod.add(m3def);

      TypeDefTrans fmtt = new TypeDefTrans("FormatType",
					  "string");
      fmtt.setBlankLines(1);
      cmod.add(fmtt);

      TypeDefTrans ftl = new TypeDefTrans("FormatTypeList",
					  "sequence<FormatType>");
      if (WRITE_IDL_COMMENTS)
	{
	  String[] ftlCom = {
	    "List of native formats supported for updates and deposition"
	  };
	  ftl.setComment(ftlCom);
	}
      cmod.add(ftl);

      TypeDefTrans nr = new TypeDefTrans("EntryRepresentation",
					 "sequence<octet>");
      if (WRITE_IDL_COMMENTS)
	{
	  String[] nrCom = {
	    "Representation of an entry in a native server format"
	  };
	  nr.setComment(nrCom);
	}
      cmod.add(nr);

      stIndexId = new StructureTrans("IndexId", null, false);
      stIndexId.add(new FieldTrans("string", "id", false));
      stIndexId.add(new FieldTrans("long", "index", false));
      cmod.add(stIndexId);

      stVectorXYZ = new StructureTrans("VectorXYZ", null, false);
      stVectorXYZ.add(new FieldTrans("float", "x", false));
      stVectorXYZ.add(new FieldTrans("float", "y", false));
      stVectorXYZ.add(new FieldTrans("float", "z", false));
      cmod.add(stVectorXYZ);

      seqIndex = new StructureTrans("SeqIndex", null, false);
      ft = new FieldTrans("IndexId", "seq", false);
      ft.setFieldStructure(stIndexId);
      seqIndex.add(ft);
      ft = new FieldTrans("IndexId", "comp", false);
      ft.setFieldStructure(stIndexId);
      seqIndex.add(ft);
      ft = new FieldTrans("IndexId", "asym", false);
      ft.setFieldStructure(stIndexId);
      seqIndex.add(ft);
      ft = new FieldTrans("IndexId", "alt", false);
      ft.setFieldStructure(stIndexId);
      seqIndex.add(ft);
      cmod.add(seqIndex);

      atomIndex = new StructureTrans("AtomIndex", null, false);
      ft = new FieldTrans("IndexId", "atom", false);
      ft.setFieldStructure(stIndexId);
      atomIndex.add(ft);
      ft = new FieldTrans("IndexId", "seq", false);
      ft.setFieldStructure(stIndexId);
      atomIndex.add(ft);
      ft = new FieldTrans("IndexId", "comp", false);
      ft.setFieldStructure(stIndexId);
      atomIndex.add(ft);
      ft = new FieldTrans("IndexId", "asym", false);
      ft.setFieldStructure(stIndexId);
      atomIndex.add(ft);
      ft = new FieldTrans("IndexId", "alt", false);
      ft.setFieldStructure(stIndexId);
      atomIndex.add(ft);
      cmod.add(atomIndex);
    }

  public void doXRayTypeDefs(ModuleTrans cmod, ModuleTrans[] modQual)
    {
      unitCell = new StructureTrans("UnitCell", modQual, false);
      unitCell.add(new FieldTrans("float", "length_a", false));
      unitCell.add(new FieldTrans("float", "length_b", false));
      unitCell.add(new FieldTrans("float", "length_c", false));
      unitCell.add(new FieldTrans("float", "angle_alpha", false));
      unitCell.add(new FieldTrans("float", "angle_beta", false));
      unitCell.add(new FieldTrans("float", "angle_gamma", false));
      cmod.add(unitCell);

      millerIndices = new StructureTrans("MillerIndices", modQual, false);
      millerIndices.add(new FieldTrans("long", "h", false));
      millerIndices.add(new FieldTrans("long", "k", false));
      millerIndices.add(new FieldTrans("long", "l", false));
      cmod.add(millerIndices);
    }

  public void doNmrTypeDefs(ModuleTrans cmod, ModuleTrans[] modQual)
    {
    }

  public void doFlags(ModuleTrans cmod)
    {
      //___________________________________________________________
      // Flag Vector
      SequenceTrans fvs = new SequenceTrans("Flags", "octet");
      fvs.setBlankLines(1);
      if (WRITE_IDL_COMMENTS)
	{
	  String[] fvsCom = {
	    "A Flags vector returned by get_presences_flags() is used",
	    "to efficiently determine those structs that are present",
	    "for a given entry, and which fields in each struct",
	    "are valid.  The Flag represents a sequence of bits",
	    "with a bit set to \"1\" indicating a struct",
	    "or field is present and valid.",
	    "The index of the octet within the sequence is determined by",
	    "integer division of the flags numeric value by eight (flag/8).",
	    "The bit within the octet is specified by the low 3 order",
	    "bits of the flags numeric value (1<<(flag&7))."
	  };
	  fvs.setComment(fvsCom);
	}
      cmod.add(fvs);
    }

  public void doEntry(ModuleTrans cmod)
    {
      Trans iface, t;

      //___________________________________________________________
      // ENTRY
      // Entry GetList Methods can throw exceptions basicExcepts
      iface = new InterfaceTrans(cmod.getEntryName(), basicExcepts);
      cmod.add(iface);
	    
      
      // method to get all the flags for an entry
      String prefix = "";
      if ( ! cmod.isCoreModule())
	{
	  prefix = coreModuleName + "::";
	}

      MethodTrans gpf;
      gpf = new MethodTrans(prefix + "Flags",
			    "get_presence_flags",
			    null, null, basicExcepts);
      if (WRITE_IDL_COMMENTS)
	{
	  String[] gpfCom = {
	    "The get_presence_flag() method retrieves the",
	    "present/valid flags for an entry.",
	    "",
	    "Flags that indicate if a struct is present are",
	    "designated with a \"S_\" prefix.  Flags indicating",
	    "the validity of optional fields within a struct",
	    "are designated with a \"F_\" prefix followed by the name",
	    "of the struct.  Flags are not provided for the",
	    "mandatory fields that are always present and valid.",
	    "",
	    "In cases where a sequence of structs contains a string",
	    "field which is sometimes but not always valid, the Flag bit",
	    "is set to true and the string data values that are undefined",
	    "are represented by a period \".\"; data values that",
	    "are unknown are represented by a question mark \"?\"."
	  };
	  gpf.setComment(gpfCom);
	}
      iface.add(gpf);

      if (cmod.isCoreModule())
	{
	  t = new MethodTrans("CosPropertyService::Properties",
			      "get_subentry_list",
			      null, null, basicExcepts);
	  iface.add(t);
	}
    }

  public void doMMSServer(ModuleTrans cmod)
    {
      Trans iface, t;

      //___________________________________________________________
      // Mms Server - EntryFactory
      //
      //___________________________________________________________
      // Entry Ids
      t = new TypeDefTrans("EntryId", "Identifier");
      t.setBlankLines(1);
      if (WRITE_IDL_COMMENTS)
	{
	  String[] eitdtCom = {
	    "Unique string identifier for each entry.",
	  };
	  t.setComment(eitdtCom);
	}
      cmod.add(t);
      cmod.add(new SequenceTrans("EntryIdList", "EntryId"));
      
      //___________________________________________________________
      // Entry Group Ids
      t = new TypeDefTrans("EntryGroupId", "Identifier");
      t.setBlankLines(1);
      if (WRITE_IDL_COMMENTS)
	{
	  String[] gitdtCom = {
	    "Entry groups form a traditional two-level hierarchy for",
	    "entry lists.  The EntryGroupId is a two-letter identifier",
	    "taken from the second and third letters of all EntryIds in",
	    "that group."
	  };
	  t.setComment(gitdtCom);
	}
      cmod.add(t);
      cmod.add(new SequenceTrans("EntryGroupIdList", "EntryGroupId"));

      StructureTrans depDateStrk = 
	new StructureTrans("ModificationDate", null, false);
      depDateStrk.add(new FieldTrans("EntryId", "entry_id", false));
      t = new FieldTrans("TimeBase::TimeT", "date", false);
      if (WRITE_IDL_COMMENTS)
	{
	  String[] dateCom = {
	    "The date field is intended mainly for use in numerical",
	    "comparisons.  Its value is set to the number of",
	    "milliseconds since 00:00:00 GMT January 1, 1970."
	  };
	  t.setComment(dateCom);
	}
      depDateStrk.add(t);
      cmod.add(depDateStrk);
      cmod.add(new SequenceTrans("ModificationDateList",
				 "ModificationDate"));

      iface = new InterfaceTrans("EntryFactory");
      cmod.add(iface);

      ConstantTrans contran = new ConstantTrans
		("string", "MODEL_IDENTIFIER",
		 "\"" + getModelIdentifier() + "\"");
      iface.add(contran);
      contran = new ConstantTrans
		("string", "REVISION_NUMBER", 
		 "\"" + getRevisionNumber() + "\"");
      iface.add(contran);
      t = new MethodTrans("string", "get_version",
			  null, null, null);
      if (WRITE_IDL_COMMENTS)
	{
	  String[] gvCom = {
	    "Retrieve a string identifying the type and version",
	    "number of the server"
	  };
	  t.setComment(gvCom);
	}
      t.setBlankLines(1);
      iface.add(t);

      t = new MethodTrans("BaseIDL::ModuleDefSet",
			  "get_extension_modules",
			  null, null, null);
      iface.add(t);
      
      t = new MethodTrans("EntryIdList",
			  "get_entry_id_list",
			  null, null, basicExcepts);
      if (WRITE_IDL_COMMENTS)
	{
	  t.setBlankLines(1);
	  String[] geilCom = {
	    "Retrieve a list of all entries in all entry groups"
	  };
	  t.setComment(geilCom);
	}
      iface.add(t);

      t = new MethodTrans("long",
			  "get_entry_id_list_size",
			  null, null, basicExcepts);
      iface.add(t);

      String[] atypes = {"in long", "in long"};
      String[] anames = {"from", "to"};
      t = new MethodTrans("EntryIdList",
			  "get_entry_id_list_block_n",
			  atypes, anames, basicExcepts);
      iface.add(t);

      
      t = new MethodTrans("ModificationDateList",
			  "get_entry_modification_dates",
			  null, null, basicExcepts);
      if (WRITE_IDL_COMMENTS)
	{
	  t.setBlankLines(1);
	  String[] gemdCom = {
	    "Retrieve a list of all entries in all entry groups along",
	    "with the date they were last modified."
	  };
	  t.setComment(gemdCom);
	}
      iface.add(t);
      
      t = new MethodTrans("ModificationDateList",
			  "get_entry_modification_dates_block_n",
			  atypes, anames, basicExcepts);
      iface.add(t);

      t = new MethodTrans("EntryGroupIdList",
			  "get_entry_group_list",
			  null, null, basicExcepts);
      if (WRITE_IDL_COMMENTS)
	{
	  t.setBlankLines(1);
	  String[] geglCom = {
	    "Retrieve a list of all entry groups"
	  };
	  t.setComment(geglCom);
	}
      iface.add(t);
      
      String[] eigTypes= {"in EntryGroupId"};
      String[] eigNames= {"group"};
      t = new MethodTrans("EntryIdList", "get_entries_in_group",
			  eigTypes, eigNames, basicExcepts);
      if (WRITE_IDL_COMMENTS)
	{
	  t.setBlankLines(1);
	  String[] geigCom = {
	    "Retrieve a list of all entries contained in a specified",
	    "entry group."
	  };
	  t.setComment(geigCom);
	}
      iface.add(t);
      
      String[] efiTypes= {"in EntryId"};
      String[] efiNames= {"entry_id"};
      t = new MethodTrans("Entry", "get_entry_from_id",
			  efiTypes, efiNames, basicExcepts);
      if (WRITE_IDL_COMMENTS)
	{
	  t.setBlankLines(1);
	  String[] gefiCom = {
	    "Retrieve a reference to the Entry interface object for",
	    "the specified EntryId string."
	  };
	  t.setComment(gefiCom);
	}
      iface.add(t);
      
      t = new MethodTrans("FormatTypeList",
			  "native_formats_supported",
			  null, null, basicExcepts);
      if (WRITE_IDL_COMMENTS)
	{
	  t.setBlankLines(1);
	  String[] nfsCom = {
	    "Retrieve a list of native formats a server supports."
	  };
	  t.setComment(nfsCom);
	}
      iface.add(t);
      
      String[] nfrTypes= {"in FormatType", "in EntryId"};
      String[] nfrNames= {"format", "entry_id"};
      t = new MethodTrans("EntryRepresentation",
			  "get_native_entry_representation",
			  nfrTypes, nfrNames, basicExcepts);
      if (WRITE_IDL_COMMENTS)
	{
	  t.setBlankLines(1);
	  String[] gnfrCom = {
	    "Retrieve a native format represention of an entry."
	  };
	  t.setComment(gnfrCom);
	}
      iface.add(t);
    }

  public StructureTrans defaultContainer(String sname,
					 ModuleTrans[] mq, boolean isCif)
    {
      // Constructor for non-Cif structures
      StructureTrans s;

      if (USE_VALUETYPES)
	{
	  s = new ValuetypeTrans(sname, null, mq, 0);
	}
      else
	{
	  s = new StructureTrans(sname, null, mq, 0);
	}
      s.setIsCif(isCif);
      return s;
    }

  public StructureTrans defaultContainer(String sname, CategoryTrans ct)
    {
      // Constructor for base module, all comments
      StructureTrans s;

      if (USE_VALUETYPES)
	{
	  s = new ValuetypeTrans(sname, ct, null, MAX_PARAGRAPHS);
	}
      else
	{
	  s = new StructureTrans(sname, ct, null, MAX_PARAGRAPHS);
	}
      return s;
    }

  public StructureTrans defaultContainer(String sname,
					 CategoryTrans ct, int descp)
    {
      // Constructor for base module, limited comment paragraphs
      StructureTrans s;

      if (USE_VALUETYPES)
	{
	  s = new ValuetypeTrans(sname, ct, null, descp);
	}
      else
	{
	  s = new StructureTrans(sname, ct, null, descp);
	}
      return s;
    }

  public StructureTrans defaultContainer(String sname, CategoryTrans ct,
			ModuleTrans[] mq)
    {
      // Constructor for opt module
      StructureTrans s;

      if (USE_VALUETYPES)
	{
	  s = new ValuetypeTrans(sname, ct, mq, MAX_PARAGRAPHS);
	}
      else
	{
	  s = new StructureTrans(sname, ct, mq, MAX_PARAGRAPHS);
	}
      return s;
    }

  public StructureTrans defaultContainer(String sname, CategoryTrans ct, 
			ModuleTrans[] mq, int descp)
    {
      // Constructor for opt module, limited comments
      StructureTrans s;

      if (USE_VALUETYPES)
	{
	  s = new ValuetypeTrans(sname, ct, mq, descp);
	}
      else
	{
	  s = new StructureTrans(sname, ct, mq, descp);
	}
      return s;
    }

}
