//
// SqlWriter.java,v 1.6 2003/05/30 18:57:02 dsg Exp
//
// Copyright 2001 The Regents of the University of California
// All Rights Reserved
//
// OpenMMS was developed by Dr. Douglas S. Greer at the San Diego
// Supercomputer Center, a research unit of the University of California,
// San Diego.  Support for this effort was provided by NSF through the
// Protein Data Bank (Grant DBI-9814284) and the National Partnership for
// Advanced Computational Infrastructure (Grant ACI-9619020)
//
// Permission to use, copy, modify and distribute any part of OpenMMS for
// educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above
// copyright notice, this paragraph and the following paragraphs appear
// in all copies.
//
// Those desiring to incorporate this OpenMMS into commercial products or
// use for commercial purposes should contact the Technology Transfer
// Office, University of California, San Diego, 9500 Gilman Drive, La
// Jolla, CA 92093-0910, Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// In no event shall the University of California be liable to any party
// for direct, indirect, special, incidental, or consequential damages,
// including lost profits, arising out of the use of this OpenMMS, even
// if the University of California has been advised of the possibility of
// such damage.
//
// The OpenMMS provided herein is on an "as is" basis, and the
// University of California has no obligation to provide maintenance,
// support, updates, enhancements, or modifications.  The University of
// California makes no representations and extends no warranties of any
// kind, either implied or express, including, but not limited to, the
// implied warranties of merchantability or fitness for a particular
// purpose, or that the use of the OpenMMS will not infringe any patent,
// trademark or other rights.

package org.rcsb.openmms.meta;

import java.io.*;
import java.util.*;

import org.rcsb.openmms.cifparse.*;
import org.rcsb.openmms.util.*;

/**
 * Common Base class for metamodel Visitors that generate SQL code
 *
 * @author Douglas S. Greer
 * @version 1.6
 */
public class SqlWriter extends TransVisitor
			       implements TypeNamesSql
{
  String dirPath;
  String oFileBase;
  CifDictionary cifDic;

  final static String INDENT = "    ";
  final static String MANDATORY = "NOT NULL";

  String[] padLongTextType = new String[NUMBER_OF_FLAVORS];

  // class  with flexible methods for creating lists of name/value pairs.
  class StringPairBlock
  {
    StringStruct pb;
    StringStruct names;
    StringStruct values;

    int pairsSoFar = 0;
    String separator;

    StringPairBlock(StringStruct sst, String pre,
		    String in, String post, String sep)
      {
	separator = sep;

	sst.writeString(pre);
	names = new StringStruct();
	sst.writeStringStruct(names);
	sst.writeString(in);
	values = new StringStruct();
	sst.writeStringStruct(values);
	sst.writeString(post);
      }
    
    protected void addStringPair(String n, String v)
      {
	if (pairsSoFar > 0)
	  {
	    names.writeString(separator);
	    values.writeString(separator);
	  }
	names.writeString(n);
	values.writeString(v);
	pairsSoFar++;
      }
  }  // StringPairBlock class

  public SqlWriter(CifDictionary d, String dir, String oBase)
    {
      cifDic = d;
      dirPath = dir;
      oFileBase = oBase;
      if (oFileBase == null)
	{
	  // default
	  oFileBase = "";
	}
      for (int i = 0; i < NUMBER_OF_FLAVORS; i++)
	{
	  padLongTextType[i] = padString(bigTextType[i], TYPE_FIELD_LEN);
	}
    }

  protected void writeIntro(StringStruct sst, String desc)
    {
      Date dt = new Date();
      GregorianCalendar cal = new GregorianCalendar();
      cal.setTime(dt);

      String[] header =
      {
	"--",
	"-- PDB Macromolecular Structure",
	"-- SQL Definitions" + desc,
	"--     written: " + dt.toString(),
	"--",
	"-- Copyright " + cal.get(Calendar.YEAR),
	"--     Douglas Greer, John Westbrook and Phil Bourne,",
	"--     Research Collaboratory for Structural Bioinfomatics,",
	"--     International Union of Crystallography",
	"--"
      };
      sst.writeLine(header);
    }

  // database specific strings
  final String[] dbSpecific =
  {
    "set escape off\nset escape \\\nset escape on",  // Oracle
    "-- <None>",                                     // Sybase
    "-- <None>",                                     // Informix
    "-- <None>",                                     // MySQL
    "-- <None>",                                     // IBM DB2
    "-- <None>"                                      // MS Access
  };

  protected void writeSysSpecific(StringStruct sst)
    {
      // db specific commands
      sst.writeLine("");
      sst.writeLine("");
      sst.writeLine("--");
      sst.writeIndentedString
	("-- Setup commands specific for database type: ", 0);
      sst.writeSelection(dbTypeNames);
      sst.writeLine("");
      sst.writeLine("--");
      sst.writeSelection(dbSpecific);
      sst.writeLine("");
      sst.writeLine("--");
      sst.writeLine("-- End of database specific commands");
      sst.writeLine("--");
    }

  public String categoryNameToTableName(CategoryTrans ct)
    {
      String categoryName = clearIllegalChars(ct.getName());

      return categoryName.toUpperCase();
    }

  public String itemNameToAttributeName(FieldTrans ft)
    {
      String refName = ft.getFullRefName();
      String iName = clearIllegalChars(refName);

      // return iName.toLowerCase();
      return iName;
    }

  // db must be set up so the the backslash '\' character can be used to
  // escape newlines, itself (\\) and the ampersand '&' character
  public String literal(String s)
    {
      StringBuffer sb = new StringBuffer(s);

      for (int i = 0; i < sb.length(); i++)
	{
	  switch (sb.charAt(i))
	    {
	    case '\'':
	      sb.insert(i, '\'');
	      i++;
	      break;
	    case '\n':
	    case '\\':
	    case '&':
	      sb.insert(i, '\\');
	      i++;
	      break;
	    default:
	      break;
	    }
	}
      // quote the whole string
      sb.insert(0, '\'');
      sb.append('\'');
      return sb.toString();
    }

  public String clearIllegalChars(String s)
    {
      StringBuffer sb = new StringBuffer(s);
      
      // march through the word replacing bad characters with "_"
      for (int i = 0; i < sb.length(); i++)
	{
	  switch (sb.charAt(i))
	    {
	    case '[':
	    case '\'':
	      sb.setCharAt(i, '_');
	      break;
	    case ']':
	      sb.deleteCharAt(i);
	      i--;
	      break;
	    }
	}
      return sb.toString();
    }

  public String cifToSqlType(String ct, int min) throws TransGenException
    {
      String rval;

      if (ct == null)
	{
	  // index to parent type
	  rval = STD_LINE_TYPE;
	}
      else if (ct.compareToIgnoreCase("float") == 0)
	{
	  // rval = "NUMBER(6, 2)";
	  rval = "FLOAT";
	}
      else if (ct.compareToIgnoreCase("int") == 0)
	{
	  // rval = "NUMBER(6, 0)";
	  rval = "INTEGER";
	}
      else if (ct.compareToIgnoreCase("uchar1") == 0)
	{
	  rval = "CHAR(1)";
	}
      else if (ct.compareToIgnoreCase("uchar3") == 0)
	{
	  rval = "CHAR(3)";
	}
      else if (ct.compareToIgnoreCase("yyyy-mm-dd") == 0)
	{
	  rval = "CHAR(10)";
	}
      else if (ct.compareToIgnoreCase("name") == 0 ||
	       ct.compareToIgnoreCase("idname") == 0 ||
	       ct.compareToIgnoreCase("ucode") == 0 ||
	       ct.compareToIgnoreCase("code") == 0 ||
	       ct.compareToIgnoreCase("atcode") == 0 ||
	       ct.compareToIgnoreCase("float-range") == 0 ||
	       ct.compareToIgnoreCase("yyyy-mm-dd:hh:mm") == 0 ||
	       ct.compareToIgnoreCase("fax") == 0 ||
	       ct.compareToIgnoreCase("phone") == 0 ||
	       ct.compareToIgnoreCase("email") == 0)
	{
	  rval = STD_LINE_TYPE;
	}
      else if (ct.compareToIgnoreCase("text") == 0 ||
	       ct.compareToIgnoreCase("line") == 0 ||
	       ct.compareToIgnoreCase("uline") == 0 ||
	       ct.compareToIgnoreCase("any") == 0 ||
	       ct.compareToIgnoreCase("symop") == 0)
	{
	  rval = "TEXT";
	}
      else
	{
	  throw new TransGenException("cifToSqlType - Unknown type: " + ct);
	}

      if (min > 0)
	{
	  rval = padString(rval, min);
	}
      return rval;
    }

  public String cifToSqlSetMethod(FieldTrans ft)
    throws TransGenException
    {
      String rval;
      DictionaryItem di = ft.getDictionaryItem();
      String ct = di.getItemType();

      if (ft.isFieldBitSet(FieldTrans.IS_A_STREAM_DATATYPE))
	{
	  rval = "setAsciiStream";
	}
      else if (ct == null)
	{
	  rval = "setString";
	}
      else if (ct.compareToIgnoreCase("float") == 0)
	{
	  rval = "setFloat";
	}
      else if (ct.compareToIgnoreCase("int") == 0)
	{
	  rval = "setInt";
	}
      else if (ct.compareToIgnoreCase("uchar1") == 0 ||
	       ct.compareToIgnoreCase("uchar3") == 0 ||
	       ct.compareToIgnoreCase("yyyy-mm-dd") == 0 ||
	       ct.compareToIgnoreCase("name") == 0 ||
	       ct.compareToIgnoreCase("idname") == 0 ||
	       ct.compareToIgnoreCase("line") == 0 ||
	       ct.compareToIgnoreCase("uline") == 0 ||
	       ct.compareToIgnoreCase("ucode") == 0 ||
	       ct.compareToIgnoreCase("code") == 0 ||
	       ct.compareToIgnoreCase("atcode") == 0 ||
	       ct.compareToIgnoreCase("text") == 0 ||
	       ct.compareToIgnoreCase("any") == 0 ||
	       ct.compareToIgnoreCase("symop") == 0 ||
	       ct.compareToIgnoreCase("float-range") == 0 ||
	       ct.compareToIgnoreCase("yyyy-mm-dd:hh:mm") == 0 ||
	       ct.compareToIgnoreCase("fax") == 0 ||
	       ct.compareToIgnoreCase("phone") == 0 ||
	       ct.compareToIgnoreCase("email") == 0)
	{
	  rval = "setString";
	}
      else
	{
	  throw new TransGenException("cifToSqlSetMethod - Unknown type: "
				      + ct);
	}
      return rval;
    }

  public String cifToSqlGetMethod(FieldTrans ft)
    throws TransGenException
    {
      String rval;
      DictionaryItem di = ft.getDictionaryItem();
      String ct = di.getItemType();

      if (ft.isFieldBitSet(FieldTrans.IS_A_STREAM_DATATYPE))
	{
	  rval = "getAsciiStream";
	}
      else if (ct == null)
	{
	  rval = "getString";
	}
      else if (ct.compareToIgnoreCase("float") == 0)
	{
	  rval = "getFloat";
	}
      else if (ct.compareToIgnoreCase("int") == 0)
	{
	  rval = "getInt";
	}
      else if (ct.compareToIgnoreCase("uchar1") == 0 ||
	       ct.compareToIgnoreCase("uchar3") == 0 ||
	       ct.compareToIgnoreCase("yyyy-mm-dd") == 0 ||
	       ct.compareToIgnoreCase("name") == 0 ||
	       ct.compareToIgnoreCase("idname") == 0 ||
	       ct.compareToIgnoreCase("line") == 0 ||
	       ct.compareToIgnoreCase("uline") == 0 ||
	       ct.compareToIgnoreCase("ucode") == 0 ||
	       ct.compareToIgnoreCase("code") == 0 ||
	       ct.compareToIgnoreCase("atcode") == 0 ||
	       ct.compareToIgnoreCase("text") == 0 ||
	       ct.compareToIgnoreCase("any") == 0 ||
	       ct.compareToIgnoreCase("symop") == 0 ||
	       ct.compareToIgnoreCase("float-range") == 0 ||
	       ct.compareToIgnoreCase("yyyy-mm-dd:hh:mm") == 0 ||
	       ct.compareToIgnoreCase("fax") == 0 ||
	       ct.compareToIgnoreCase("phone") == 0 ||
	       ct.compareToIgnoreCase("email") == 0)
	{
	  rval = "getString";
	}
      else
	{
	  throw new TransGenException("cifToSqlSetMethod - Unknown type: "
				      + ct);
	}
      return rval;
    }

  public String padString(String s)
    {
      return padString(s, 30);
    }

  public String padString(String s, int min)
    {
      StringBuffer sb = new StringBuffer(s);
      
      // march through the word replacing bad characters with "_"
      for (int i = sb.length(); i < min; i++)
	{
	  sb.append(' ');
	}
      return sb.toString();
    }

  protected void buffer(StringStruct sst, String s1, String[] sa, String s2)
    {
      flush(sst, false);
      sst.writeIndentedString(s1, 0);
      sst.writeSelection(sa);
      sst.writeString(s2);
      sst.setAuxString("");
    }


  protected void buffer(StringStruct sst, String s)
    {
      flush(sst, false);
      sst.setAuxString(s);
    }

  protected void flush(StringStruct sst, boolean last)
    {
      String buf = sst.getAuxString();

      if (buf != null)
	{
	  // last attribute in table is not terminate with a ","
	  if (last)
	    {
	      sst.writeLine(buf);
	    }
	  else
	    {
	      if (buf.length() > 0)
		{
		  sst.writeLine(buf + ",");
		}
	      else
		{
		  sst.writeString(",\n");
		}
	    }
	  sst.setAuxString(null);
	}
    }
}  

