//
// StringStruct.java,v 1.5 2003/03/21 23:59:00 dsg Exp
//
// Copyright 2001 The Regents of the University of California
// All Rights Reserved
//
// OpenMMS was developed by Dr. Douglas S. Greer at the San Diego
// Supercomputer Center, a research unit of the University of California,
// San Diego.  Support for this effort was provided by NSF through the
// Protein Data Bank (Grant DBI-9814284) and the National Partnership for
// Advanced Computational Infrastructure (Grant ACI-9619020)
//
// Permission to use, copy, modify and distribute any part of OpenMMS for
// educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above
// copyright notice, this paragraph and the following paragraphs appear
// in all copies.
//
// Those desiring to incorporate this OpenMMS into commercial products or
// use for commercial purposes should contact the Technology Transfer
// Office, University of California, San Diego, 9500 Gilman Drive, La
// Jolla, CA 92093-0910, Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// In no event shall the University of California be liable to any party
// for direct, indirect, special, incidental, or consequential damages,
// including lost profits, arising out of the use of this OpenMMS, even
// if the University of California has been advised of the possibility of
// such damage.
//
// The OpenMMS provided herein is on an "as is" basis, and the
// University of California has no obligation to provide maintenance,
// support, updates, enhancements, or modifications.  The University of
// California makes no representations and extends no warranties of any
// kind, either implied or express, including, but not limited to, the
// implied warranties of merchantability or fitness for a particular
// purpose, or that the use of the OpenMMS will not infringe any patent,
// trademark or other rights.

package org.rcsb.openmms.meta;

import java.io.*;
import java.util.*;

/**
 * StringStruct can be used to create a tree of string buffers
 * where new text can be dynamicly added to any buffer in the
 * the tree. String "selectors" can be used for key words to
 * create multiple versions by choosing a single String from
 * an array. The tree is then traversed and the strings are
 * written to a file or to one huge string.
 *
 * @author Douglas S. Greer
 * @version 1.5
 */
public class StringStruct
{
  private static String indentString = "    ";
  private String auxString = null;

  private int indentLevel = 0;

  // selector must be static so when it is set it is seen by 
  // all StringStructs
  private static int selector = 0;

  private int blankBefore = 0;
  private int blankAfter = 0;

  class Kid 
  {
    static final short BUFFER = 0;
    static final short SELECTOR = 1;
    static final short BLOCK = 2;

    short type;
    StringBuffer sbuf;
    StringStruct block;
    String[] selections;
  }

  protected Kid current = null;
  protected ArrayList kidList = new ArrayList();

  public StringStruct()
    {
    }

  public StringStruct(StringStruct parent)
    {
      // inherit indentation
      indentLevel = parent.getIndentLevel();
      indentString = parent.getIndentString();
      parent.writeStringStruct(this);
    }

  public void beginIndent()
    {
      indentLevel++;
    }

  public void endIndent()
    {
      if (indentLevel > 0)
	{
	  indentLevel--;
	}
    }

  public void beginIndent(int n)
    {
      indentLevel += n;
    }

  public void endIndent(int n)
    {
      indentLevel = Math.max(0, indentLevel - n);
    }

  public int getIndentLevel()
    {
      return indentLevel;
    }

  public void setIndentString(String s)
    {
      indentString = s;
    }

  public String getIndentString()
    {
      return indentString;
    }

  public void setAuxString(String s)
    {
      auxString = s;
    }

  public String getAuxString()
    {
      return auxString;
    }

  public void setBlankLines(int before, int after)
    {
      blankBefore = before;
      blankAfter = after;
    }

  public void setSelector(int n)
    {
      selector = n;
    }

  public int getSelector()
    {
      return selector;
    }

  public void writeSelection(String[] sels)
    {
      current = new Kid();
      current.type = Kid.SELECTOR;
      current.selections = sels;
      kidList.add(current);
    }

  public void writeStringStruct(StringStruct b)
    {
      current = new Kid();
      current.type = Kid.BLOCK;
      current.block = b;
      kidList.add(current);
    }

  public StringStruct newBlock(String opening, String closing)
    {
      String[] tmpo =
      {
	opening
      };
      String[] tmpc =
      {
	closing
      };
      return newBlock(tmpo, tmpc);
    }

  public StringStruct newBlock(String[] opening, String closing)
    {
      String[] tmpc =
      {
	closing
      };
      return newBlock(opening, tmpc);
    }

  public StringStruct newBlock(String[] opening, String[] closing)
    {
      writeLine(opening);
      StringStruct rval = new StringStruct(this);
      writeLine(closing);
      return rval;
    }

  public StringStruct newMethodBlock(String methodName, String tosses)
    {
      return newMethodBlock(methodName, null, tosses);
    }

  public StringStruct newMethodBlock(String methodName, String line2,
				     String tosses)
    {
      writeLine("");
      writeLine(methodName);
      if (line2 != null)
	{
	  writeLine("        " + line2);
	}
      if (tosses != null)
	{
	  writeLine("  " + tosses);
	}
      writeLine("  {");
      beginIndent();
      StringStruct rval = new StringStruct(this);
      endIndent();
      writeLine("  }");
      return rval;
    }

  protected void needBuffer()
    {
      if (current == null || current.type != Kid.BUFFER)
	{
	  current = new Kid();
	  current.type = Kid.BUFFER;
	  current.sbuf = new StringBuffer();
	  kidList.add(current);
	}
    }

  public void traverseToFile(String fname)
    throws IOException, TransGenException
    {
      BufferedWriter ofw = new BufferedWriter(new FileWriter(fname));
      String s = traverseToStringBuffer().toString();
      ofw.write(s);
      ofw.close();
    }

  public StringBuffer traverseToStringBuffer()
    throws TransGenException
    {
      Kid k;
      StringBuffer rval = new StringBuffer();

      for (int i = 0; i < kidList.size(); i++)
 	{
 	  k = (Kid) kidList.get(i);
	  switch (k.type)
	    {
	    case Kid.BUFFER:
	      rval.append(k.sbuf);
	      break;
	    case Kid.SELECTOR:
	      rval.append(k.selections[selector]);
	      break;
	    case Kid.BLOCK:
	      // recurse subblocks
	      rval.append(k.block.traverseToStringBuffer());
	      break;
	    default:
	      throw new TransGenException("StringStruct.tranverseToString(): "
				  + "Unknown Kid type");
	    }
 	}
      return rval;
    }

  public void writeLine(String s)
    {
      writeLine(s, 0);
    }

  public void writeString(String s)
    {
      needBuffer(); // make sure we're appending to a StringBuffer.
      current.sbuf.append(s);
    }

  public void writeIndentedString(String s, int ilev)
    {
      needBuffer(); // make sure we're appending to a StringBuffer.
      if (s != null && s.length() > 0)
	{
	  for (int i = 0; i < indentLevel + ilev; i++)
	    {
	      current.sbuf.append(indentString);
	    }
	  current.sbuf.append(s);
	}
    }

  public void writeLine(String s, int ilev)
    {
      needBuffer(); // make sure we're appending to a StringBuffer.
      writeIndentedString(s, ilev);
      current.sbuf.append("\n");
    }

  public void writeLine(String[] sl)
    {
      for (int i = 0; i < sl.length; i++)
	{
	  writeLine(sl[i], 0);
	}
    }

  public void writeLine(String[] sl, int ilev)
    {
      for (int i = 0; i < sl.length; i++)
	{
	  writeLine(sl[i], ilev);
	}
    }
}  

