//
// StructureTrans.java,v 1.4 2002/06/19 01:55:29 dsg Exp
//
// Copyright 2001 The Regents of the University of California
// All Rights Reserved
//
// OpenMMS was developed by Dr. Douglas S. Greer at the San Diego
// Supercomputer Center, a research unit of the University of California,
// San Diego.  Support for this effort was provided by NSF through the
// Protein Data Bank (Grant DBI-9814284) and the National Partnership for
// Advanced Computational Infrastructure (Grant ACI-9619020)
//
// Permission to use, copy, modify and distribute any part of OpenMMS for
// educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above
// copyright notice, this paragraph and the following paragraphs appear
// in all copies.
//
// Those desiring to incorporate this OpenMMS into commercial products or
// use for commercial purposes should contact the Technology Transfer
// Office, University of California, San Diego, 9500 Gilman Drive, La
// Jolla, CA 92093-0910, Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// In no event shall the University of California be liable to any party
// for direct, indirect, special, incidental, or consequential damages,
// including lost profits, arising out of the use of this OpenMMS, even
// if the University of California has been advised of the possibility of
// such damage.
//
// The OpenMMS provided herein is on an "as is" basis, and the
// University of California has no obligation to provide maintenance,
// support, updates, enhancements, or modifications.  The University of
// California makes no representations and extends no warranties of any
// kind, either implied or express, including, but not limited to, the
// implied warranties of merchantability or fitness for a particular
// purpose, or that the use of the OpenMMS will not infringe any patent,
// trademark or other rights.

package org.rcsb.openmms.meta;

import java.io.IOException;

import org.rcsb.openmms.cifparse.*;

/**
 * Class for representing "Structures" in Cif metamodel
 *
 * @author Douglas S. Greer
 * @version 1.4
 */
public class StructureTrans extends Trans
{
  CategoryTrans cat;
  // Variable name for references to an array of structs
  protected String listVarName;
  protected String constantFlagName;
  // Dictionary Category associated with this structure
  protected DictionaryCategory dictCategory;
  protected boolean isCifStructure = true;
  protected boolean useBlockMethods;

  // containing submodules
  ModuleTrans[] modQualifier = null;

  public int descriptionParagraphs;

  public StructureTrans(String sname, ModuleTrans[] mq, boolean isCif)
    {
      // Constructor for non-Cif structures
      this(sname, null, mq, 0, Trans.TYPE_STRUCTURE);
      setIsCif(isCif);
    }

  public StructureTrans(String sname, CategoryTrans ct,
			ModuleTrans[] mq, int descp)
    {
      this(sname, ct, mq, descp, Trans.TYPE_STRUCTURE);
    }

  public StructureTrans(String sname, CategoryTrans ct, 
			ModuleTrans[] mq, int descp, int type)
    {
      super(sname, type);
      cat = ct;
      descriptionParagraphs = descp;
      modQualifier = mq;     
      listVarName = convertToUnderbar(sname) + "_list";
      constantFlagName = "S" + convertToFlagName(sname);
    }

  public void accept(TransVisitor v)
    throws IOException, TransGenException
    {
      v.openStructure(this);
      // visit children
      Trans t;
      for (int i = 0; i < numberOfChildren(); i++)
	{
	  t = getChild(i);
	  int type = t.getTransType();
	  switch (type)
	    {
	    case Trans.TYPE_FIELD:
	    case Trans.TYPE_VECTOR:
	    case Trans.TYPE_MATRIX:
	      t.accept(v);
	      break;
	    default:
	      throw new TransGenException
		("Invalid StructureTrans Child Type: " + type );
	    }
	}
      v.closeStructure(this);
    }

  public boolean isValuetype()
    {
      return false;
    }

  /**
   * Return a name of the Java class (Type) that will be used to
   * declare instances of this Strucute in generated code.
   */
  public String getTypeName()
    {
      return getName() + getVTypeSuffix();
    }

  private String getVTypeSuffix()
    {
      if (isValuetype())
	{
	  return "Impl";
	}
      else
	{
	  return "";
	}
    }

  /**
   * Return a fully qualified (with module name prefixes) structure
   * name that can be used in calling context ctx.
   */
  public String getFullyQualifiedName(ModuleTrans ctx)
    {
      StringBuffer sb = new StringBuffer("");
      String s;

      // this assumes that the calling context does not contain
      // nested modules
      if (modQualifier != null)
	{
	  // use full module prefix string if calling context module
	  // name does not match that of the top-level module
	  // for the structure
	  int skip = 0;
	  if (ctx.getName().equals(modQualifier[0].getName()))
	    {
	      skip++;
	    }
	  if (skip < modQualifier.length)
	    {
	      for (int i = 0; i + skip < modQualifier.length; i++)
		{
		  if (i > 0)
		    {
		      sb.append('.');
		    }
		  sb.append(modQualifier[i+skip].getName());
		}
	      sb.append("::");
	    }
	}
      else
	{
	  // a null modQualifier signifies the Core module
	  if ( ! ctx.isCoreModule())
	    {
	      sb.append(RootTrans.coreModuleName + "::");
	    }
	}
      return sb.toString() + getName();
    }

  public void setIsCif(boolean isc)
    {
      isCifStructure = isc;
    }

  public boolean isCifStructure()
    {
      return isCifStructure;
    }

  public void setDictionaryCategory(DictionaryCategory di)
    {
      dictCategory = di;
    }

  public DictionaryCategory getDictionaryCategory()
    {
      return dictCategory;
    }
  
  public CategoryTrans getCategory()
    {
      return cat;
    }

  public ModuleTrans[] getModuleQualifier()
    {
      return modQualifier;
    }

  public String getListVarName()
    {
      return listVarName;
    }

  public String getFullListVarName()
    {
      String rval;

      rval = listVarName;
      if (modQualifier != null)
	{
	  rval = modQualifier[0].getVarName() + "." + rval;
	}
      return rval;
    }

  public String getFlagName()
    {
      return constantFlagName;
    }

  public void setUseBlockMethod(boolean ubm)
    {
      useBlockMethods = ubm;
    }

  public boolean useBlockMethod()
    {
      return useBlockMethods;
    }
}
