//
// TransVisitor.java,v 1.3 2001/10/27 02:02:53 dsg Exp
//
// Copyright 2001 The Regents of the University of California
// All Rights Reserved
//
// OpenMMS was developed by Dr. Douglas S. Greer at the San Diego
// Supercomputer Center, a research unit of the University of California,
// San Diego.  Support for this effort was provided by NSF through the
// Protein Data Bank (Grant DBI-9814284) and the National Partnership for
// Advanced Computational Infrastructure (Grant ACI-9619020)
//
// Permission to use, copy, modify and distribute any part of OpenMMS for
// educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above
// copyright notice, this paragraph and the following paragraphs appear
// in all copies.
//
// Those desiring to incorporate this OpenMMS into commercial products or
// use for commercial purposes should contact the Technology Transfer
// Office, University of California, San Diego, 9500 Gilman Drive, La
// Jolla, CA 92093-0910, Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// In no event shall the University of California be liable to any party
// for direct, indirect, special, incidental, or consequential damages,
// including lost profits, arising out of the use of this OpenMMS, even
// if the University of California has been advised of the possibility of
// such damage.
//
// The OpenMMS provided herein is on an "as is" basis, and the
// University of California has no obligation to provide maintenance,
// support, updates, enhancements, or modifications.  The University of
// California makes no representations and extends no warranties of any
// kind, either implied or express, including, but not limited to, the
// implied warranties of merchantability or fitness for a particular
// purpose, or that the use of the OpenMMS will not infringe any patent,
// trademark or other rights.

package org.rcsb.openmms.meta;

import java.io.*;
import java.util.*;

/**
 * Base class TransVisitor or "Transformation Visitor".
 * MMSExpr uses the "Visitor" design pattern (see Gamma et. al.) 
 * to traverse the metamodel heirarchy.  This superclass provides
 * empty (default) implementations for all the Visitor subclasses
 *
 * @author Douglas S. Greer
 * @version 1.3
 */
public class TransVisitor
{
  String oFilename;
  StringStruct primarySS;

  int indentLevel = 0;
  final static String INDENT = "    ";  // four spaces is the default indent
  protected String indent;

  public TransVisitor()
    {
      indent = INDENT;
    }

  static String[] COPYRIGHT = 
  {
    "// All Rights Reserved",
    "//",
    "// OpenMMS was developed by Dr. Douglas S. Greer at the San Diego",
    "// Supercomputer Center, a research unit of the University of California,",
    "// San Diego.  Support for this effort was provided by NSF through the",
    "// Protein Data Bank (Grant DBI-9814284) and the National Partnership for",
    "// Advanced Computational Infrastructure (Grant ACI-9619020)",
    "//",
    "// Permission to use, copy, modify and distribute any part of OpenMMS for",
    "// educational, research and non-profit purposes, without fee, and",
    "// without a written agreement is hereby granted, provided that the above",
    "// copyright notice, this paragraph and the following paragraphs appear",
    "// in all copies.",
    "//",
    "// Those desiring to incorporate this OpenMMS into commercial products or",
    "// use for commercial purposes should contact the Technology Transfer",
    "// Office, University of California, San Diego, 9500 Gilman Drive, La",
    "// Jolla, CA 92093-0910, Ph: (619) 534-5815, FAX: (619) 534-7345.",
    "//",
    "// In no event shall the University of California be liable to any party",
    "// for direct, indirect, special, incidental, or consequential damages,",
    "// including lost profits, arising out of the use of this OpenMMS, even",
    "// if the University of California has been advised of the possibility of",
    "// such damage.",
    "//",
    "// The OpenMMS provided herein is on an \"as is\" basis, and the",
    "// University of California has no obligation to provide maintenance,",
    "// support, updates, enhancements, or modifications.  The University of",
    "// California makes no representations and extends no warranties of any",
    "// kind, either implied or express, including, but not limited to, the",
    "// implied warranties of merchantability or fitness for a particular",
    "// purpose, or that the use of the OpenMMS will not infringe any patent,",
    "// trademark or other rights."
  };

  public void writeCopyright()
    {
      if (primarySS != null)
 	{
	  writeCopyright(primarySS);
	}
    }

  public void writeCopyright(StringStruct ss)
    {
      Date dt = new Date();
      GregorianCalendar cal = new GregorianCalendar();
      cal.setTime(dt);
      String firstline = "// Copyright " + cal.get(Calendar.YEAR)
	+ " The Regents of the University of California";
      ss.writeLine(firstline);
      ss.writeLine(COPYRIGHT);
    }

  public void setIndentString(String s)
    {
      indent = s;
      if (primarySS != null)
 	{
	  primarySS.setIndentString(s);
	}
    }

  public void open(String fname)
    throws TransGenException
    {
      if (primarySS != null)
 	{
	  throw new TransGenException
	    ("Transvisitor: previous StringStruct still open");
 	}
      oFilename = fname;
      primarySS = new StringStruct();
    }
  
  public void close()
    throws IOException, TransGenException
    {
      primarySS.traverseToFile(oFilename);
      primarySS = null;
    }

  public void writeStr(StringStruct oss, String[] sl, int ilev)
    throws IOException
    {
      for (int i = 0; i < sl.length; i++)
	{
	  writeStr(oss, sl[i], ilev);
	}
    }

  public void writeStr(String[] sl) throws IOException
    {
      for (int i = 0; i < sl.length; i++)
	{
	  writeStr(primarySS, sl[i], 0);
	}
    }

  public void writeStr(StringStruct oss, String s, int ilev)
    throws IOException
    {
      indentLevel += ilev;
      if (s != null && s.length() > 0)
	{
	  for (int i = 0; i < indentLevel; i++)
	    {
	      oss.writeString(indent);
	    }
	  oss.writeString(s);
	}
      oss.writeString("\n");
      indentLevel -= ilev;
    }

  public void writeBuffer(StringStruct a)
    {
      primarySS.writeStringStruct(a);
    }

  public void writeStr(String s) throws IOException
    {
      writeStr(primarySS, s, 0);
    }

  public void openRoot(RootTrans rt)
    throws IOException, TransGenException
    {
    }

  public void closeRoot(RootTrans rt)
    throws IOException, TransGenException
    {
    }

  public void openFile(FileTrans ft) 
    throws IOException, TransGenException
    {
    }

  public void closeFile(FileTrans ft) 
    throws IOException, TransGenException
    {
    }

  public void openModule(ModuleTrans mt) 
    throws IOException, TransGenException
    {
    }
      
  public void closeModule(ModuleTrans mt) 
    throws IOException, TransGenException
    {
    }

  public void openInterface(InterfaceTrans it) 
    throws IOException, TransGenException
    {
    }
      
  public void closeInterface(InterfaceTrans it) 
    throws IOException, TransGenException
    {
    }

  public void openCategory(CategoryTrans ct) 
    throws IOException, TransGenException
    {
    }
      
  public void closeCategory(CategoryTrans ct) 
    throws IOException, TransGenException
    {
    }

  public void openStructure(StructureTrans st) 
    throws IOException, TransGenException
    {
    }
      
  public void closeStructure(StructureTrans st) 
    throws IOException, TransGenException
    {
    }

  public void openValuetype(ValuetypeTrans st) 
    throws IOException, TransGenException
    {
    }
      
  public void closeValuetype(ValuetypeTrans st) 
    throws IOException, TransGenException
    {
    }

  public void openMethod(MethodTrans mt) 
    throws IOException, TransGenException
    {
    }

  public void closeMethod(MethodTrans mt) 
    throws IOException, TransGenException
    {
    }

  public void openGetListMethod(GetListMethodTrans glmt) 
    throws IOException, TransGenException
    {
    }

  public void closeGetListMethod(GetListMethodTrans glmt) 
    throws IOException, TransGenException
    {
    }

  public void openField(FieldTrans ft) 
    throws IOException, TransGenException
    {
    }

  public void closeField(FieldTrans ft) 
    throws IOException, TransGenException
    {
    }

  public void openSubfield(SubfieldTrans ft) 
    throws IOException, TransGenException
    {
    }

  public void closeSubfield(SubfieldTrans ft) 
    throws IOException, TransGenException
    {
    }

  public void openVector(VectorTrans ft) 
    throws IOException, TransGenException
    {
    }

  public void closeVector(VectorTrans ft) 
    throws IOException, TransGenException
    {
    }

  public void openMatrix(MatrixTrans ft) 
    throws IOException, TransGenException
    {
    }

  public void closeMatrix(MatrixTrans ft) 
    throws IOException, TransGenException
    {
    }

  public void createList(ListTrans lt) 
    throws IOException, TransGenException
    {
    }

  public void createTypeDef(TypeDefTrans lt) 
    throws IOException, TransGenException
    {
    }

  public void createDirective(DirectiveTrans dt) 
    throws IOException, TransGenException
    {
    }

  public void createConstant(ConstantTrans dt) 
    throws IOException, TransGenException
    {
    }

  public void createSequence(SequenceTrans lt) 
    throws IOException, TransGenException
    {
    }

  public void openException(ExceptionTrans st) 
    throws IOException, TransGenException
    {
    }
      
  public void closeException(ExceptionTrans st) 
    throws IOException, TransGenException
    {
    }

  public void createExceptionField(ExceptionFieldTrans ft) 
    throws IOException, TransGenException
    {
    }
}  

