//
// DBLoader.java,v 1.13 2003/05/09 22:32:19 dsg Exp
//
// Copyright 2001 The Regents of the University of California
// All Rights Reserved
//
// OpenMMS was developed by Dr. Douglas S. Greer at the San Diego
// Supercomputer Center, a research unit of the University of California,
// San Diego.  Support for this effort was provided by NSF through the
// Protein Data Bank (Grant DBI-9814284) and the National Partnership for
// Advanced Computational Infrastructure (Grant ACI-9619020)
//
// Permission to use, copy, modify and distribute any part of OpenMMS for
// educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above
// copyright notice, this paragraph and the following paragraphs appear
// in all copies.
//
// Those desiring to incorporate this OpenMMS into commercial products or
// use for commercial purposes should contact the Technology Transfer
// Office, University of California, San Diego, 9500 Gilman Drive, La
// Jolla, CA 92093-0910, Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// In no event shall the University of California be liable to any party
// for direct, indirect, special, incidental, or consequential damages,
// including lost profits, arising out of the use of this OpenMMS, even
// if the University of California has been advised of the possibility of
// such damage.
//
// The OpenMMS provided herein is on an "as is" basis, and the
// University of California has no obligation to provide maintenance,
// support, updates, enhancements, or modifications.  The University of
// California makes no representations and extends no warranties of any
// kind, either implied or express, including, but not limited to, the
// implied warranties of merchantability or fitness for a particular
// purpose, or that the use of the OpenMMS will not infringe any patent,
// trademark or other rights.

package org.rcsb.openmms.apps.rdb;

import java.io.*;
import java.text.*;
import java.sql.*;
import java.util.Date;
import java.util.HashSet;

// For definition of ORB functions
import org.omg.CORBA.*;
// For definition of DataAccessException
import org.omg.DsLSRMacromolecularStructure.*;

import org.rcsb.openmms.cifparse.*;
import org.rcsb.openmms.util.*;
import org.rcsb.openmms.entry.*;
import org.rcsb.openmms.loader.*;

/**
 * Database loader class that performs load and delete Entry operations.
 *
 * @author Douglas S. Greer
 * @version 1.13
 */
public class DBLoader
  implements TypeNamesSql
{
  // basic status messages
  public static int debug;
  int entriesLoaded = 0;
  int entryKey;

  ServerHandler servh;
  DataItemList dil;
  Connection conn = null;

  EntryFoundry foundry;

  public DBLoader(ServerHandler a, DataItemList dl)
    {
      servh = a;
      dil = dl;
    }

  public void setDebug(int d)
    {
      debug = d;
    }

  public Connection connect(String url, String driver,
				    String user, String passwd)
    throws SQLException, ClassNotFoundException
    {
      if (conn != null)
	{
	  closeConnection();
	}
      Class.forName(driver);
      conn = DriverManager.getConnection(url, user, passwd);
      return conn;
    }

  public void closeConnection()
    {
      try
	{
	  if (conn != null)
	    {
	      conn.close();
	      conn = null;
	    }
	}
      catch (SQLException se)
	{
	  servh.logException(se, "CLOSE JDBC CONNECTION FAILED",
		       "SQLException", "DBLoader.closeConnection()",
		       null, null, false);
	}
    }
  
  public String[] readManifest(String manifest, HashSet exclude)
      throws IOException
    {
      String[] r = FileEntryFoundry.readManifest(manifest, exclude);
      return r;
    }

  boolean Stop = false;

  synchronized public void setStop(boolean stp)
    {
      Stop = stp;
    }

  synchronized public boolean getStop()
    {
      return Stop;
    }

  /**
   * This routine is done in the swing thread on start-up to insure
   * that the database loading was not aborted in the middle of
   * an entry load.  This is needed for systems such as mySQL that
   * do not support transaction control.
   */
  public void dbCheck()
      throws MmsTranslationException
    {
      try
	{
	  InsertEntryFinal edbl = new InsertEntryFinal(servh, debug, conn);
	  edbl.dbCheck(servh);
	}
      catch (SQLException se)
	{
	  servh.logException(se, "SQL Error During Database Check",
		       "SQLException", "DBLoader.dbCheck()",
		       null, null, false);
	}
      catch (Exception e)
	{
	  e.printStackTrace();
	}
    }

  // CAUTION: This method is invoked by the *worker* thread
  public String loadEntries(String[] list, String dataPath)
    throws Error
    {
      String rval = "Load Completed";
      String scond = "Load Stopped";

      entriesLoaded = 0;
      setStop(false);

      foundry = new FileEntryFoundry(servh, dil, dataPath, null);
      InsertEntryFinal edbl = new InsertEntryFinal(servh, debug, conn);
      for (int i = 0; i < list.length; i++)
	{
	  if (getStop())
	    {
	      setStop(false);
	      rval = scond;
	      break;
	    }
	  loadSingleEntry(edbl, list[i]);
	}
      if (entriesLoaded >= 1)
	{
	  printElapsedTime(entriesLoaded);
	}
      return rval;
    }

  // CAUTION: This method is invoked by the *worker* thread
  public String loadEntries(Reader r, String dataPath, String manifest)
    throws Error
    {
      String rval = "Load Completed";
      String scond = "Load Stopped";

      entriesLoaded = 0;
      setStop(false);

      foundry = new FileEntryFoundry(servh, dil, dataPath, manifest);
      InsertEntryFinal edbl = new InsertEntryFinal(servh, debug, conn);
      StreamTokenizer st = new StreamTokenizer(r);
      st.resetSyntax();
      st.wordChars('A', 'Z');
      st.wordChars('a', 'z');
      // numbers are treated as words
      st.wordChars('0', '9');
      st.wordChars('_', '_');  // underscore allowed
      st.wordChars('.', '.');  // dot allowed
      st.whitespaceChars(',', ',');
      st.commentChar('#');     // file may have comment lines

      Runtime rt = Runtime.getRuntime();
      servh.logMessage("Total Memory=" + stringMB(rt.totalMemory())
			      + " Free Memory=" + stringMB(rt.freeMemory()));

      int ttype;
    eof_found:
      while (true)
	{
	  if (getStop())
	    {
	      setStop(false);
	      rval = scond;
	      break;
	    }
	  try
	    {
	      ttype = st.nextToken();
	    }
	  catch (IOException ie)
	    {
	      servh.logException(ie, "ENTRY LIST SYNTAX ERROR",
					"IO Exception",
					"DBLoader.loadEntries()",
					"near PDB_ID=" + st.sval,
					null, false);
	      scond = "List Syntax Error";
	      continue;
	    }
	  switch(ttype)
	    {
	    case StreamTokenizer.TT_EOF:
	      break eof_found;
	    case StreamTokenizer.TT_WORD:
	      loadSingleEntry(edbl, st.sval);
	      break;
	    }
	}
      if (entriesLoaded >= 1)
	{
	  printElapsedTime(entriesLoaded);
	}
      return rval;
    }


  private void loadSingleEntry(InsertEntryFinal edbl, String id)
    {
      try
	{
	  processSingleEntry(edbl, id);
	}
      catch (CifParseException cpe)
	{
	  abortDBOperation();
	  servh.logException(cpe, "LOAD ENTRY FAILED",
			     "CIF Parse Exception",
			     "DBLoader.loadEntries()",
			     "while loading Entry PDB_ID=" + id,
			     null, true);
	}
      catch (SQLException se)
	{
	  abortDBOperation();
	  SQLException first = se;
	  StringBuffer errList = new StringBuffer();
	  while ( se != null)
	    {
	      errList.append("  SQL Message:  " + se.getMessage());
	      errList.append("  SQL State:  " + se.getSQLState());
	      errList.append("  SQL ErrorCode: " +se.getErrorCode());
	      se = se.getNextException();
	    }
	  servh.logException(first, "LOAD ENTRY FAILED",
			     "SQL Exception",
			     "DBLoader.loadEntries()",
			     "while loading Entry PDB_ID=" + id,
			     errList.toString(), true);
	}
      catch (DataAccessException dae)
	{
	  // If target entry is not on the server just print a simple
	  // message
	  if (dae.description.indexOf("not found") >= 0)
	    {
	      String serr = "  Load Failed: " + dae.description;
	      servh.logMessage(serr);
	    }
	  else
	    {
	      abortDBOperation();
	      String daerr = "DataAccessException thrown in method "
		  + dae.method_name
		  + "\n    Description: " + dae.description;
	      servh.logException(dae, "LOAD ENTRY FAILED",
				 "Data Access Exception",
				 "DBLoader.loadEntries()",
				 "while loading Entry PDB_ID=" + id,
				 daerr, true);
	    }
	}
      catch (MmsTranslationException mte)
	{
	  abortDBOperation();
	  servh.logException(mte, "LOAD ENTRY FAILED",
			       "Mms Translation Exception",
			     "DBLoader.loadEntries()",
			     "while loading Entry PDB_ID=" + id,
			     mte.toString(), true);
	}
      catch (Exception e)
	{
	  abortDBOperation();
	  servh.logException(e, "LOAD ENTRY FAILED",
			     "(General) Exception",
			     "DBLoader.loadEntries()",
			     "while loading Entry PDB_ID=" + id,
			     e.toString(), true);
	}
      catch (Throwable t)
	{
	  // try to continue loading other entries after a
	  // OutOfMemory Exception by calling the GC.
	  if (t instanceof OutOfMemoryError)
	    {
	      runGC();
	      servh.logException(t, "LOAD ENTRY FAILED",
				 "OUT OF MEMORY ERROR ",
				 "DBLoader.loadEntries()",
				 "while loading Entry PDB_ID=" + id,
				 null, false);
	      abortDBOperation();
	    }
	  else
	    {
	      // otherwise re-throw it
	      throw (Error) t;
	    }
	}
    }

  // CAUTION: This method is invoked in the *worker* thread
  public void processSingleEntry(InsertEntryFinal edbl, String ename)
    throws DataAccessException, SQLException, MmsTranslationException,
	   CifParseException, IOException
    {

      if (getEntryKey(ename) > 0)
	{
	  servh.logMessage("Entry: " + ename 
		  + " has already been loaded into this DB");
	}
      else
	{
	  if (entriesLoaded == 0)
	    {
	      startTotalTime();
	    }
	  else if (entriesLoaded%100 == 0)
	    {
	      printElapsedTime(entriesLoaded);
	    }
	  servh.logMessage("Loading PDB_ID: " + ename);
	  if (ename.endsWith(".cif") || ename.endsWith(".CIF"))
	    {
	      ename = ename.substring(0, ename.length()-4);
	    }
	  startTimer();
	  EntryImpl entry = null;
	  try
	    {
	      entry = readSingleEntry(ename);
	      long rtime = getElapsedTime();
	      servh.logString("    Read time: " 
			      + timeToSeconds(rtime) + " s.");
	      startTimer();
	      insertSingleEntry(edbl, ename, entry);
	    }
	  finally
	    {
	      // null out for garbage-collector cleanup on exceptions
	      entry = null;
	      // Run Garbage Collector after each readEntry/insertEntry
// 	      runGC();
	    }
	  long ltime = getElapsedTime();
	  servh.logMessage("  Load time: " + timeToSeconds(ltime) 
				  + " s.  Entry_key: " + entryKey);
	  entriesLoaded++;
	}
    }

  final static String geks = "SELECT " + ENTRY_TABLE_KEY_NAME 
      + " FROM " + ENTRY_TABLE_NAME 
      + " WHERE " + ENTRY_TABLE_ID_NAME + " = '";

  // CAUTION: This method is invoked in the *worker* thread
  public int getEntryKey(String ename)
    throws SQLException
    {
      int rval = -1;
      Statement stmt = null;
      try
	{
	  stmt = conn.createStatement();
	  ResultSet rs = stmt.executeQuery(geks + ename + "'");
	  while (rs.next())
	    {
	      if (rval > 0)
		{
		  servh.logMessage("MMS SYSTEM ERROR");
		  servh.logMessage("  Entry: "
					  + ename + " in DB multiple times");
		}
	      rval = rs.getInt(1);
	    }
	}
      finally
	{
	  if (stmt != null)
	    {
	      stmt.close();
	    }
	}
      return rval;
    }

  // CAUTION: This method is invoked in the *worker* thread
  public EntryImpl readSingleEntry(String id)
    throws DataAccessException, SQLException, MmsTranslationException,
	   CifParseException, IOException
    {
      EntryImpl ei = foundry.createEntryImpl(id);
      // actualy read the data with init Entry;
      ei.setDebug(debug);
      ei.initEntry();
      return ei;
    }

  // CAUTION: This method is invoked in the *worker* thread
  public void insertSingleEntry(InsertEntryFinal edbl, String ename, 
				EntryImpl ei)
    throws SQLException, MmsTranslationException
    {
      // Writing an entry should be all or nothing
      // use transaction control to enforce commits and rollbacks
      conn.setAutoCommit(false);
      entryKey = edbl.buildAndInsertEntry(ename, ei);
      conn.commit();
      conn.setAutoCommit(true);
    }

  // CAUTION: This method is invoked in the *worker* thread
  public String deleteEntries(Reader r)
    {
      InsertEntryFinal edbl = new InsertEntryFinal(servh, debug, conn);
      StreamTokenizer st = new StreamTokenizer(r);
      String rval = "Delete Completed";
      String scond = "Delete Stopped";

      st.resetSyntax();
      st.wordChars('A', 'Z');
      st.wordChars('a', 'z');
      // numbers are treated as words
      st.wordChars('0', '9');
      st.wordChars('_', '_');  // underscore allowed
      st.wordChars('.', '.');  // dot allowed
      st.whitespaceChars(',', ',');
      st.commentChar('#');     // file may have comment lines

      int ttype;
    eof_found:
      while (true)
	{
	  if (getStop())
	    {
	      setStop(false);
	      rval = scond;
	      break;
	    }
	  try
	    {
	      ttype = st.nextToken();
	    }
	  catch (IOException ie)
	    {
	      servh.logException(ie, "ENTRY LIST SYNTAX ERROR",
					"IO Exception",
					"DBLoader.deleteEntries()",
					"near PDB_ID=" + st.sval,
					null, false);
	      scond = "List Syntax Error";
	      continue;
	    }
	  switch(ttype)
	    {
	    case StreamTokenizer.TT_EOF:
	      break eof_found;
	    case StreamTokenizer.TT_WORD:
	      try
		{
		  deleteSingleEntry(edbl, st.sval);
		}
	      catch (SQLException se)
		{
		  abortDBOperation();
		  SQLException first = se;
		  StringBuffer errList = new StringBuffer();
		  while ( se != null)
		    {
		      errList.append("  SQL State:  " + se.getSQLState());
		      errList.append("  SQL ErrorCode: " +se.getErrorCode());
		      se = se.getNextException();
		    }
		  servh.logException(first, "LOAD ENTRY FAILED",
			       "SQL Exception",
			       "DBLoader.deleteEntries()",
			       "while deleting Entry PDB_ID=" + st.sval,
			       errList.toString(), true);
		}
	      catch (DataAccessException dae)
		{
		  abortDBOperation();
		  String daerr = "DataAccessException thrown in method "
		    + dae.method_name
		      + "\n    Description: " + dae.description;
		  servh.logException(dae, "LOAD ENTRY FAILED",
			       "Data Access Exception",
			       "DBLoader.deleteEntries()",
			       "while deleting Entry PDB_ID=" + st.sval,
			       daerr, true);
		}
	      catch (MmsTranslationException mte)
		{
		  abortDBOperation();
		  servh.logException(mte, "LOAD ENTRY FAILED",
			       "Mms Translation Exception",
			       "DBLoader.deleteEntries()",
			       "while loading Entry PDB_ID=" + st.sval,
			       mte.toString(), true);
		}
	      catch (Exception e)
		{
		  abortDBOperation();
		  servh.logException(e, "LOAD ENTRY FAILED",
			       "(General) Exception",
			       "DBLoader.deleteEntries()",
			       "while deleting Entry PDB_ID=" + st.sval,
			       e.toString(), true);
		}
	      break;
	    }
	}
      return rval;
    }

  // CAUTION: This method is invoked in the *worker* thread
  public void deleteSingleEntry(InsertEntryFinal edbl, String ename)
    throws DataAccessException, SQLException, MmsTranslationException
    {
      int eKey = getEntryKey(ename);
      
      if (eKey <= 0)
	{
	  servh.logMessage("Entry: " + ename 
		  + " has not been loaded into this DB");
	}
      else
	{
	  deleteEntry(edbl, eKey, ename);
	  servh.logMessage("  Entry " + ename + " Deleted from DB");
	}
    }

  // CAUTION: This method is invoked in *both* the event (swing) thread
  // and the worker thread
  public void deleteEntry(InsertEntryFinal edbl, int key, String ename)
    throws SQLException, MmsTranslationException
    {
      // Writing an entry should be all or nothing
      // use transaction control to enforce commits and rollbacks
      if (ename.endsWith(".cif") || ename.endsWith(".CIF"))
	{
	  ename = ename.substring(0, ename.length()-4);
	}
      conn.setAutoCommit(false);
      edbl.removeEntry(key, ename);
      conn.commit();
      conn.setAutoCommit(true);
    }

  // CAUTION: This method is invoked in the *worker* thread
  protected void abortDBOperation()
    {
      try
	{
	  if ( ! conn.getAutoCommit())
	    {
	      conn.rollback();
	    }
	}
      catch (SQLException se)
	{
	  // can't win for losing ...
	  servh.logMessage("");
	  servh.logMessage
	    ("********** SQL EXCEPTION DURING ROLLBACK **********");
	  servh.logMessage("");
	  se.printStackTrace();
	}
    }

  // CAUTION: This method is invoked in the *worker* thread
  public void runGC()
    {
      // EntryImpl's contain all the data, null out entry var
      // before running the GC.
      Runtime rt = Runtime.getRuntime();
//       long wasFree = rt.freeMemory();
      rt.gc();
      rt.runFinalization();
//       long isFree = rt.freeMemory();
//       servh.logMessage("Total Memory=" + stringMB(rt.totalMemory())
// 			      + " Free Memory=" + stringMB(isFree));
    }

  long stime;
  Date totalTimeStart;

  // CAUTION: This method is invoked in the *worker* thread
  void startTimer()
    {
      stime = System.currentTimeMillis();
    }

  // CAUTION: This method is invoked in the *worker* thread
  long getElapsedTime()
    {
      long etime = System.currentTimeMillis();
      return etime - stime;
    }

  // CAUTION: This method is invoked in the *worker* thread
   String timeToSeconds(long time)
    {
      String rval = Long.toString(time / 1000) + "." 
	+ Long.toString(time % 1000);
      return rval;
    }

  // CAUTION: This method is invoked in the *worker* thread
   String timeToHours(long time)
    {
      DecimalFormat df1 = new DecimalFormat("0");
      DecimalFormat df2 = new DecimalFormat("00");
      FieldPosition fp = new FieldPosition(NumberFormat.INTEGER_FIELD);

      long h = time/(60*60*1000);
      long m = (time/(60*1000))%60;
      long s = (time/1000)%60;

      StringBuffer rval = new StringBuffer();
      df1.format(h, rval, fp);
      rval.append(":");
      df2.format(m, rval, fp);
      rval.append(":");
      df2.format(s, rval, fp);
      return rval.toString();
    }

  // CAUTION: This method is invoked in the *worker* thread
  void startTotalTime()
    {
      totalTimeStart = new Date();
      servh.logMessage("Load Started " + totalTimeStart.toString());
    }

  // CAUTION: This method is invoked in the *worker* thread
  void printElapsedTime(int n)
    {
      long c = System.currentTimeMillis() - totalTimeStart.getTime();
      servh.logMessage
	("\n" + entriesLoaded +  " entries loaded.  " 
	 + (new Date()).toString() 
	 + "\n  Total time = " + timeToHours(c)
	 +  "  Time per entry = " + timeToSeconds(c/n));
    }

  // CAUTION: This method is invoked in the *worker* thread
   String stringMB(long sz)
    {
      sz = sz/1000;
      Float f = new Float(((float) sz)/1000);
      return f.toString() + "MB";
    }
}

