//
// CFileServer.java,v 1.6 2003/04/05 01:40:05 dsg Exp
//
// Copyright 2001 The Regents of the University of California
// All Rights Reserved
//
// OpenMMS was developed by Dr. Douglas S. Greer at the San Diego
// Supercomputer Center, a research unit of the University of California,
// San Diego.  Support for this effort was provided by NSF through the
// Protein Data Bank (Grant DBI-9814284) and the National Partnership for
// Advanced Computational Infrastructure (Grant ACI-9619020)
//
// Permission to use, copy, modify and distribute any part of OpenMMS for
// educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above
// copyright notice, this paragraph and the following paragraphs appear
// in all copies.
//
// Those desiring to incorporate this OpenMMS into commercial products or
// use for commercial purposes should contact the Technology Transfer
// Office, University of California, San Diego, 9500 Gilman Drive, La
// Jolla, CA 92093-0910, Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// In no event shall the University of California be liable to any party
// for direct, indirect, special, incidental, or consequential damages,
// including lost profits, arising out of the use of this OpenMMS, even
// if the University of California has been advised of the possibility of
// such damage.
//
// The OpenMMS provided herein is on an "as is" basis, and the
// University of California has no obligation to provide maintenance,
// support, updates, enhancements, or modifications.  The University of
// California makes no representations and extends no warranties of any
// kind, either implied or express, including, but not limited to, the
// implied warranties of merchantability or fitness for a particular
// purpose, or that the use of the OpenMMS will not infringe any patent,
// trademark or other rights.

package org.rcsb.openmms.apps.refserv;

import java.io.*;
import java.util.*;
import javax.naming.*;
import org.omg.CORBA.ORB;
import org.omg.PortableServer.POAPackage .*;

import org.rcsb.openmms.cifparse.*;
import org.rcsb.openmms.util.*;
import org.rcsb.openmms.entry.*;
import org.rcsb.openmms.loader.*;

/**
 * Top-level main program for MMS Corba reference server.
 *
 * @author Douglas S. Greer
 * @version 1.6
 */
public class CFileServer
{
  // basic status debug messages
  public static int debug = 3;

  public static ORB orb;

  String frameName;
  ServFrame sframe;

  ServerHandler servh;
  DataItemList dilist;
  String dataPath;
  String manifest;

  String iorFileName = "CFileServer.ior";
  String iorFilePath;

  String logDir;
  String logFile;
  String nameServURL;
  String serviceName;
  boolean useGUI;

  public static void main(String args[])
    {
      CFileServer cfs = new CFileServer("MMS Cif File Server");
      cfs.init(args);
      // Must be last call; does not return
      cfs.eventLoop();
    }

  public CFileServer(String name)
    {
      frameName = name;
    }

  public void init(String[] args)
    {
      try
	{
	  logDir = parseArguments("logdir", args);
	  logFile = parseArguments("logfile", args);
	  useGUI = parseArgumentsBool("GUI", args);
	  dataPath = parseArguments("data", args);
	  manifest = parseArguments("manifest", args);
	  if (dataPath == null)
	    {
	      throw new CifParseException(
		  "CFileServer: Must specify data path - \"data=<path>\"");
	    }

	  nameServURL = parseArguments("nameserv", args);
	  if (nameServURL == null)
	    {
	      nameServURL = "iiopname://corba.sdsc.edu:5001";
	    }
	  serviceName = parseArguments("servicename", args);
	  if (serviceName == null)
	    {
	      serviceName = "OpenMMS_Server";
	    }

	  String debugStr = parseArguments("debug", args);
	  if (debugStr != null)
	    {
	      // debug <= 0 report only fatal errors
	      // debug >= 1 report recoverable error messages
	      // debug >= 3 basic status messages
	      // debug >= 5 trace
	      // debug >= 7 dump various data structures
	      debug = Integer.parseInt(debugStr);
	    }

	  // Bind DataItems to CatLoaders
	  LinkTable ltable = new LinkTable();
	  LoaderLink[] links = ltable.createLinkTable();
	  dilist = new DataItemList();
	  buildDataItemList(dilist, links);

	  initOrb(args);
 	  servh = new ServerHandler(debug, orb);

	  // No messages can be logged until GUI is created.
	  initGUI(servh);

	  servh.logMessage("Reference MMS Corba Server Started at "
			  + (new Date()).toString());
	  EntryCache ec = new BasicRAMEntryCache();
	  ec.setDebug(debug);
	  servh.setEntryCache(ec);

	  // Where to find find files on cache misses
	  FileEntryFoundry efed 
	    = new FileEntryFoundry(servh, dilist, dataPath, manifest);
	  efed.setDebug(debug);
	  servh.addEntryFoundry(efed);
	  servh.setEntryDirectory(efed);
	}
      catch(CifParseException e)
	{
	  servh.fatalError
	    ("Parse Exception during CFileServer.init()", e);
	}
      // Initialize Corba Server
      startCorbaServer(args);
      String msg = "Cif Corba Server Ready";
      servh.logMessage(msg);
      System.out.println(msg);
    }

  private void buildDataItemList(DataItemList dilist, LoaderLink[] links)
    throws CifParseException
    {
      DataItem di;
      String itemName;

      for (int i = 0; i < links.length; i++)
	{
	  itemName = "_" + links[i].getCatName() + 
	    "." + links[i].getItemName();
	  di = new DataItem(itemName);
	  di.setFieldCode(links[i].getFieldCode());
	  di.setCatLoader(links[i].getCatLoader());
	  dilist.add(di);
	}
    }

  public void initGUI(ServerHandler ah)
    {
      if (useGUI)
	{
	  sframe = new ServFrame(frameName, ah, logDir, logFile);
	  sframe.pack();
	  sframe.setVisible(true);
	}
      else
	{
	  ServPanel sp = new ServPanel(false, logDir, logFile);
	  ah.setMessageHandler(sp);
	}
    }
  
  public void initOrb(String args[])
    {
      // create and initialize Default ORB
      orb = ORB.init(args, null);
    }

  public void startCorbaServer(String args[])
    {
      CFileServerImpl si = null;
      org.omg.CORBA.Object ref = null;
      try
	{
	  // create servant and register it with the ORB
	  si = new CFileServerImpl(servh, debug);
	  ref = servh.getRootPoa().servant_to_reference(si);
	}
      catch (Exception e)
	{
	  servh.fatalError("during CFileServer.startCorbaServer()", e);
	}
      try
	{
	  bindNameServer(ref, nameServURL, serviceName);
	}
      catch (NamingException ne)
	{
	  iorFilePath = logDir + "/" + iorFileName;
	  String msg = "Bind to Name Server: " + nameServURL + " FAILED"
	    + "\nWriting IOR to file: " + iorFilePath;
	  System.err.println(msg);
	  servh.logMessage(msg);
	  try
	    {
	      writeIorFile(ref);
	    }
	  catch (Exception e)
	    {
	      servh.fatalError
		("Unable to write IOR File in startCorbaServer()", e);
	    }
	}
    }
  
  private void bindNameServer(org.omg.CORBA.Object ref,
			      String nsurl, String sname)
    throws NamingException
    {
	  Hashtable nameEnv = new Hashtable(11);
	  
     	  nameEnv.put("java.naming.provider.url", nsurl);
     	  nameEnv.put("java.naming.factory.initial",
     		  "com.sun.jndi.cosnaming.CNCtxFactory");
	  Context ctx = new InitialContext(nameEnv);
	  ctx.rebind(sname, ref);
	  String msg = "Bind to JNDI URL:" + nsurl
			  + "\n\twith name " + sname + " Successful";
	  servh.logMessage(msg);
	  System.out.println(msg);
    }

  private void writeIorFile(org.omg.CORBA.Object ref)
    throws IOException
    {
      // Remove the old ior file and create a new one.
      File f = new File(iorFilePath);
      if(f.exists())
	{
	  f.delete();
	}
      
      FileWriter fout = new FileWriter(f);
      
      // create the IOR string
      String ior = orb.object_to_string(ref);
      servh.logMessage(ior);
      // write the IOR string out
      fout.write(ior);
      fout.close();
    }

  public String parseArguments(String param, String[] argv)
    {
      for(int i=0; i<argv.length; i++)
	{
	  if(argv[i].startsWith(param))
	    {
	      int index=argv[i].indexOf('=');
	      if(index!=-1) return (argv[i].substring(index+1));
	    }
	}
      return null;
    }

  public boolean parseArgumentsBool(String param, String[] argv)
    {
      for(int i=0; i<argv.length; i++)
	if(argv[i].equalsIgnoreCase(param)) return true;

      return false;
    }

  public void eventLoop()
    {
      try
	{
	  // wait for invocations from clients
	  java.lang.Object sync = new java.lang.Object();
	  synchronized (sync) {
	    sync.wait();
	  }
	}
      catch (Exception e)
	{
	  servh.fatalError("Event Loop Error", e);
	}
    }
}
