//
// CFileServerImpl.java,v 1.4 2002/06/19 01:55:34 dsg Exp
//
// Copyright 2001 The Regents of the University of California
// All Rights Reserved
//
// OpenMMS was developed by Dr. Douglas S. Greer at the San Diego
// Supercomputer Center, a research unit of the University of California,
// San Diego.  Support for this effort was provided by NSF through the
// Protein Data Bank (Grant DBI-9814284) and the National Partnership for
// Advanced Computational Infrastructure (Grant ACI-9619020)
//
// Permission to use, copy, modify and distribute any part of OpenMMS for
// educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above
// copyright notice, this paragraph and the following paragraphs appear
// in all copies.
//
// Those desiring to incorporate this OpenMMS into commercial products or
// use for commercial purposes should contact the Technology Transfer
// Office, University of California, San Diego, 9500 Gilman Drive, La
// Jolla, CA 92093-0910, Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// In no event shall the University of California be liable to any party
// for direct, indirect, special, incidental, or consequential damages,
// including lost profits, arising out of the use of this OpenMMS, even
// if the University of California has been advised of the possibility of
// such damage.
//
// The OpenMMS provided herein is on an "as is" basis, and the
// University of California has no obligation to provide maintenance,
// support, updates, enhancements, or modifications.  The University of
// California makes no representations and extends no warranties of any
// kind, either implied or express, including, but not limited to, the
// implied warranties of merchantability or fitness for a particular
// purpose, or that the use of the OpenMMS will not infringe any patent,
// trademark or other rights.

package org.rcsb.openmms.apps.refserv;

import java.io.*;
import java.util.*;

import org.rcsb.openmms.cifparse.*;
import org.rcsb.openmms.entry.*;

import org.omg.CORBA.*;
import org.omg.PortableServer.*;
import org.omg.PortableServer.POAPackage .*;
import org.omg.DsLSRMacromolecularStructure.*;
import org.omg.ccm.BaseIDL.*;

/**
 * Implements the EntryFactory for the MMS Corba reference server.
 *
 * @author Douglas S. Greer
 * @version 1.4
 */
public class CFileServerImpl extends EntryFactoryPOA
{
  private static int debug = 0;
  private static boolean trace = false;

  ServerHandler servh;

  boolean inited = false;
  // Entry Group data, Entries and ModificationDate structures
  String[] entries;  // all entries
  String[] entryGroups; // all entry groups
  ModificationDate[] modDates;  // all entries with mod dates

  public CFileServerImpl(ServerHandler a, int dbug)
    throws DataAccessException
    {
      servh = a;
      debug = dbug;
      if (debug >= 5)
	{
	  trace = true;
	}
      initEntryLists();
    }

  public Entry get_entry_from_id(String id)
    throws DataAccessException
    {
      if (trace)
	{
	  servh.logMessage
	    ("CFileServerImpl.get_entry_from_id(): id=" + id);
	}

      // get object reference from the servant
      EntryImpl ei = servh.getEntryFromId(id);

      org.omg.CORBA.Object ref = null;
      try
	{
	  ref = servh.getRootPoa().servant_to_reference(ei);
	}
      catch(org.omg.CORBA.UserException cue)
	{
	  servh.fatalError
	    ("CFileServerImpl - could not convert to reference", cue);
	}
      Entry e = EntryHelper.narrow(ref);

      if (e == null)
 	{
 	  String err = "Non-existant Entry Id: "
 	    + id + "\n";
 	    if (debug >= 2) 
 	      {
 		servh.logMessage(err);
 	      }
 	    // throw a Corba DataAccess exception
 	    throw new DataAccessException("getEntry", err);
 	}
      return e;
    }

  protected void checkInit()
    throws DataAccessException
    {
      if ( ! inited)
	{
	  initEntryLists();
	  inited = true;
	}
    }

  private void initEntryLists()
    throws DataAccessException
    {
      EntryDirectory ed =  servh.getEntryDirectory();

      entries = ed.getEntryIdList();
      entryGroups = ed.getEntryGroupList();
      modDates = ed.getEntryModificationDates();
    }

  public String get_version()
    {
      if (trace)
	{
	  servh.logMessage
	    ("CFileServerImpl.get_version() Called");
	}
      return EntryFactory.MODEL_IDENTIFIER +"_"+ EntryFactory.REVISION_NUMBER;
    }

  public String[] get_entry_id_list()
    throws DataAccessException
    {
      if (trace)
	{
	  servh.logMessage
	    ("CFileServerImpl.get_entry_id_list() Called");
	}
      checkInit();
      return entries;
    }

  public int get_entry_id_list_size()
    throws DataAccessException
    {
      if (trace)
	{
	  servh.logMessage
	    ("CFileServerImpl.get_entry_id_list_size() Called");
	}
      checkInit();
      return entries.length;
    }

  public String[] get_entry_id_list_block_n(int from, int to)
    throws DataAccessException
    {
      if (trace)
	{
	  servh.logMessage
	    ("CFileServerImpl.get_entry_id_list_block() Called");
	}
      checkInit();
      String[] t = new String[to-from+1];
      System.arraycopy(entries, from, t, 0, to-from+1);
      return t;
    }

  public ModificationDate[] get_entry_modification_dates()
    {
      if (trace)
	{
	  servh.logMessage
	    ("CFileServerImpl.get_entry_modification_dates() Called");
	}
      return modDates;
    }

  public ModificationDate[] get_entry_modification_dates_block_n
      (int from, int to)
    throws DataAccessException
    {
      if (trace)
	{
	  servh.logMessage
	    ("CFileServerImpl.get_entry_modification_dates() Called");
	}
      checkInit();
      ModificationDate[] t = new ModificationDate[to-from+1];
      System.arraycopy(modDates, from, t, 0, to-from+1);
      return t;
    }

  public String[] get_entry_group_list()
    throws DataAccessException
    {
      if (trace)
	{
	  servh.logMessage
	    ("CFileServerImpl.get_entry_group_list() Called");
	}
      checkInit();
      return entryGroups;
    }

  public String[] get_entries_in_group(String group)
    throws DataAccessException
    {
      if (trace)
	{
	  servh.logMessage
	    ("CFileServerImpl.get_entries_in_group() Called");
	}
      checkInit();
      String[] rval =  servh.getEntryDirectory().getEntriesInGroup(group);
      return rval;
    }

  public String[] native_formats_supported()
    throws DataAccessException
    {
      if (trace)
	{
	  servh.logMessage
	    ("CFileServerImpl.native_formats_supported() Called");
	}
      String[] r =  servh.getEntryDirectory().getNativeFormatsSupported();
      return r;
    }

  public ModuleDef[] get_extension_modules()
    {
      return null;
    }

  public byte[] get_native_entry_representation(String format,
						 String entry_id)
    throws DataAccessException
    {
      if (trace)
	{
	  servh.logMessage
	    ("CFileServerImpl.get_native_format_representation()"
	     + "\n\tformat=" + format + " entry_id=" + entry_id);
	}
      byte[] buf = servh.getEntryDirectory()
	.getNativeEntryRepresentation(format, entry_id);
      return buf;
    }
}


