//
// CifStats.java,v 1.5 2002/07/30 00:31:06 dsg Exp
//
// Copyright 2001 The Regents of the University of California
// All Rights Reserved
//
// OpenMMS was developed by Dr. Douglas S. Greer at the San Diego
// Supercomputer Center, a research unit of the University of California,
// San Diego.  Support for this effort was provided by NSF through the
// Protein Data Bank (Grant DBI-9814284) and the National Partnership for
// Advanced Computational Infrastructure (Grant ACI-9619020)
//
// Permission to use, copy, modify and distribute any part of OpenMMS for
// educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above
// copyright notice, this paragraph and the following paragraphs appear
// in all copies.
//
// Those desiring to incorporate this OpenMMS into commercial products or
// use for commercial purposes should contact the Technology Transfer
// Office, University of California, San Diego, 9500 Gilman Drive, La
// Jolla, CA 92093-0910, Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// In no event shall the University of California be liable to any party
// for direct, indirect, special, incidental, or consequential damages,
// including lost profits, arising out of the use of this OpenMMS, even
// if the University of California has been advised of the possibility of
// such damage.
//
// The OpenMMS provided herein is on an "as is" basis, and the
// University of California has no obligation to provide maintenance,
// support, updates, enhancements, or modifications.  The University of
// California makes no representations and extends no warranties of any
// kind, either implied or express, including, but not limited to, the
// implied warranties of merchantability or fitness for a particular
// purpose, or that the use of the OpenMMS will not infringe any patent,
// trademark or other rights.

package org.rcsb.openmms.apps.stats;

import java.awt.*;
import java.io.*;
import java.net.*;
import java.util.*;

import org.rcsb.openmms.cifparse.*;

/**
 * Contains main method for mmCIF statistics gathering program.
 *
 * @author Douglas S. Greer
 * @version 1.5
 */
public class CifStats
{
  // basic status debug messages
  public static int debug = 1;

  String dataPath;
  ArrayList dct;
  StatBuilder buildr;
  boolean readTime;
  Date beginTime;

  // can start statistic gathering part way through the list
  boolean started;
  String startingEntry;

  String oFileName = "Stats.txt";
  PrintWriter logFileWriter;

  CifDictionary cifDic = new CifDictionary();

  public static void main(String[] argv)
    {
      CifStats cifStats = new CifStats();
      cifStats.execute(argv);
    }

  public void execute(String[] argv)
    {
      CifParser cparser = new CifParser();
      CifTokenizer ct = null;

      try
	{
	  initParams(argv);
	  logFileWriter = new PrintWriter(new FileWriter(oFileName));

	  for(int i = 0; i < dct.size(); i++)
	    {
	      String s = (String) dct.get(i);
	      System.out.println("Reading dictionary from " + s);
	      try
		{
		  ct = new CifTokenizer(s);
		}
	      catch(IOException e)
		{
		  System.out.println("Error in opening dictionary: " + s);
		  System.exit(1);
		}
	      cparser.readDictionary(ct, cifDic, true);
	    }

	  DataItemList diList = cifDic.getDictionaryItemList();
	  int len = diList.size();
	  DataItem di;
	  for (int i = 0; i < len; i++)
	    {
	      di = (DataItem) diList.elementAt(i);
	      di.setItemDefaultValue(null);
	    }
	  // Builder is used in reading data files, not dictionary
	  buildr = new StatBuilder(cparser, diList);
	  buildr.setDebug(debug);
	  cparser.setBuilder(buildr);

	  // run through data files and gather statistics
	  beginTime = new Date();
	  logMessage("\nItem statistics for mmCIF files.  Date: "
		   + beginTime.toString());

	  processEntryList(dataPath);
	  if (entriesRead <= 0)
	    {
	      String msg = "\nERROR: Finished with No Entries Read\n";
	      System.err.println(msg);
	      logMessage(msg);
	      System.exit(1);
	    }

	  writeStats(cifDic, entriesRead,
	     "Statistics for complete list of " + entriesRead + " Entries");
	}
      catch(IOException e)
	{
	  System.out.println("IOException: " + e.getMessage());
	  e.printStackTrace();
	  System.exit(1);
	}
      catch(CifParseException e)
	{
	  if (ct != null)
	    {
	      System.out.println("Fatal Error parsing Cif file: " 
		 + ct.filename() + " at line number "
		 + ct.getLineNumber());
	    }
	  else
	    {
	      System.out.println("Fatal Error");
	    }
	  e.printStackTrace();
	  System.exit(1);
	}
    }

  public void initParams(String[] argv)
    throws CifParseException
    {
      String debugStr = parseArguments("debug", argv);
      if(debugStr != null)
	{
	  // debug <= 0 no messages
	  // debug == 1 basic status messages
	  // debug >= 5 trace
	  debug = Integer.parseInt(debugStr);
	}
      dataPath = parseArguments("data", argv);
      startingEntry = parseArguments("start", argv);
      String t = parseArguments("output", argv);
      readTime = parseArgumentsBool("Time", argv);
      if (t != null)
	{
	  oFileName = t;
	}
      dct = parseArgumentsMult("dct", argv);
      if(dct.size() == 0)
	{
	  throw new CifParseException
	    ("CifStats: Insufficient input " +
	     "- must specify dictionary (dct=<>)");
	}
    }

  private void processEntryList(String dDir)
    {
      // if no startingEntry given process everything
      started = startingEntry == null;

      // read groups and entries
      try
	{
	  File dFile = new File(dDir);
	  if ( ! dFile.isDirectory())
	    {
	      throw new Exception("-data argument is not a directory");
	    }
	  String[] level1 = dFile.list();
	  for (int i = 0; i < level1.length; i++)
	    {
	      String gdir = dDir + "/" + level1[i];
	      File gfile = new File(gdir);
	      if ( ! gfile.isDirectory())
		{
		  processEntry(dDir, level1[i]);
		}
	      else
		{

		  // traverse subdirectory
		  String[] level2 = gfile.list();
		  for (int j = 0; j < level2.length; j++)
		    {
		      String kname = gdir + "/" + level2[j];
		      if (kname.endsWith(".cif"))
			{
			  File kfile = new File(kname);
			  if (kfile.isFile())
			    {
			      processEntry(gdir, level2[j]);
			    }
			} 
		    }
		}
	    }
	}
      catch(IOException ioe)
	{
	  logException(ioe, "ERROR READING DATA DIRECTORY",
				    "I/O Exception",
				    "processEntryList()",
				    "while reading directory " + dDir,
				    null, true);
	}
      catch (Exception e)
	{
	  logException(e, "ERROR READING DATA DIRECTORY",
				    "(General) Exception",
				    "processEntryList()",
				    "while reading directory " + dDir,
				    null, true);
	}
    }

  int entriesRead = 0;

  public void processEntry(String dir, String name)
    {
      String id;

      if (startingEntry != null)
	{
	  if (name.startsWith(startingEntry))
	    {
	      started = true;
	    }
	}
      if ( ! started)
	{
	  return;
	}
      if (name.endsWith(".cif"))
	{
	  // have to open each file to find the 
	  // modification date
	  id = name.substring(0, name.length() - 4);
	}
      else
	{
	  // File must have ".cif" suffix
	  return;
	}
      CifTokenizer ct = null;
      try
	{
	  if (readTime)
	    {
	      startTimer();
	    }
	  ct = new CifTokenizer(dir + "/" + name);
	  buildr.readDataBlock(id, ct);
	  if (readTime)
	    {
	      long ltime = getElapsedTime();
	      logMessage("Entry: " + id + " Parse time: "
			 + timeToSeconds(ltime));
	    }
	}
      catch(IOException ioe)
	{
	  logException(ioe, "PARSE ENTRY FAILED",
		       "I/O Exception",
		       "processEntry()",
		       "while reading Entry PDB_ID=" + id,
		       null, true);
	}
      catch (CifParseException cpe)
	{
	  logException(cpe, "PARSE ENTRY FAILED",
		       "mmCIF Parse Exception",
		       "processEntry()",
		       "while reading Entry PDB_ID=" + id,
		       "At line number: " + ct.getLineNumber(),
		       true);
	}
      catch (Exception e)
	{
	  logException(e, "PARSE ENTRY FAILED",
		       "(General) Exception",
		       "processEntry()",
		       "while reading Entry PDB_ID=" + id,
		       null, true);
	}
      finally
	{
	  if (ct != null)
	    {
	      try
		{
		  ct.close();
		}
	      catch(IOException ioe)
		{
		  // ignore it
		}
	    }
	  ct = null;
	}
      entriesRead++;

      if (entriesRead != 0 &&
	  ( entriesRead <= 100 & (entriesRead % 25 == 0)) ||
	  ( entriesRead <= 1000 & (entriesRead % 250 == 0)) ||
	  ( entriesRead <= 10000 & (entriesRead % 1000 == 0)))
	{
	  writeStats(cifDic, entriesRead, "Statistics after reading " 
		     + entriesRead + " entries (upto " + id + ")");
	}
    }

  private void writeStats(CifDictionary cd, int total, String msg)
    {
      DictionaryCategory dc;
      Date dt = new Date();
      int rfs;
      String lstr;
      float prs;

      long time = dt.getTime() - beginTime.getTime();

      logMessage("\n\n" +  msg);
      logMessage("  Total Entries: " + total
	       + " Total Time (H:M:S) : " + timeString(time)
	       + " Time per entry: " 
	       + ((float)(((double)time)*0.001 / total)) + " seconds");

      if (false)
	{
	  DictionaryCategoryList dcl = cd.getDictionaryCategoryList();
	  Collection cm = dcl.getMap().values();
	  logMessage("Number of Categories = " +  cm.size());
	  Iterator ci = cm.iterator();
	  while (ci.hasNext())
	    {
	      dc = (DictionaryCategory) ci.next();
	      String mand = "(Mand code undefined)";
	      switch (dc.getCategoryMandatoryCode())
		{
		case DictionaryCategory.CATEGORY_MANDATORY_CODE_YES:
		  mand = "Mandatory";
		  break;
		case DictionaryCategory.CATEGORY_MANDATORY_CODE_NO:
		  mand = "Optional";
		  break;
		}
	      logMessage(mand + " Category " +  dc.getCategoryName());
	    }
	}

      DictionaryItemList dil = cd.getDictionaryItemList();
      int len = dil.size();
      DictionaryItem di;

      // logMessage("Number of Items = " +  len);
      for (int i = 0; i < len; i++)
	{
	  di = (DictionaryItem) dil.elementAt(i);
	  String mand = "(Mand code undefined)";
	  switch (di.getItemMandatoryCode())
	    {
	    case DictionaryItem.ITEM_MANDATORY_CODE_YES:
	      mand = "Mandatory";
	      break;
	    case DictionaryItem.ITEM_MANDATORY_CODE_NO:
	      mand = "Optional";
	      break;
	    }
	  rfs = di.getFieldCode();
	  lstr =  di.getItemDefaultValue();
	  if (lstr == null)
	    {
	      lstr =  "";
	    }
	  else
	    {
	      lstr =  " last= " + di.getItemDefaultValue() + ".cif";
	    }
	  prs = (float)((float)(rfs*10000/total)/100.0) ;
	  logMessage("    " + mand + " Item " +  di.getItemName()
		   + "\n\tRefs " + rfs 
		   + " (" + prs + " %) " + lstr);
	}
    }

  public String timeString(long t)
    {
      long hrs = t / (60*60*1000);
      long mins = (t / (60*1000)) % 60;
      long secs = (t / 1000) % 60;
      return Long.toString(hrs) + ":" 
	+ Long.toString(mins) + ":" + Long.toString(secs);
    }

  public void logMessage(String s)
    {
      logString(s + "\n");
    }

  public void logString(String s)
    {
	logFileWriter.write(s);
	logFileWriter.flush();
    }

  public void logException(Throwable t, String title,
			   String exceptionType, String where,
			   String cause, String other,
			   boolean doStackTrace)
    {
      StringBuffer m = new StringBuffer();
      m.append("\n\n********** " + title + " **********");
      m.append("\n  " + exceptionType + " in " + where);
      if (cause != null)
	{
	  m.append("\n  " + cause);
	}
      m.append("\n  Exception Description: " + t.toString());
      if (other != null)
	{
	  m.append("\n  " + other);
	}
      if (doStackTrace)
	{
	  m.append("\n  Stack Trace\n");
	  StringWriter sw = new StringWriter();
	  t.printStackTrace(new PrintWriter(sw));
	  m.append(sw.toString());
	}
      m.append("\n********** end of exception report **********\n");
      logString(m.toString());
    }

  long stime;
  void startTimer()
    {
      stime = System.currentTimeMillis();
    }

  long getElapsedTime()
    {
      long etime = System.currentTimeMillis();
      return etime - stime;
    }

   String timeToSeconds(long time)
    {
      String rval = Long.toString(time / 1000) + "." 
	+ Long.toString(time % 1000);
      return rval;
    }

  public String parseArguments(String param, String[] argv)
    {
      for(int i = 0; i < argv.length; i++)
	{
	  if(argv[i].startsWith(param))
	    {
	      int index = argv[i].indexOf('=');
	      if (index != -1)
		{
		  return (argv[i].substring(index+1));
		}
	    }
	}
      return null;
    }

  public boolean parseArgumentsBool(String param, String[] argv)
    {
      for(int i = 0; i < argv.length; i++)
	if(argv[i].compareTo(param) == 0) return true;

      return false;
    }

  public ArrayList parseArgumentsMult(String param, String[] argv)
    {
      ArrayList values = new ArrayList();
      for(int i = 0; i < argv.length; i++)
	{
	  if(argv[i].startsWith(param))
	    {
	      int index = argv[i].indexOf('=');
	      if(index != -1) 
		{
		  values.add(argv[i].substring(index+1));
		}
	    }
	}
      return values;
    }
}	
