//
// ConvertBuilder.java,v 1.1 2003/07/29 23:56:23 dsg Exp
//
// Copyright 2001 The Regents of the University of California
// All Rights Reserved
//
// OpenMMS was developed by Dr. Douglas S. Greer at the San Diego
// Supercomputer Center, a research unit of the University of California,
// San Diego.  Support for this effort was provided by NSF through the
// Protein Data Bank (Grant DBI-9814284) and the National Partnership for
// Advanced Computational Infrastructure (Grant ACI-9619020)
//
// Permission to use, copy, modify and distribute any part of OpenMMS for
// educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above
// copyright notice, this paragraph and the following paragraphs appear
// in all copies.
//
// Those desiring to incorporate this OpenMMS into commercial products or
// use for commercial purposes should contact the Technology Transfer
// Office, University of California, San Diego, 9500 Gilman Drive, La
// Jolla, CA 92093-0910, Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// In no event shall the University of California be liable to any party
// for direct, indirect, special, incidental, or consequential damages,
// including lost profits, arising out of the use of this OpenMMS, even
// if the University of California has been advised of the possibility of
// such damage.
//
// The OpenMMS provided herein is on an "as is" basis, and the
// University of California has no obligation to provide maintenance,
// support, updates, enhancements, or modifications.  The University of
// California makes no representations and extends no warranties of any
// kind, either implied or express, including, but not limited to, the
// implied warranties of merchantability or fitness for a particular
// purpose, or that the use of the OpenMMS will not infringe any patent,
// trademark or other rights.

package org.rcsb.openmms.apps.xconv;

import java.util.*;
import java.io.*;
import org.rcsb.openmms.cifparse.*;

/**
 * Builder subclass for reading the mmCIF datafiles and
 * translating it to XML format.
 *
 * @see org.rcsb.openmms.cifparse.Builder
 *
 * @author Douglas S. Greer
 * @version 1.1
 */
public class ConvertBuilder implements Builder
{
  private static boolean trace = false;
  private static int debug = 0;

  private XmlWriter xmlWriter;
  private DictionaryItem dictI;

  // This string is used to control the looping on tables and rows
  private String categoryName = null;
  private DictionaryCategory dictionaryCat;

  // Categories may contain multiple rows. This var true if we
  // have already written the open category tag
  private boolean aCategoryIsOpen = false;

  private Record record;
  private DictionaryCategoryList dictCatList;

  public ConvertBuilder(CifDictionary cifDic)
    {
      record = new Record(cifDic);
      dictCatList = cifDic.getDictionaryCategoryList();
    }

  public void setWriter(XmlWriter xwriter)
    {
      xmlWriter = xwriter;
    }
  
  public void setDebug(int d)
    {
      debug = d;
      if (debug >= 5)
	{
	  trace = true;
	}
    }

  //
  // Builder Methods called by CifParser
  //
  public void beginCompound()
    throws CifParseException
    {
      if (trace)
	{
	  System.out.println("XBuilder.beginCompound()");
	}

      try
	{
	  xmlWriter.openEntry();
	}
      catch (IOException ioe)
	{
	  outputError(ioe);
	}
      categoryName = null;
      record.clear();
    }

  public void endCompound()
    throws CifParseException
    {
      try
	{
	  checkEndCategory();
	}
      catch (IOException ioe)
	{
	  outputError(ioe);
	}
      if (trace)
	{
	  System.out.println("XBuilder.endCompound()");
	}

      try
	{
	  xmlWriter.closeEntry();
	}
      catch (IOException ioe)
	{
	  outputError(ioe);
	}
    }

  public void setSingleItem(DataItem di)
    throws CifParseException
    {
      String name = null;
      if (di != null)
	{
	  name = di.getItemName();
	}
      checkCategory(name);
      if (trace)
	{
	  System.out.println("XBuilder.setSingleValue(): " + name);
	}
    }

  public void insertSingleValue(DataItem di, String val) 
    throws CifParseException
    {
      if (trace)
	{
	  System.out.println
	    ("XBuilder.insertSingleValue(): "
	     + di.getItemName() + ":" + val);
	}
      setValue(di, val);
    }

  public void beginLoop()
    throws CifParseException
    {
      if (trace)
	{
	  System.out.println("XBuilder.beginLoop()");
	}
      // loop may be preceeded by single values -- need to write
      // out any existing data and close the category
      try
	{
	  endRecord();
	  checkEndCategory();
	}
      catch (IOException ioe)
	{
	  outputError(ioe);
	}
    }

  public void endLoop()
    throws CifParseException
    {
      if (trace)
	{
	  System.out.println("XBuilder.endLoop()");
	}
      try
	{
	  checkEndCategory();
	}
      catch (IOException ioe)
	{
	  outputError(ioe);
	}
      newCategory(null);
    }

  public void beginRow()
    throws CifParseException
    {
      if (trace)
	{
	  System.out.println("XBuilder.beginRow()");
	}
    }

  public void endRow()
    throws CifParseException
    {
      if (trace)
	{
	  System.out.println("XBuilder.endRow()");
	}
      try
	{
	  endRecord();
	}
      catch (IOException ioe)
	{
	  outputError(ioe);
	}
    }

  public void setLoopItem(int column, DataItem di)
    throws CifParseException
    {
      String iname = "(Null DataItem)";
      if (di != null)
	{
	  iname = di.getItemName();
	}
      if (trace)
	{
	  System.out.println("XBuilder.setLoopItem(): " + iname);
	}
      // If this is the first item in the loop start up a new 
      // category.
      if (categoryName == null)
	{
	  newCategory(XmlWriter.parseCategoryName(iname));
	}
    }

  public void insertLoopValue(int column, DataItem d, String val)
    throws CifParseException
    {
      if (trace)
	{
	  System.out.println
	    ("XBuilder.insertLoopValue(): " + d.getItemName() + ":" + val);
	}
      setValue(d, val);
    }

  private void setValue(DataItem d, String val)
    throws CifParseException
    {
      dictI = (DictionaryItem) d;
      if (dictI != null)
	{
	  record.add(dictI, val);
	}
    }
  
  // called by setSingleValue()
  private void checkCategory(String name)
    {
      String catName = XmlWriter.parseCategoryName(name);
      try
	{
	  if (categoryName == null)
	    {
	      newCategory(catName);
	    }
	  else
	    {
	      // we're already in a category is it a new one?
	      if (! categoryName.equals(catName))
		{
		  // new category --> write old data if needed
		  endRecord();
		  checkEndCategory();
		  newCategory(catName);
		}
	    }
	}
      catch (IOException ioe)
	{
	  outputError(ioe);
	}
    }
  
  private void newCategory(String name)
    {
      // Clear any left over values
      record.clear();
      categoryName = name;
      if (name != null)
	{
	  dictionaryCat = dictCatList.findDictionaryCategory
	      (name.toUpperCase());
	  if (dictionaryCat == null)
	    {
	      System.err.println("Cannot find dictionary Category = " + name);
	    }
	}
      else
	{
	  dictionaryCat = null;
	}
    }

  private void checkEndCategory()
      throws IOException
    {
      if (aCategoryIsOpen)
	{
	  xmlWriter.finishCategory(categoryName);
	  aCategoryIsOpen = false;
	}
      // If there was no partial record we will end the
      // category without writing anything
      // We are now done with this category.
      newCategory(null);
    }

  private void endRecord()
      throws IOException
    {
      if (categoryName != null && record.haveData(categoryName, dictionaryCat))
	{
	  if ( ! aCategoryIsOpen)
	    {
	      xmlWriter.startCategory(categoryName);
	      aCategoryIsOpen = true;
	    }
	  xmlWriter.writeRecord(record, categoryName, dictionaryCat);
	}
      record.clear();
    }

  private void outputError(IOException ioe)
    {
      try
	{
	  System.err.println("ERROR Writing output file"
			     + xmlWriter.getEntryName());
	  System.err.println(ioe.toString());
	  ioe.printStackTrace();
	}
      catch (IOException e)
	{}
    }
}
