//
// XmlConvert.java,v 1.1 2003/07/29 23:56:25 dsg Exp
//
// Copyright 2001 The Regents of the University of California
// All Rights Reserved
//
// OpenMMS was developed by Dr. Douglas S. Greer at the San Diego
// Supercomputer Center, a research unit of the University of California,
// San Diego.  Support for this effort was provided by NSF through the
// Protein Data Bank (Grant DBI-9814284) and the National Partnership for
// Advanced Computational Infrastructure (Grant ACI-9619020)
//
// Permission to use, copy, modify and distribute any part of OpenMMS for
// educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above
// copyright notice, this paragraph and the following paragraphs appear
// in all copies.
//
// Those desiring to incorporate this OpenMMS into commercial products or
// use for commercial purposes should contact the Technology Transfer
// Office, University of California, San Diego, 9500 Gilman Drive, La
// Jolla, CA 92093-0910, Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// In no event shall the University of California be liable to any party
// for direct, indirect, special, incidental, or consequential damages,
// including lost profits, arising out of the use of this OpenMMS, even
// if the University of California has been advised of the possibility of
// such damage.
//
// The OpenMMS provided herein is on an "as is" basis, and the
// University of California has no obligation to provide maintenance,
// support, updates, enhancements, or modifications.  The University of
// California makes no representations and extends no warranties of any
// kind, either implied or express, including, but not limited to, the
// implied warranties of merchantability or fitness for a particular
// purpose, or that the use of the OpenMMS will not infringe any patent,
// trademark or other rights.

package org.rcsb.openmms.apps.xconv;

import java.awt.*;
import java.io.*;
import java.net.*;
import java.util.*;

import org.rcsb.openmms.cifparse.*;

/**
 * Program to convert mmCIF representation of MMS data to XML format
 *
 * @author Douglas S. Greer
 * @version 1.1
 */
public class XmlConvert
{
  // basic status debug messages
  public static int debug = 1;

 /**
  * Set to true if we are writing a DTD for this Model
  */
  String xmlPath, dataPath;
  String cifListFile;
  String msaPackageName;
  // dct = List of Dictionary filenames 
  ArrayList dct;

  CifTokenizer ct;
  CifParser cifParser = new CifParser();
  CifDictionary cifDic = new CifDictionary();

 /**
  * Looks for CreateDTD, debug=, xmldir=, data=, dct= and filelist=
  * in argv input command line
  */
  public static void main(String[] argv)
    throws Exception 
    {
      XmlConvert xc = new XmlConvert();
      xc.init(argv);
      xc.execute(argv);
    }

  private void init(String[] argv)
    {

      String debugStr = parseArguments("debug", argv);
      if(debugStr != null)
	{
	  // debug <= 0 no messages
	  // debug == 1 basic status messages
	  // debug >= 5 trace
	  debug = Integer.parseInt(debugStr);
	}
      try
	{
	  xmlPath = parseArguments("xmldir", argv);
	  if(xmlPath == null )
	    {
	      throw new CifParseException
		("XmlConvert: Insufficient input - must specify xml pathname" +
		 "\n\t and xmp pathname (xmldir=<>)");
	    }
	  dataPath = parseArguments("data", argv);
	  if(dataPath == null )
	    {
	      throw new CifParseException
		("XmlConvert: Insufficient input - must specify data pathname" +
		 " (data=<>)");
	    }
	  dct = parseArgumentsMult("dct", argv);
	  if(dct.size() == 0)
	    {
	      throw new CifParseException
		("XmlConvert: Insufficient input " +
		 "- must specify dictionary (dct=<>)");
	    }
	  cifListFile = parseArguments("filelist", argv);
	}
      catch(CifParseException e)
	{
	  System.out.println("FATAL COMMAND LINE ERROR\n" + e);
	  System.exit(1);
	}
    }

  public void execute(String[] argv)
    {
      CifTokenizer ct = null;
      try
	{
	  for(int i = 0; i < dct.size(); i++)
	    {
	      String s = (String) dct.get(i);
	      System.out.println("Reading mmCIF dictionary from " + s);
	      try
		{
		  ct = new CifTokenizer(s);
		}
	      catch(IOException e)
		{
		  System.out.println("Error in opening File: " + s);
		  System.exit(1);
		}
	      cifParser.readDictionary(ct, cifDic, true);
	    }

	  // generate the xml files
	  writeFiles();

	  System.out.println("Finished");
	}
      catch(IOException e)
	{
	  System.out.println("\n>>> IOException <<<\n");
	  e.printStackTrace();
	  System.exit(1);
	}
      catch(XConvException e)
	{
	  System.out.println("\n>>> XmlGenException <<<\n");
	  e.printStackTrace();
	  System.exit(1);
	}
      catch(CifParseException e)
	{
	  System.out.println("\n>>> CifParseException <<<\n");
	  if (ct != null)
	    {
	      System.out.println("Fatal Error parsing Cif file: " 
		 + ct.filename() + " at line number " + ct.getLineNumber());
	    }
	  else
	    {
	      System.out.println("Fatal Error");
	    }
	  e.printStackTrace();
	  System.exit(1);
	}
    }
  void writeFiles() 
    throws IOException, XConvException
    {
      if (cifListFile != null)
	{
	  if (debug > 0)
	    {
	      System.out.println("Writing XML files");
	    }
	  // one Convert Builder for all entries
	  ConvertBuilder convBuilder = new ConvertBuilder(cifDic);
	  cifParser.setBuilder(convBuilder);

	  BufferedReader cf = new BufferedReader(new FileReader(cifListFile));
	  String line, baseName;
	  int px;
	  while ((line = cf.readLine()) != null)
	    {
	      line = line.trim();
	      if ( line.length() == 0 || line.charAt(0) == '#')
		{
		  continue;
		}
	      // parse file name
	      for (px = 0; px < line.length(); px++)
		{
		  if (line.charAt(px) == '.')
		    {
		      break;
		    }
		}
	      if (px >= line.length() 
		  || ! line.substring(px).equals(".cif"))
		{
		  throw new XConvException
		    ("XmlConvert.writeFiles: invalid Cif Input file name: "
		     + line);
		}
	      baseName = line.substring(0, px);

	      // convert data set
	      convertCifDataFile(baseName, convBuilder);
	    }
	}
    }

  // This routine is called once for each file that needs to be
  // converted. The same XConvBuilder and CifParser are used
  // for all file converstions
  private void convertCifDataFile(String base, ConvertBuilder xBuilder) 
    throws XConvException
    {
      String cifFPath;
      CifTokenizer ct = null;

      cifFPath = getCifDataFilePath(base);
      try
	{
	  ct = new CifTokenizer(cifFPath);
	  // write XML file in the current directory
	  XmlWriter writer = new XmlWriter(xmlPath, base);
	  xBuilder.setWriter(writer);
	  cifParser.readDataBlock(ct, cifDic.getDictionaryItemList());
	}
      catch(CifParseException e)
	{
	  if (ct != null)
	    {
		try
		  {
		    ct.close();
		  }
		catch (IOException e2)
		  { }
	    }
	  // something appears to be incorrect inside an mmCIF file
	    throw new XConvException
	      ("ParseException while reading cif file: "
	       + cifFPath + "\n" + e.getMessage());
	}
      catch(IOException e)
	{
	    if (ct != null)
	      {
		try
		  {
		    ct.close();
		  }
		catch (IOException e2)
		  { }
	      }
	    throw new XConvException
	      ("Cannot read file: " + cifFPath + "\n" + e.getMessage());
	}
    }

  public String getCifDataFilePath(String id)
    {
      return dataPath + "/" + idToGroup(id) + "/" + id + ".cif";
    }

  private String idToGroup(String id)
    {
      if (id == null)
	{
	  return null;
	}
      return id.substring(1,3);
    }

  private String parseArguments(String param, String[] argv)
    {
      for(int i=0; i<argv.length; i++)
	{
	  if(argv[i].startsWith(param))
	    {
	      int index=argv[i].indexOf('=');
	      if(index!=-1) return (argv[i].substring(index+1));
	    }
	}
      return null;
    }

  private boolean parseArgumentsBool(String param, String[] argv)
    {
      for(int i=0; i<argv.length; i++)
	if(argv[i].compareTo(param)==0) return true;

      return false;
    }

  private ArrayList parseArgumentsMult(String param, String[] argv)
    {
      ArrayList values = new ArrayList();
      for(int i=0; i<argv.length; i++)
	{
	  if(argv[i].startsWith(param))
	    {
	      int index=argv[i].indexOf('=');
	      if(index!=-1) 
		{
		  values.add(argv[i].substring(index+1));
		}
	    }
	}
      return values;
    }
}	
