//
// XmlWriter.java,v 1.1 2003/07/29 23:56:25 dsg Exp
//
// Copyright 2001 The Regents of the University of California
// All Rights Reserved
//
// OpenMMS was developed by Dr. Douglas S. Greer at the San Diego
// Supercomputer Center, a research unit of the University of California,
// San Diego.  Support for this effort was provided by NSF through the
// Protein Data Bank (Grant DBI-9814284) and the National Partnership for
// Advanced Computational Infrastructure (Grant ACI-9619020)
//
// Permission to use, copy, modify and distribute any part of OpenMMS for
// educational, research and non-profit purposes, without fee, and
// without a written agreement is hereby granted, provided that the above
// copyright notice, this paragraph and the following paragraphs appear
// in all copies.
//
// Those desiring to incorporate this OpenMMS into commercial products or
// use for commercial purposes should contact the Technology Transfer
// Office, University of California, San Diego, 9500 Gilman Drive, La
// Jolla, CA 92093-0910, Ph: (619) 534-5815, FAX: (619) 534-7345.
//
// In no event shall the University of California be liable to any party
// for direct, indirect, special, incidental, or consequential damages,
// including lost profits, arising out of the use of this OpenMMS, even
// if the University of California has been advised of the possibility of
// such damage.
//
// The OpenMMS provided herein is on an "as is" basis, and the
// University of California has no obligation to provide maintenance,
// support, updates, enhancements, or modifications.  The University of
// California makes no representations and extends no warranties of any
// kind, either implied or express, including, but not limited to, the
// implied warranties of merchantability or fitness for a particular
// purpose, or that the use of the OpenMMS will not infringe any patent,
// trademark or other rights.

package org.rcsb.openmms.apps.xconv;

import java.io.*;
import java.util.*;

import org.rcsb.openmms.cifparse.*;

/**
 * Model vistor base class
 *
 * @author Douglas S. Greer
 * @version 1.1
 */
public class XmlWriter
{
  private static final int MAX_LINE = 68;

  private String xPath;
  private String entryName;
  private String uCaseEId;

  private BufferedWriter outputFileWriter;
  private int indentLevel = 0;
  private static String indent = "   ";
  private boolean atNewLine = true;

  public XmlWriter(String xpath, String base)
    {
      xPath = xpath;
      entryName = base;
      uCaseEId = entryName.toUpperCase();
    }

  static public String parseCategoryName(String name)
      {
	String rval = null;
	int rndx = name.indexOf(".");
	if (rndx >= 0)
	  {
	    rval = name.substring(1, rndx);
	  }
	else
	  {
	    rval = name;
	  }
	return rval;
      }

  static public String parseItemName(String name)
      {
	String si = null;
	StringBuffer sb = new StringBuffer();
	int rndx = name.indexOf(".");
	for (int i = rndx + 1; i < name.length(); i++)
	  {
	    char c = name.charAt(i);
	    switch (c)
	      {
	      case '[':
	      case ']':
		// strip matrix and vector brackets
		break;
	      default:
		sb.append(c);
		break;
	      }
	  }
	return sb.toString();
      }

  public String getEntryName()
      throws IOException
    {
      return entryName;
    }

  public void openEntry()
      throws IOException
    {
      openFile(xPath + "/" + entryName + ".xml");
      writeOpening();
    }

  public void closeEntry()
      throws IOException
    {
      writeClosing();
      closeFile();
    }

  public void startCategory(String name)
      throws IOException
    {
      newLine();
      writeStr("<PDBx:" + name + "Category>");
      indentLevel++;
    }

  public void finishCategory(String name)
      throws IOException
    {
      indentLevel--;
      newLine();
      writeStr("</PDBx:" + name + "Category>");
    }

  public void writeRecord(Record r, String catName, DictionaryCategory dc)
      throws IOException
    {
      boolean bodyItem = false;
      ArrayList a = r.getArrayList();
      newLine();
      writeStr("<PDBx:" + catName);
      for (int i = 0; i < a.size(); i++)
	{
	  ItemElem ie = (ItemElem) a.get(i);
	  String fullName = ie.getDictionaryItem().getItemName();
	  boolean key = dc.isKey(fullName);
	  if ( ! key)
	    {
	      continue;
	    }
	  String name = parseItemName(fullName);
	  String val = ie.getValue();
	  writeAttributeElem(name, val, key);
	}
      writeStr(">");
      indentLevel++;
      for (int i = 0; i < a.size(); i++)
	{
	  ItemElem ie = (ItemElem) a.get(i);
	  String fullName = ie.getDictionaryItem().getItemName();
	  boolean key = dc.isKey(fullName);
	  if (key)
	    {
	      continue;
	    }
	  String name = parseItemName(fullName);
	  String val = ie.getValue();
	  writeBodyElem(name, val, key);
	  bodyItem = true;
	}
      indentLevel--;
      if (bodyItem)
	{
	  newLine();
	}
      writeStr("</PDBx:" + catName + ">");

    }

  public void writeAttributeElem(String name, String val, boolean key)
      throws IOException
    {
      writeStr(" " + name + "=\"" + val + "\"");
    }

  public void writeBodyElem(String name, String val, boolean key)
      throws IOException
    {
      newLine();
      writeStr("<PDBx:" + name + ">"
	       + val + "</PDBx:" + name + ">");
    }

  private void writeOpening()
      throws IOException
    {
      writeStr("<?xml version=\"1.0\" encoding=\"UTF-8\" ?>");
      newLine();
      writeStr("<PDBx:datablock  datablockName=\"" + uCaseEId + "\" "); 
      newLine();
      writeStr(
	  "xmlns:PDBx=\"http://deposit.pdb.org/pdbML/pdbx-v0.905.xsd\"", 4);
      newLine();
      writeStr(
	  "xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"", 4);
      newLine();
      writeStr(
	  "xsi:schemaLocation=\"http://deposit.pdb.org/pdbML/pdbx-v0.905.xsd pdbx-v0.905.xsd\">", 4);
      indentLevel++;
    }

  private void writeClosing()
      throws IOException
    {
      indentLevel--;
      newLine();
      writeStr("</PDBx:datablock>");
      newLine();
    }

  private void openFile(String fname)
      throws IOException
    {
      if (outputFileWriter != null)
	{
	  closeFile();
	}
      outputFileWriter = new BufferedWriter(new FileWriter(fname));
    }
  
  private void closeFile()
      throws IOException
    {
      outputFileWriter.close();
      outputFileWriter = null;
    }

  private void writeStr(String s)
      throws IOException
    {
      writeStr(outputFileWriter, s, 0);
    }

  private void writeStr(String s, int ilev)
      throws IOException
    {
      writeStr(outputFileWriter, s, ilev);
    }

  private void writeStr(String[] sl)
      throws IOException
    {
      for (int i = 0; i < sl.length; i++)
	{
	  writeStr(outputFileWriter, sl[i], 0);
	}
    }

  private void writeStr(BufferedWriter o, String s, int ilev)
      throws IOException
    {
      indentLevel += ilev;
      if (s != null && s.length() > 0)
	{
	  if (atNewLine)
	    {
	      for (int i = 0; i < indentLevel; i++)
		{
		  o.write(indent);
		}
	      atNewLine = false;
	    }
	  o.write(s);
	}
      indentLevel -= ilev;
    }

  private void newLine()
      throws IOException
    {
      outputFileWriter.newLine();
      atNewLine = true;
    }
}  


